{
    This file is part of the Free Pascal run time library.
    Copyright (c) 1999-2000 by Florian Klaempfl

    This file implements the win32 gui support for the graph unit

    See the file COPYING.FPC, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}
unit Graph;
interface

uses
  windows;

{

    This file is part of the Free Pascal run time library.
    Copyright (c) 1999-2000 by the Free Pascal development team

    Interface include file for graph unit

    See the file COPYING.FPC, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}
{-------------------------------------------------------}
{ Differences with TP Graph unit:                       }
{ -  default putimage and getimage only support a max.  }
{    of 64K colors on screen, because all pixels are    }
{    saved as words.                                    }
{ -  Set RGB Palette is not used, SetPalette must be    }
{    used instead.                                      }
{ -  In the TP graph unit, Clipping is always performed }
{    on strings written with OutText, and this clipping }
{    is done on a character per character basis (for    }
{    example, if ONE part of a character is outside the }
{    viewport , then that character is not written at   }
{    all to the screen. In FPC Pascal, clipping is done }
{    on a PIXEL basis, not a character basis, so part of }
{    characters which are not entirely in the viewport  }
{    may appear on the screen.                          }
{ -  SetTextStyle only conforms to the TP version when  }
{    the correct (and expected) values are used for     }
{    CharSize for stroked fonts (4 = stroked fonts)     }
{ -  InstallUserDriver is not supported, so always      }
{    returns an error.                                  }
{ -  RegisterBGIDriver is not supported, so always      }
{    returns an error.                                  }
{ - DrawPoly XORPut mode is not exactly the same as in  }
{   the TP graph unit.                                  }
{ - Imagesize returns a longint instead of a word       }
{ - ImageSize cannot return an error value              }
{-------------------------------------------------------}
{ AUTHORS:                                                                      }
{   Gernot Tenchio      - original version              }
{   Florian Klaempfl    - major updates                 }
{   Pierre Mueller      - major bugfixes                }
{   Carl Eric Codere    - complete rewrite              }
{   Thomas Schatzl      - optimizations,routines and    }
{                           suggestions.                }
{   Jonas Maebe         - bugfixes and optimizations    }
{ Credits (external):                                   }
{   - Original FloodFill code by                        }
{        Menno Victor van der star                      }
{     (the code has been heavily modified)              }
{-------------------------------------------------------}
{-------------------------------------------------------}
{ For significant speed improvements , is is recommended }
{ that these routines be hooked (otherwise the default,  }
{ slower routines will be used) :                        }
{   HLine()                                              }
{   VLine()                                              }
{   PatternLine()                                        }
{   ClearViewPort()                                      }
{   PutImage()                                           }
{   GetImage()  - ImageSize() should also be changed     }
{   InternalEllipse()                                    }
{   Line()                                               }
{   GetScanLine()                                        }
{--------------------------------------------------------}
{ FPC unit requirements:                                 }
{  All modes should at least have 1 graphics page to     }
{  make it possible to create animation on all supported }
{  systems , this can be done either by double-buffering }
{  yourself in the heap if no hardware is available to do}
{  it.                                                   }
{--------------------------------------------------------}
{ COMPATIBILITY WARNING: Some of the compatibility tests }
{ were done using the CGA and other the VGA drivers.     }
{ Within the BGI drivers themselves the BEHAVIOUR is not }
{ the same, so be warned!!!                              }
{--------------------------------------------------------}
{ History log:                                           }
{   15th February 1999:                                  }
{   + Added support for system font in vertical mode     }
{   + system font is now available for all platforms     }
{   * font support routines now compile                  }
{   * textHeight would not return correct size for system }
{     font                                               }
{   * Alignment of fonts partly fixed                    }
{   17th Feb. 1999:                                      }
{   + First support for stroked fonts                    }
{   18th Feb. 1999:                                      }
{   * bugfix of line drawing which fixes stroked font    }
{     displays.                                          }
{   23rd Feb. 1999:                                      }
{   + Applied Pierre's patches to font                   }
{   + Added scaling of bitmapped fonts                   }
{   + Vertical stroked fonts                             }
{  24th Feb. 1999:                                       }
{   * Scaling of stroked fonts must be done using FPs    }
{     to be 100% compatible with turbo pascal            }
{   + Sped up by 40% stroked font scaling calculations   }
{   + RegisterBGIFont                                    }
{  9th march 1999:                                       }
{   + Starting implementing Fillpoly()                   }
{  15th march 1999:                                      }
{   + SetFillStyle()                                     }
{   + patternLine()                                      }
{   + Bar()                                              }
{   * GraphDefaults would not make the Default color     }
{     of the fill pattern to the Max. Palette entry.     }
{   + SetFillPattern()                                   }
{  20th march 1999:                                      }
{   * GraphDefaults would not reset to the text system   }
{   * DefaultFont would write one character too much to  }
{     the screen                                         }
{   + Sloped thick lines in Line()                       }
{   + Sloped patterned lines in Line()                   }
{   * GraphDefaults would not reset the User Fill pattern}
{     to $ff                                             }
{   + DirectPutPixel takes care of XOR mode writes       }
{     improves speed by about 30% over old method of     }
{     GetPixel XOR CurrentColor                          }
{   * Dashed LineStyle exactly like BP version now       }
{   + Center LineStyle (checked against CGA driver)      }
{   * GraphDefaults() now resets linepattern array       }
{ 1st april  1999:                                       }
{   + First implementation of FillPoly (incomplete)      }
{ 2nd april  1999:                                       }
{   * FillPoly did not Reset PatternLine index           }
{   * FillPoly did not use correct color                 }
{   * PatternLine was writing modes in reverse direction }
{   * PatternLine would not work with non-rectangular    }
{     shapes.                                            }
{   * PatternLine must fill up the ENTIRE pattern,       }
{     with either the foreground or background color.    }
{   * GraphDefaults() would not call SetBkColor()        }
{   * Fixed some memory leaks in FillPoly()              }
{ 11th April  1999:                                      }
{   * PatternLine() was drawing one pixel less then      }
{     requested                                          }
{ 12th April  1999:                                      }
{   + FloodFill - first working implementation           }
{      Horrbly slow even on very fast cpu's              }
{   + Some suggestions of Thomas implemented             }
{ 13th April  1999:                                      }
{   * FloodFill() vertical index was off by one pixel    }
{   * FloodFill() would never draw the last line in the  }
{      list                                              }
{   - Removed ClearViewPort320 which was wrong anyways,  }
{     will need to be implemented later.                 }
{   * PatternLine() would not always restore write mode  }
{   + Circle() uses NormalPut always with NormWidth lines}
{   + FillEllipse() initial version                      }
{   * InternalEllipse() - 0 to 360 now supported as      }
{     angles.                                            }
{ 14th April  1999:                                      }
{   * mod x = and (x-1)(from Thomas Schatzl) gives a     }
{     significant speed improvement.                     }
{ 15th april  1999:                                      }
{   + Arc() ok except for Aspect Ratio, which does not   }
{     give us the correct ratio on a 320x200 screen.     }
{   + Added FillPoly() from Thomas Schatzl               }
{   + More hookable routines                             }
{  16th april  1999:                                     }
{   + Line() checked ok.                                 }
{  17th april  1999:                                     }
{   * GraphDefaults() would not reset CP                 }
{   + GetX(), GetY(), MoveTo() checked for viewports     }
{   * OutTextXY() should not update the CP               }
{   * ClearViewPort() would not update the CP            }
{   * ClearDevice() would not update the CP              }
{   * Sector() would update the CP by calling LineTo     }
{   * Bar3D() would update the CP                        }
{   * PieSlice() would update the CP                     }
{  18th april  1999:                                     }
{   + Clipping algorithm                                 }
{  19th april  1999:                                     }
{   + Adapterinfo structure                              }
{  20th april 1999:                                      }
{   + GetModeName                                        }
{   + GetGraphMode                                       }
{   + GetModeRange                                       }
{--------------------------------------------------------}


type
{$ifndef fpc}
    smallint = -32768..32767;
{$endif}

    TResolutionRec = record
      x,y: longint;
    end;

    const
       maxsmallint = high(smallint);
       { error codes }
       grOk =  0;
       grNoInitGraph = -1;
       grNotDetected = -2;
       grFileNotFound = -3;
       grInvalidDriver = -4;
       grNoLoadMem = -5;
       grNoScanMem = -6;
       grNoFloodMem = -7;
       grFontNotFound = -8;
       grNoFontMem = -9;
       grInvalidMode = -10;
       grError = -11;
       grIOerror = -12;
       grInvalidFont = -13;
       grInvalidFontNum = -14;
       grInvalidVersion = -18;


       { Color constants for setpalette }
       black     = 0;
       blue      = 1;
       green     = 2;
       cyan      = 3;
       red       = 4;
       magenta   = 5;
       brown     = 6;
       lightgray = 7;
       darkgray  = 8;
       lightblue = 9;
       lightgreen = 10;
       lightcyan = 11;
       lightred  = 12;
       lightmagenta = 13;
       yellow    = 14;
       white     = 15;

       EGABlack    =  0;
       EGABlue     =  1;
       EGAGreen    =  2;
       EGACyan     =  3;
       EGARed      =  4;
       EGAMagenta  =  5;
       EGALightgray=  7;
       EGABrown    = 20;
       EGADarkgray   = 56;
       EGALightblue  = 57;
       EGALightgreen = 58;
       EGALightcyan  = 59;
       EGALightred   = 60;
       EGALightmagenta=61;
       EGAYellow     = 62;
       EGAWhite      = 63;



       { Line styles for GetLineStyle/SetLineStyle }
       SolidLn = 0;
       DottedLn = 1;
       CenterLn = 2;
       DashedLn = 3;
       UserBitLn = 4;

       NormWidth = 1;
       ThickWidth = 3;

       { Set/GetTextStyle Konstanten: }
       DefaultFont = 0;
       TriplexFont = 1;
       SmallFont = 2;
       SansSerifFont = 3;
       GothicFont = 4;
       ScriptFont = 5;
       SimpleFont = 6;
       TSCRFont = 7;
       LCOMFont = 8;
       EuroFont = 9;
       BoldFont = 10;

       HorizDir = 0;
       VertDir = 1;

       UserCharSize = 0;

       ClipOn = true;
       ClipOff = false;

       { Bar3D constants }
       TopOn = true;
       TopOff = false;

       { fill pattern for Get/SetFillStyle: }
       EmptyFill      = 0;
       SolidFill      = 1;
       LineFill       = 2;
       LtSlashFill    = 3;
       SlashFill      = 4;
       BkSlashFill    = 5;
       LtBkSlashFill  = 6;
       HatchFill      = 7;
       XHatchFill     = 8;
       InterleaveFill = 9;
       WideDotFill    = 10;
       CloseDotFill   = 11;
       UserFill       = 12;

       { bitblt operators  }
       NormalPut     = 0;
       CopyPut       = 0;
       XORPut        = 1;
       OrPut         = 2;
       AndPut        = 3;
       NotPut        = 4;

       { SetTextJustify constants }
       LeftText   = 0;
       CenterText = 1;
       RightText  = 2;

       BottomText = 0;
       TopText    = 2;

       { graphic drivers }
       CurrentDriver = -128;
       Detect        = 0;
       CGA           = 1;
       MCGA          = 2;
       EGA           = 3;
       EGA64         = 4;
       EGAMono       = 5;
       LowRes        = 6; { nickysn: used to be 1, but moved to 6, because I added CGA }
       HercMono      = 7;
       VGA           = 9;
       VESA          = 10;


       D1bit = 11;
       D2bit = 12;
       D4bit = 13;
       D6bit = 14;  { 64 colors Half-brite mode - Amiga }
       D8bit = 15;
       D12bit = 16; { 4096 color modes HAM mode - Amiga }
       D15bit = 17;
       D16bit = 18;
       D24bit = 19; { not yet supported }
       D32bit = 20; { not yet supported }
       D64bit = 21; { not yet supported }

       lowNewDriver = 11;
       highNewDriver = 21;


       { graph modes }
       Default = 0;

       { CGA Driver modes }
       CGAC0 = 0;
       CGAC1 = 1;
       CGAC2 = 2;
       CGAC3 = 3;
       CGAHi = 4;

       { MCGA Driver modes }
       MCGAC0 = 0;
       MCGAC1 = 1;
       MCGAC2 = 2;
       MCGAC3 = 3;
       MCGAMed = 4;
       MCGAHi = 5;

       { EGA Driver modes }
       EGALo      = 0;  { 640x200 16 color 4 page }
       EGAHi      = 1;  { 640x350 16 color 2 page }

       { EGA64 Driver modes }
       EGA64Lo    = 0;  { 640x200 16 color 1 page }
       EGA64Hi    = 1;  { 640x350 4 color  1 page }

       { EGAMono Driver modes }
       EGAMonoHi  = 3;  { 640x350 64K on card, 1 page; 256K on card, 2 page }

       { VGA Driver modes }
       VGALo   = 0;
       VGAMed  = 1;
       VGAHi   = 2;

       { They start at such a high number to make sure they don't clash     }
       { with the mode numbers returned by getmoderange (those mode numbers }
       { are about the same modes, but all supported modes must have a      }
       { consecutive mode number for that, so they are generated on-the-fly }
       { starting from 1)                                                   }

       detectMode = 30000;
       m320x200 = 30001;
       m320x256 = 30002; { amiga resolution (PAL) }
       m320x400 = 30003; { amiga/atari resolution }
       m512x384 = 30004; { mac resolution }
       m640x200 = 30005; { vga resolution }
       m640x256 = 30006; { amiga resolution (PAL) }
       m640x350 = 30007; { vga resolution }
       m640x400 = 30008;
       m640x480 = 30009;
       m800x600 = 30010;
       m832x624 = 30011; { mac resolution }
       m1024x768 = 30012;
       m1280x1024 = 30013;
       m1600x1200 = 30014;
       m2048x1536 = 30015;

       lowNewMode = 30001;
       highNewMode = 30015;

       resolutions: array[lowNewMode..highNewMode] of TResolutionRec =
       ((x:320;y:200),
        (x:320;y:256),
        (x:320;y:400),
        (x:512;y:384),
        (x:640;y:200),
        (x:640;y:256),
        (x:640;y:350),
        (x:640;y:400),
        (x:640;y:480),
        (x:800;y:600),
        (x:832;y:624),
        (x:1024;y:768),
        (x:1280;y:1024),
        (x:1600;y:1200),
        (x:2048;y:1536));


       { Hercules mono card }
       HercMonoHi = 0;

       MaxColors   = 255;   { Maximum possible colors using a palette }
                            { otherwise, direct color encoding        }


    type
       RGBRec = packed record
         Red: smallint;
         Green: smallint;
         Blue : smallint;
       end;

       PaletteType = record
             Size   : longint;
             Colors : array[0..MaxColors] of RGBRec;
       end;

       LineSettingsType = record
             linestyle : word;
             pattern : word;
             thickness : word;
       end;

       TextSettingsType = record
             font : word;
             direction : word;
             charsize : word;
             horiz : word;
             vert : word;
       end;

       FillSettingsType = record
             pattern : word;
             color : word;
       end;

       FillPatternType = array[1..8] of byte;

       PointType = record
             x,y : smallint;
       end;

       ViewPortType = record
             x1,y1,x2,y2 : smallint;
             Clip : boolean;
       end;

       ArcCoordsType = record
             x,y : smallint;
             xstart,ystart : smallint;
             xend,yend : smallint;
       end;

        graph_float = single;   { the platform's preferred floating point size }

  const
       fillpatternTable : array[0..12] of FillPatternType = (
           ($00,$00,$00,$00,$00,$00,$00,$00),     { background color  }
           ($ff,$ff,$ff,$ff,$ff,$ff,$ff,$ff),     { foreground color  }
           ($ff,$ff,$00,$00,$ff,$ff,$00,$00),     { horizontal lines  }
           ($01,$02,$04,$08,$10,$20,$40,$80),     { slashes           }
           ($07,$0e,$1c,$38,$70,$e0,$c1,$83),     { thick slashes     }
           ($07,$83,$c1,$e0,$70,$38,$1c,$0e),     { thick backslashes }
           ($5a,$2d,$96,$4b,$a5,$d2,$69,$b4),     { backslashes       }
           ($ff,$88,$88,$88,$ff,$88,$88,$88),     { small boxes       }
           ($18,$24,$42,$81,$81,$42,$24,$18),     { rhombus           }
           ($cc,$33,$cc,$33,$cc,$33,$cc,$33),     { wall pattern      }
           ($80,$00,$08,$00,$80,$00,$08,$00),     { wide points       }
           ($88,$00,$22,$00,$88,$00,$22,$00),     { dense points      }
           (0,0,0,0,0,0,0,0)                      { user defined line style }
          );



  { ******************** PROCEDURAL VARIABLES ********************* }
  { * These are hooks which have device specific stuff in them,   * }
  { * therefore to add new modes or to redirect these routines    * }
  { * then declare variables of these types as shown below.       * }
  {-----------------------------------------------------------------}

TYPE


       { This is the standard putpixel routine used by all function }
       { drawing routines, it will use the viewport settings, as    }
       { well as clip, and use the current foreground color to plot }
       { the desired pixel.                                         }
       defpixelproc = procedure(X,Y: smallint);

       { standard plot and get pixel                                }
       getpixelproc = function(X,Y: smallint): word;
       putpixelproc = procedure(X,Y: smallint; Color: Word);

       { clears the viewport, also used to clear the device         }
       clrviewproc  = procedure;

       { putimage procedure, can be hooked to accomplish transparency }
       putimageproc = procedure (X,Y: smallint; var Bitmap; BitBlt: Word);
       getimageproc = procedure(X1,Y1,X2,Y2: smallint; Var Bitmap);
       imagesizeproc= function (X1,Y1,X2,Y2: smallint): longint;

       graphfreememprc = procedure (var P: Pointer; size: word);
       graphgetmemprc  = procedure (var P: pointer; size: word);

       { internal routines -- can be hooked for much faster drawing }

       { draw filled horizontal lines using current color }
       { on entry coordinates are already clipped.        }
       hlineproc = procedure (x, x2,y : smallint);
       { on entry coordinates are already clipped.        }
       { draw filled vertical line using current color    }
       vlineproc = procedure (x,y,y2: smallint);

       { this routine is used to draw filled patterns for all routines }
       { that require it. (FillPoly, FloodFill, Sector, etc...         }
       { clipping is verified, uses current Fill settings for drawing  }
       patternlineproc = procedure (x1,x2,y: smallint);

       { this routine is used to draw all circles/ellipses/sectors     }
       { more info... on this later...                                 }
       ellipseproc = procedure (X,Y: smallint;XRadius: word;
         YRadius:word; stAngle,EndAngle: word; fp: PatternLineProc);

       { Line routine - draws lines thick/norm widths with current     }
       { color and line style - LINE must be clipped here.             }
       lineproc = procedure (X1, Y1, X2, Y2 : smallint);

       { this routine is used for FloodFill - it returns an entire      }
       { screen scan line with a word for each pixel in the scanline.   }
       { Also handy for GetImage, so I added x coords as well (JM)      }
       getscanlineproc = procedure (X1, X2, Y : smallint; var data);

       { changes the active display screen where we draw to... }
       setactivepageproc = procedure (page: word);

       { changes the active display screen which we see ... }
       setvisualpageproc = procedure (page: word);

       { this routine actually switches to the desired video mode.     }
       initmodeproc = procedure;

       { this routine is called to save the sate just before a mode set }
       savestateproc = procedure;
       { this routine is called in closegraph to cleanup...             }
       restorestateproc = procedure;

       { This routine is a hook for SetRGBPalette                       }
       setrgbpaletteproc =
         procedure(ColorNum, RedValue, GreenValue, BlueValue: smallint);

       { This routine is a hook for GetRGBPalette                       }
       getrgbpaletteproc =
         procedure(ColorNum: smallint; var
            RedValue, GreenValue, BlueValue: smallint);

       SetAllPaletteProc = procedure(const Palette:PaletteType);

       OutTextXYProc = procedure(x,y : SmallInt;const TextString : string);

       CircleProc = procedure(X, Y: smallint; Radius:Word);

       SetBkColorProc = procedure(ColorNum: Word);
       GetBkColorProc = function: Word;


TYPE
    {-----------------------------------}
    { Linked list for mode information  }
    { This list is set up by one of the }
    { following routines:               }
    { It lists all available resolutions}
    { on this display adapter.          }
    {-----------------------------------}
    {   QueryAdapter()                  }
    {   DetectGraph()                   }
    {   InitGraph()                     }
    {-----------------------------------}
    PModeInfo = ^TModeInfo;
    TModeInfo = record
      DriverNumber: smallint;
      ModeNumber: smallint;
      internModeNumber: smallint;
      MaxColor: Longint;            { Maximum colors on screen        }
      PaletteSize : Longint;        { Maximum palette entry we can change }
      XAspect : word;            { XAspect ratio correction factor }
      YAspect : word;            { YAspect ratio correction factor }
      MaxX: word;                { Max-X row                       }
      MaxY: word;                { Max. column.                    }
      DirectColor: boolean;         { Is this a direct color mode??   }
      Hardwarepages: byte;          { total number of image pages - 1 }
      ModeName: String[18];
      { necessary hooks ... }
      DirectPutPixel : DefPixelProc;
      GetPixel       : GetPixelProc;
      PutPixel       : PutPixelProc;
      SetRGBPalette  : SetRGBPaletteProc;
      GetRGBPalette  : GetRGBPaletteProc;
      SetAllPalette  : SetAllPaletteProc;
      { defaults possible ... }
      SetVisualPage  : SetVisualPageProc;
      SetActivePage  : SetActivePageProc;
      ClearViewPort  : ClrViewProc;
      PutImage       : PutImageProc;
      GetImage       : GetImageProc;
      ImageSize      : ImageSizeProc;
      GetScanLine    : GetScanLineProc;
      Line           : LineProc;
      InternalEllipse: EllipseProc;
      PatternLine    : PatternLineProc;
      HLine          : HLineProc;
      VLine          : VLineProc;
      Circle         : CircleProc;
      InitMode       : InitModeProc;
      OutTextXY : OutTextXYProc;
      SetBKColor     : SetBkColorProc;
      GetBKColor     : GetBkColorProc;
      next: PModeInfo;
    end;

    TNewModeInfo = record
      modeInfo: array[lowNewDriver..highNewDriver] of PModeInfo;
      loHiModeNr: array[lowNewDriver..highNewDriver] of record
        lo,hi: smallint;
      end;
    end;



VAR
  DirectPutPixel : DefPixelProc;
  ClearViewPort  : ClrViewProc;
  PutPixel       : PutPixelProc;
  PutImage       : PutImageProc;
  GetImage       : GetImageProc;
  ImageSize      : ImageSizeProc;
  GetPixel       : GetPixelProc;
  SetVisualPage  : SetVisualPageProc;
  SetActivePage  : SetActivePageProc;
  SetRGBPalette  : SetRGBPaletteProc;
  GetRGBPalette  : GetRGBPaletteProc;
  SetAllPalette  : SetAllPaletteProc;
  OutTextXY      : OutTextXYProc;
  SetBkColor     : SetBkColorProc;
  GetBkColor     : GetBkColorProc;

  GraphFreeMemPtr: graphfreememprc;
  GraphGetMemPtr : graphgetmemprc;

  GetScanLine    : GetScanLineProc;
  Line           : LineProc;
  InternalEllipse: EllipseProc;
  PatternLine    : PatternLineProc;
  HLine          : HLineProc;
  VLine          : VLineProc;
  Circle         : CircleProc;

  SaveVideoState : SaveStateProc;
  RestoreVideoState: RestoreStateProc;


type
  TCharsetTransTable = array[Char] of Char;
  PCharsetTransTable = ^TCharsetTransTable;

const

  { The following table can be used for translating characters from the
    Ansi charset (ISO8859-1) to the DOS ASCII charset (CP437).
    To use this table, add the following line of code to your program:
    GraphStringTransTable := @AnsiToASCIITransTable;
  }

  AnsiToASCIITransTable: TCharsetTransTable =
    (#$00, #$01, #$02, #$03, #$04, #$05, #$06, #$07,   { $00 - $07 }
     #$08, #$09, #$0a, #$0b, #$0c, #$0d, #$0e, #$0f,   { $08 - $0f }
     #$10, #$11, #$12, #$13, #$14, #$15, #$16, #$17,   { $10 - $17 }
     #$18, #$19, #$1a, #$1b, #$1c, #$1d, #$1e, #$1f,   { $18 - $1f }
     #$20, #$21, #$22, #$23, #$24, #$25, #$26, #$27,   { $20 - $27 }
     #$28, #$29, #$2a, #$2b, #$2c, #$2d, #$2e, #$2f,   { $28 - $2f }
     #$30, #$31, #$32, #$33, #$34, #$35, #$36, #$37,   { $30 - $37 }
     #$38, #$39, #$3a, #$3b, #$3c, #$3d, #$3e, #$3f,   { $38 - $3f }
     #$40, #$41, #$42, #$43, #$44, #$45, #$46, #$47,   { $40 - $47 }
     #$48, #$49, #$4a, #$4b, #$4c, #$4d, #$4e, #$4f,   { $48 - $4f }
     #$50, #$51, #$52, #$53, #$54, #$55, #$56, #$57,   { $50 - $57 }
     #$58, #$59, #$5a, #$5b, #$5c, #$5d, #$5e, #$5f,   { $58 - $5f }
     #$60, #$61, #$62, #$63, #$64, #$65, #$66, #$67,   { $60 - $67 }
     #$68, #$69, #$6a, #$6b, #$6c, #$6d, #$6e, #$6f,   { $68 - $6f }
     #$70, #$71, #$72, #$73, #$74, #$75, #$76, #$77,   { $70 - $77 }
     #$78, #$79, #$7a, #$7b, #$7c, #$7d, #$7e, #$7f,   { $78 - $7f }
     '?' , '?' , '?' , '?' , '?' , '?' , '?' , '?' ,   { $80 - $87 }
     '?' , '?' , '?' , '?' , '?' , '?' , '?' , '?' ,   { $88 - $8f }
     '?' , '?' , '?' , '?' , '?' , '?' , '?' , '?' ,   { $90 - $97 }
     '?' , '?' , '?' , '?' , '?' , '?' , '?' , '?' ,   { $98 - $9f }
     #$ff, #$ad, #$9b, #$9c, '?' , #$9d, '?' , '?' ,   { $a0 - $a7 }
     '?' , '?' , #$a6, #$ae, #$aa, '?' , '?' , '?' ,   { $a8 - $af }
     #$f8, #$f1, #$fd, '?' , '?' , #$e6, '?' , #$fa,   { $b0 - $b7 }
     '?' , '?' , #$a7, #$af, #$ac, #$ab, '?' , #$a8,   { $b8 - $bf }
     '?' , '?' , '?' , '?' , #$8e, #$8f, #$92, #$80,   { $c0 - $c7 }
     '?' , #$90, '?' , '?' , '?' , '?' , '?' , '?' ,   { $c8 - $cf }
     '?' , #$a5, '?' , '?' , '?' , '?' , #$99, '?' ,   { $d0 - $d7 }
     '?' , '?' , '?' , '?' , #$9a, '?' , '?' , #$e1,   { $d8 - $df }
     #$85, #$a0, #$83, '?' , #$84, #$86, #$91, #$87,   { $e0 - $e7 }
     #$8a, #$82, #$88, #$89, #$8d, #$a1, #$8c, #$8b,   { $e8 - $ef }
     '?' , #$a4, #$95, #$a2, #$93, '?' , #$94, #$f6,   { $f0 - $f7 }
     '?' , #$97, #$a3, #$96, #$81, '?' , '?' , #$98);  { $f8 - $ff }


  GraphStringTransTable: PCharsetTransTable = nil;
  DrawTextBackground : boolean = false;

function queryadapterinfo : pmodeinfo;

Procedure Closegraph;
procedure SetLineStyle(LineStyle: word; Pattern: word; Thickness: word);
function  GraphErrorMsg(ErrorCode: smallint): string;
Function  GetMaxX: smallint;
Function  GetMaxY: smallint;
Procedure SetViewPort(X1, Y1, X2, Y2: smallint; Clip: Boolean);
Function  GraphResult: smallint;
function  GetModeName(ModeNumber: smallint): string;
procedure SetGraphMode(Mode: smallint);
function GetGraphMode: smallint;
function GetMaxMode: smallint;
procedure RestoreCrtMode;
procedure GetModeRange(GraphDriver: smallint; var LoMode, HiMode: smallint);
Function  GetX: smallint;
Function  GetY: smallint;
procedure GraphDefaults;
procedure ClearDevice;
procedure GetViewSettings(var viewport : ViewPortType);
procedure SetWriteMode(WriteMode : smallint);
procedure GetFillSettings(var Fillinfo:Fillsettingstype);
procedure GetFillPattern(var FillPattern:FillPatternType);
procedure GetLineSettings(var ActiveLineInfo : LineSettingsType);
procedure InitGraph(var GraphDriver:smallint;var GraphMode:smallint;const PathToDriver:String);
procedure DetectGraph(var GraphDriver:smallint;var GraphMode:smallint);
function InstallUserDriver(Name: string; AutoDetectPtr: Pointer): smallint;
function RegisterBGIDriver(driver: pointer): smallint;
procedure SetFillStyle(Pattern : word; Color: word);
procedure SetFillPattern(Pattern: FillPatternType; Color: word);
Function GetDriverName: string;
 procedure MoveRel(Dx, Dy: smallint);
 procedure MoveTo(X,Y: smallint);

 procedure SetDirectVideo(DirectAccess: boolean);
 function GetDirectVideo: boolean;

 { -------------------- Color/Palette ------------------------------- }
 function  GetColor: Word;
 procedure SetColor(Color: Word);
 function  GetMaxColor: word;

 procedure SetPalette(ColorNum: word; Color: shortint);
 procedure GetPalette(var Palette: PaletteType);
 function GetPaletteSize: smallint;
 procedure GetDefaultPalette(var Palette: PaletteType);


 { -------------------- Shapes/Lines -------------------------------- }
 procedure Rectangle(x1,y1,x2,y2:smallint);
 procedure Bar(x1,y1,x2,y2:smallint);
 procedure Bar3D(x1, y1, x2, y2 : smallint;depth : word;top : boolean);
 procedure FillPoly(NumPoints: word; Var PolyPoints);
 procedure DrawPoly(NumPoints : word;var polypoints);
 procedure LineRel(Dx, Dy: smallint);
 procedure LineTo(X,Y : smallint);
 procedure FloodFill(x : smallint; y : smallint; Border: word);

 { -------------------- Circle related routines --------------------- }
 procedure GetAspectRatio(var Xasp,Yasp : word);
 procedure SetAspectRatio(Xasp, Yasp : word);
 procedure GetArcCoords(var ArcCoords: ArcCoordsType);


 procedure Arc(X,Y : smallint; StAngle,EndAngle,Radius: word);
 procedure PieSlice(X,Y,stangle,endAngle:smallint;Radius: Word);
 procedure FillEllipse(X, Y: smallint; XRadius, YRadius: Word);
 procedure Sector(x, y: smallint; StAngle,EndAngle, XRadius, YRadius: Word);
 procedure Ellipse(X,Y : smallint; stAngle, EndAngle: word; XRadius,
   YRadius: word);

 { --------------------- Text related routines --------------------- }
 function  InstallUserFont(const FontFileName : string) : smallint;
 function  RegisterBGIfont(font : pointer) : smallint;
 procedure GetTextSettings(var TextInfo : TextSettingsType);
 function  TextHeight(const TextString : string) : word;
 function  TextWidth(const TextString : string) : word;
 procedure SetTextJustify(horiz,vert : word);
 procedure SetTextStyle(font,direction : word;charsize : word);
 procedure SetUserCharSize(Multx,Divx,Multy,Divy : word);

 procedure OutText(const TextString : string);




  var
    { this procedure allows to hook keyboard messages }
    charmessagehandler : WndProc;
    { this procedure allows to hook mouse messages }
    mousemessagehandler : WndProc;
    { this procedure allows to wm_command messages }
    commandmessagehandler : WndProc;
    NotifyMessageHandler : WndProc;

    OnGraphWindowCreation : procedure;

    GraphWindow,ParentWindow : HWnd;
    // this allows direct drawing to the window
    bitmapdc : hdc;
    windc : hdc;

  const
    { predefined window style }
    { we shouldn't set CS_DBLCLKS here }
    { because most dos applications    }
    { handle double clicks on it's own }
    graphwindowstyle : DWord = cs_hRedraw or cs_vRedraw;

    windowtitle : pchar = 'Graph window application';
    menu : hmenu = 0;
    icon : hicon = 0;
    drawtoscreen : boolean = true;
    drawtobitmap : boolean = true;
    // the graph window can be a child window, this allows to add toolbars
    // to the main window
    UseChildWindow : boolean = false;
    // this allows to specify an offset for the child child window
    ChildOffset : rect = (left:0;top:0;right:0;bottom:0);

CONST

  m640x200x16       = VGALo;
  m640x400x16       = VGAMed;
  m640x480x16       = VGAHi;

  { VESA Specific video modes. }
  m320x200x32k      = $10D;
  m320x200x64k      = $10E;

  m640x400x256      = $100;

  m640x480x256      = $101;
  m640x480x32k      = $110;
  m640x480x64k      = $111;

  m800x600x16       = $102;
  m800x600x256      = $103;
  m800x600x32k      = $113;
  m800x600x64k      = $114;

  m1024x768x16      = $104;
  m1024x768x256     = $105;
  m1024x768x32k     = $116;
  m1024x768x64k     = $117;

  m1280x1024x16     = $106;
  m1280x1024x256    = $107;
  m1280x1024x32k    = $119;
  m1280x1024x64k    = $11A;

  { some extra modes which applies only to GUI }
  mLargestWindow16  = $f0;
  mLargestWindow256 = $f1;
  mLargestWindow32k = $f2;
  mLargestWindow64k = $f3;
  mLargestWindow16M = $f4;
  mMaximizedWindow16 = $f5;
  mMaximizedWindow256 = $f6;
  mMaximizedWindow32k = $f7;
  mMaximizedWindow64k = $f8;
  mMaximizedWindow16M = $f9;


implementation

uses
  strings;

{
   Remarks:
      Colors in 16 color mode:
      ------------------------
        - the behavior of xor/or/and put isn't 100%:
          it is done using the RGB color getting from windows
          instead of the palette index!
        - palette operations aren't supported
      To solve these drawbacks, setpalette must be implemented
      by exchanging the colors in the DCs, further GetPaletteEntry
      must be used when doing xor/or/and operations
}


const
   InternalDriverName = 'WIN32GUI';

{
    This file is part of the Free Pascal run time library.
    Copyright (c) 1999-2000 by the Free Pascal development team

    Graph unit implementation part

    See the file COPYING.FPC, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}

var
  ExitSave: pointer;

const
  firstCallOfInitGraph: boolean = true;


{$ifdef logging}
var debuglog: text;

function strf(l: longint): string;
begin
  str(l, strf)
end;

Procedure Log(Const s: String);
Begin
  Append(debuglog);
  Write(debuglog, s);
  Close(debuglog);
End;

Procedure LogLn(Const s: string);
Begin
  Append(debuglog);
  Writeln(debuglog,s);
  Close(debuglog);
End;
{$endif logging}

const
   StdBufferSize = 4096;   { Buffer size for FloodFill }

type
  tinttable = array[0..16383] of smallint;
  pinttable = ^tinttable;

  WordArray = Array [0..StdbufferSize] Of word;
  PWordArray = ^WordArray;


const
   { Mask for each bit in byte used to determine pattern }
   BitArray: Array[0..7] of byte =
     ($01,$02,$04,$08,$10,$20,$40,$80);
   RevbitArray: Array[0..7] of byte =
     ($80,$40,$20,$10,$08,$04,$02,$01);

    { pre expanded line patterns }
    { 0 = LSB of byte pattern    }
    { 15 = MSB of byte pattern   }
    LinePatterns: Array[0..15] of BOOLEAN =
     (TRUE,TRUE,TRUE,TRUE,TRUE,TRUE,TRUE,TRUE,
      TRUE,TRUE,TRUE,TRUE,TRUE,TRUE,TRUE,TRUE);

const
  BGIPath : string = '.';


  { Default font 8x8 system from IBM PC }
  {
}

{******************************************}
{  Bitmapped font data - unrolled for      }
{  faster access. Each character is an     }
{  8x8 byte array representing the full    }
{  bitmap. 0 = nothing, 1 = color          }
{******************************************}


TYPE
  TBitMapChar = array[0..7,0..7] of byte;

CONST
   DefaultFontData: Array[#0..#255] of TBitmapChar = (
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,1,1,0),
(1,0,0,0,0,0,0,1),
(1,0,1,0,0,1,0,1),
(1,0,0,0,0,0,0,1),
(1,0,1,1,1,1,0,1),
(1,0,0,1,1,0,0,1),
(1,0,0,0,0,0,0,1),
(0,1,1,1,1,1,1,0)),
(
(0,1,1,1,1,1,1,0),
(1,1,1,1,1,1,1,1),
(1,1,0,1,1,0,1,1),
(1,1,1,1,1,1,1,1),
(1,1,0,0,0,0,1,1),
(1,1,1,0,0,1,1,1),
(1,1,1,1,1,1,1,1),
(0,1,1,1,1,1,1,0)),
(
(0,1,1,0,1,1,0,0),
(1,1,1,1,1,1,1,0),
(1,1,1,1,1,1,1,0),
(1,1,1,1,1,1,1,0),
(0,1,1,1,1,1,0,0),
(0,0,1,1,1,0,0,0),
(0,0,0,1,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,0,0,0,0),
(0,0,1,1,1,0,0,0),
(0,1,1,1,1,1,0,0),
(1,1,1,1,1,1,1,0),
(0,1,1,1,1,1,0,0),
(0,0,1,1,1,0,0,0),
(0,0,0,1,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,1,0,0,0),
(0,1,1,1,1,1,0,0),
(0,0,1,1,1,0,0,0),
(1,1,1,1,1,1,1,0),
(1,1,1,1,1,1,1,0),
(0,1,1,1,1,1,0,0),
(0,0,1,1,1,0,0,0),
(0,1,1,1,1,1,0,0)),
(
(0,0,0,1,0,0,0,0),
(0,0,0,1,0,0,0,0),
(0,0,1,1,1,0,0,0),
(0,1,1,1,1,1,0,0),
(1,1,1,1,1,1,1,0),
(0,1,1,1,1,1,0,0),
(0,0,1,1,1,0,0,0),
(0,1,1,1,1,1,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,1,1,1,1,0,0),
(0,0,1,1,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1),
(1,1,1,0,0,1,1,1),
(1,1,0,0,0,0,1,1),
(1,1,0,0,0,0,1,1),
(1,1,1,0,0,1,1,1),
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1)),
(
(0,0,0,0,0,0,0,0),
(0,0,1,1,1,1,0,0),
(0,1,1,0,0,1,1,0),
(0,1,0,0,0,0,1,0),
(0,1,0,0,0,0,1,0),
(0,1,1,0,0,1,1,0),
(0,0,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,1,1,1,1,1),
(1,1,0,0,0,0,1,1),
(1,0,0,1,1,0,0,1),
(1,0,1,1,1,1,0,1),
(1,0,1,1,1,1,0,1),
(1,0,0,1,1,0,0,1),
(1,1,0,0,0,0,1,1),
(1,1,1,1,1,1,1,1)),
(
(0,0,0,0,1,1,1,1),
(0,0,0,0,0,1,1,1),
(0,0,0,0,1,1,1,1),
(0,1,1,1,1,1,0,1),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0)),
(
(0,0,1,1,1,1,0,0),
(0,1,1,0,0,1,1,0),
(0,1,1,0,0,1,1,0),
(0,1,1,0,0,1,1,0),
(0,0,1,1,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,1,1,1,1,1,1,0),
(0,0,0,1,1,0,0,0)),
(
(0,0,1,1,1,1,1,1),
(0,0,1,1,0,0,1,1),
(0,0,1,1,1,1,1,1),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,1,0,0,0,0),
(1,1,1,1,0,0,0,0),
(1,1,1,0,0,0,0,0)),
(
(0,1,1,1,1,1,1,1),
(0,1,1,0,0,0,1,1),
(0,1,1,1,1,1,1,1),
(0,1,1,0,0,0,1,1),
(0,1,1,0,0,0,1,1),
(0,1,1,0,0,1,1,1),
(1,1,1,0,0,1,1,0),
(1,1,0,0,0,0,0,0)),
(
(1,0,0,1,1,0,0,1),
(0,1,0,1,1,0,1,0),
(0,0,1,1,1,1,0,0),
(1,1,1,0,0,1,1,1),
(1,1,1,0,0,1,1,1),
(0,0,1,1,1,1,0,0),
(0,1,0,1,1,0,1,0),
(1,0,0,1,1,0,0,1)),
(
(1,0,0,0,0,0,0,0),
(1,1,1,0,0,0,0,0),
(1,1,1,1,1,0,0,0),
(1,1,1,1,1,1,1,0),
(1,1,1,1,1,0,0,0),
(1,1,1,0,0,0,0,0),
(1,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,1,0),
(0,0,0,0,1,1,1,0),
(0,0,1,1,1,1,1,0),
(1,1,1,1,1,1,1,0),
(0,0,1,1,1,1,1,0),
(0,0,0,0,1,1,1,0),
(0,0,0,0,0,0,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,1,1,1,1,0,0),
(0,1,1,1,1,1,1,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,1,1,1,1,1,1,0),
(0,0,1,1,1,1,0,0),
(0,0,0,1,1,0,0,0)),
(
(0,1,1,0,0,1,1,0),
(0,1,1,0,0,1,1,0),
(0,1,1,0,0,1,1,0),
(0,1,1,0,0,1,1,0),
(0,1,1,0,0,1,1,0),
(0,0,0,0,0,0,0,0),
(0,1,1,0,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,1,1,1),
(1,1,0,1,1,0,1,1),
(1,1,0,1,1,0,1,1),
(0,1,1,1,1,0,1,1),
(0,0,0,1,1,0,1,1),
(0,0,0,1,1,0,1,1),
(0,0,0,1,1,0,1,1),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,1,1,1,0),
(0,1,1,0,0,0,1,1),
(0,0,1,1,1,0,0,0),
(0,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(0,0,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,1,1,0),
(0,1,1,1,1,1,1,0),
(0,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,1,1,1,1,0,0),
(0,1,1,1,1,1,1,0),
(0,0,0,1,1,0,0,0),
(0,1,1,1,1,1,1,0),
(0,0,1,1,1,1,0,0),
(0,0,0,1,1,0,0,0),
(1,1,1,1,1,1,1,1)),
(
(0,0,0,1,1,0,0,0),
(0,0,1,1,1,1,0,0),
(0,1,1,1,1,1,1,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,1,1,1,1,1,1,0),
(0,0,1,1,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,1,1,0,0),
(1,1,1,1,1,1,1,0),
(0,0,0,0,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0),
(1,1,1,1,1,1,1,0),
(0,1,1,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,1,0,0,1,0,0),
(0,1,1,0,0,1,1,0),
(1,1,1,1,1,1,1,1),
(0,1,1,0,0,1,1,0),
(0,0,1,0,0,1,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,1,1,1,1,0,0),
(0,1,1,1,1,1,1,0),
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1),
(0,1,1,1,1,1,1,0),
(0,0,1,1,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,1,1,1,1,0,0,0),
(0,1,1,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(1,1,1,1,1,1,1,0),
(0,1,1,0,1,1,0,0),
(1,1,1,1,1,1,1,0),
(0,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,0,0,0,0),
(0,1,1,1,1,1,0,0),
(1,1,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,1,1,0,0),
(1,1,1,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,1,0,0,0),
(0,1,1,0,1,1,0,0),
(0,0,1,1,1,0,0,0),
(0,1,1,1,0,1,1,0),
(1,1,0,1,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,0,0,0,0,0),
(0,1,1,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0),
(0,1,1,0,0,0,0,0),
(0,1,1,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,1,1,0,0,1,1,0),
(0,0,1,1,1,1,0,0),
(1,1,1,1,1,1,1,1),
(0,0,1,1,1,1,0,0),
(0,1,1,0,0,1,1,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,1,1,0),
(0,0,0,0,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,0,0,0,0),
(0,1,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(0,0,0,0,1,1,0,0),
(0,0,1,1,1,0,0,0),
(0,1,1,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(0,0,0,0,1,1,0,0),
(0,0,1,1,1,0,0,0),
(0,0,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,1,0,0),
(0,0,1,1,1,1,0,0),
(0,1,1,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,1,1,1,1,1,0),
(0,0,0,0,1,1,0,0),
(0,0,0,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,0,0,0),
(1,1,1,1,1,0,0,0),
(0,0,0,0,1,1,0,0),
(0,0,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,1,0,0,0),
(0,1,1,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,1,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,0,0,0,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,1,0,0),
(0,0,0,0,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,1,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(0,1,1,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(0,0,0,0,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,1,1,1,1,0),
(1,1,0,1,1,1,1,0),
(1,1,0,1,1,1,1,0),
(1,1,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,1,0,0,0),
(0,1,1,0,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,1,1,1,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,1,1,0),
(0,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,1,1,0,0),
(1,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,1,1,1,1,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,1,1,1,1,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,1,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,1,1,1,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,1,1,0),
(0,0,0,0,0,1,1,0),
(0,0,0,0,0,1,1,0),
(0,0,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,0,1,1,0),
(1,1,0,0,1,1,0,0),
(1,1,0,1,1,0,0,0),
(1,1,1,1,0,0,0,0),
(1,1,0,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,0,1,1,0),
(1,1,1,0,1,1,1,0),
(1,1,1,1,1,1,1,0),
(1,1,0,1,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,0,1,1,0),
(1,1,1,0,0,1,1,0),
(1,1,1,1,0,1,1,0),
(1,1,0,1,1,1,1,0),
(1,1,0,0,1,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,1,1,1,1,1,0,0),
(0,0,0,0,0,1,1,0)),
(
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,1,1,1,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,1,1,0,1,1,0,0),
(0,0,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,1,0,1,1,0),
(1,1,1,1,1,1,1,0),
(1,1,1,0,1,1,1,0),
(1,1,0,0,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,1,1,0,1,1,0,0),
(0,0,1,1,1,0,0,0),
(0,1,1,0,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,0,0,1,1),
(1,1,0,0,0,0,1,1),
(0,1,1,0,0,1,1,0),
(0,0,1,1,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,1,1,1,1,0),
(0,0,0,0,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,1,1,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,0,0,0,0),
(0,1,1,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,1,1,0,0),
(0,0,0,0,0,1,1,0),
(0,0,0,0,0,0,1,1),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,1,1,0,0),
(0,0,0,0,1,1,0,0),
(0,0,0,0,1,1,0,0),
(0,0,0,0,1,1,0,0),
(0,0,0,0,1,1,0,0),
(0,0,0,0,1,1,0,0),
(0,0,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,1,1,1,0,0,0),
(0,1,1,0,1,1,0,0),
(1,1,0,0,0,1,1,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,1,1)),
(
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,1,0,0),
(0,0,0,0,0,1,1,0),
(0,1,1,1,1,1,1,0),
(1,1,0,0,0,1,1,0),
(0,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,1,0,0,1,1,0),
(1,1,0,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(0,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,1,1,0),
(0,0,0,0,0,1,1,0),
(0,1,1,1,1,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,1,1,1,0),
(0,1,1,1,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,1,1,1,1,1,0),
(1,1,0,0,0,0,0,0),
(0,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,1,1,0),
(0,0,1,1,0,0,0,0),
(0,1,1,1,1,1,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,1,1,1,0),
(0,1,1,1,0,1,1,0),
(0,0,0,0,0,1,1,0),
(0,1,1,1,1,1,0,0)),
(
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,1,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,1,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(1,1,1,1,0,0,0,0)),
(
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,1,1,0,0,0),
(1,1,1,1,0,0,0,0),
(1,1,0,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,1,1,1,1,1,0),
(1,1,0,1,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,1,0,0,1,1,0),
(1,1,0,1,1,1,0,0),
(1,1,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,1,1,1,0),
(0,1,1,1,0,1,1,0),
(0,0,0,0,0,1,1,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,0,1,1,1,0),
(0,1,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0),
(0,1,1,0,0,0,0,0),
(0,1,1,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,1,0,0),
(1,1,0,0,0,0,0,0),
(0,1,1,1,1,1,0,0),
(0,0,0,0,0,1,1,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,1,1,1,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,1,1,0,1,1,0,0),
(0,0,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,1,0,1,1,0),
(1,1,1,1,1,1,1,0),
(0,1,1,0,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,0,0,0,1,1,0),
(0,1,1,0,1,1,0,0),
(0,0,1,1,1,0,0,0),
(0,1,1,0,1,1,0,0),
(1,1,0,0,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,1,1,1,0),
(0,1,1,1,0,1,1,0),
(0,0,0,0,0,1,1,0),
(0,1,1,1,1,1,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,1,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(1,1,1,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,1,1,1,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(1,1,1,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,0,1,1,0),
(1,1,0,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,1,0,0,0,0),
(0,0,1,1,1,0,0,0),
(0,1,1,0,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(0,0,0,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,1,1,0),
(1,1,0,0,0,0,1,1),
(0,0,1,1,1,1,0,0),
(0,0,0,0,0,1,1,0),
(0,0,1,1,1,1,1,0),
(0,1,1,0,0,1,1,0),
(0,0,1,1,1,1,1,1),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,1,1,0,0),
(0,1,1,1,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,1,1,0,0),
(0,1,1,1,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,1,1,0,0),
(0,1,1,1,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,1,1,0,0),
(0,0,1,1,1,0,0,0)),
(
(0,1,1,1,1,1,1,0),
(1,1,0,0,0,0,1,1),
(0,0,1,1,1,1,0,0),
(0,1,1,0,0,1,1,0),
(0,1,1,1,1,1,1,0),
(0,1,1,0,0,0,0,0),
(0,0,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,1,0,0),
(1,1,0,0,0,1,1,0),
(0,0,1,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,0,1,1,0),
(0,0,1,1,1,0,0,0),
(0,1,1,0,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,1,1,1,1,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,1,1,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,1,1,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,1,1,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,1,1,1),
(0,0,0,0,1,1,0,0),
(0,1,1,1,1,1,1,1),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,1,1,1),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,1,1,1,0),
(0,1,1,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,1,1,1,1,1,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(1,1,1,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(0,0,0,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(1,1,1,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(0,0,0,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,1,0,0),
(0,0,0,0,1,1,0,0),
(1,1,1,1,1,0,0,0)),
(
(1,1,0,0,0,0,1,1),
(0,0,0,1,1,0,0,0),
(0,0,1,1,1,1,0,0),
(0,1,1,0,0,1,1,0),
(0,1,1,0,0,1,1,0),
(0,0,1,1,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,1,1,0,0),
(0,0,0,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,1,1,1,1,1,1,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(0,1,1,1,1,1,1,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0)),
(
(0,0,1,1,1,0,0,0),
(0,1,1,0,1,1,0,0),
(0,1,1,0,0,1,0,0),
(1,1,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0),
(1,1,1,0,0,1,1,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,1,1,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0)),
(
(1,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,1,1,1,0,1,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,1,1,1,1),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,1)),
(
(0,0,0,0,1,1,1,0),
(0,0,0,1,1,0,1,1),
(0,0,0,1,1,0,0,0),
(0,0,1,1,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(1,1,0,1,1,0,0,0),
(0,1,1,1,0,0,0,0)),
(
(0,0,0,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,1,1,0,0),
(0,1,1,1,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,1,0,1,1,0,0),
(1,1,1,1,1,1,0,0),
(1,1,0,1,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,1,1,0,0),
(0,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(0,0,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,1,0,0,0),
(0,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(0,0,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,1,1,0,0),
(0,0,0,0,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,0,0,0,0,1,1),
(1,1,0,0,0,1,1,0),
(1,1,0,0,1,1,0,0),
(1,1,0,1,1,1,1,0),
(0,0,1,1,0,0,1,1),
(0,1,1,0,0,1,1,0),
(1,1,0,0,1,1,0,0),
(0,0,0,0,1,1,1,1)),
(
(1,1,0,0,0,0,1,1),
(1,1,0,0,0,1,1,0),
(1,1,0,0,1,1,0,0),
(1,1,0,1,1,0,1,1),
(0,0,1,1,0,1,1,1),
(0,1,1,0,1,1,1,1),
(1,1,0,0,1,1,1,1),
(0,0,0,0,0,0,1,1)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,1,1,0,0,1,1),
(0,1,1,0,0,1,1,0),
(1,1,0,0,1,1,0,0),
(0,1,1,0,0,1,1,0),
(0,0,1,1,0,0,1,1),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,0,0,1,1,0),
(0,0,1,1,0,0,1,1),
(0,1,1,0,0,1,1,0),
(1,1,0,0,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,0,0,0,1,0),
(1,0,0,0,1,0,0,0),
(0,0,1,0,0,0,1,0),
(1,0,0,0,1,0,0,0),
(0,0,1,0,0,0,1,0),
(1,0,0,0,1,0,0,0),
(0,0,1,0,0,0,1,0),
(1,0,0,0,1,0,0,0)),
(
(0,1,0,1,0,1,0,1),
(1,0,1,0,1,0,1,0),
(0,1,0,1,0,1,0,1),
(1,0,1,0,1,0,1,0),
(0,1,0,1,0,1,0,1),
(1,0,1,0,1,0,1,0),
(0,1,0,1,0,1,0,1),
(1,0,1,0,1,0,1,0)),
(
(1,1,0,1,1,0,1,1),
(0,1,1,1,0,1,1,1),
(1,1,0,1,1,0,1,1),
(1,1,1,0,1,1,1,0),
(1,1,0,1,1,0,1,1),
(0,1,1,1,0,1,1,1),
(1,1,0,1,1,0,1,1),
(1,1,1,0,1,1,1,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(1,1,1,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(1,1,1,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(1,1,1,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0)),
(
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(1,1,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(1,1,1,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0)),
(
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(1,1,1,1,0,1,1,0),
(0,0,0,0,0,1,1,0),
(1,1,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0)),
(
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,1,0),
(0,0,0,0,0,1,1,0),
(1,1,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0)),
(
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(1,1,1,1,0,1,1,0),
(0,0,0,0,0,1,1,0),
(1,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(1,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(1,1,1,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(1,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,1,1,1),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(1,1,1,1,1,1,1,1),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,1,1),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,1,1,1),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,1,1),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(1,1,1,1,1,1,1,1),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,1,1,1),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,1,1,1),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0)),
(
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,1),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0)),
(
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,1),
(0,0,1,1,0,0,0,0),
(0,0,1,1,1,1,1,1),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,1,1,1,1,1,1),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,1,1,1),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0)),
(
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(1,1,1,1,0,1,1,1),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,1,1),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,1,1),
(0,0,0,0,0,0,0,0),
(1,1,1,1,0,1,1,1),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0)),
(
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,1),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,1,1,1),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,1,1),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,1,1),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(1,1,1,1,0,1,1,1),
(0,0,0,0,0,0,0,0),
(1,1,1,1,0,1,1,1),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(1,1,1,1,1,1,1,1),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,1,1),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(1,1,1,1,1,1,1,1),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,1,1),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,1,1),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,1,1),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0)),
(
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,1,1,1,1),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,1,1,1),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,1,1,1),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,1,1,1,1,1),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,1,1,1),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,1,1,1,1,1,1),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0)),
(
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(1,1,1,1,1,1,1,1),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0),
(0,0,1,1,0,1,1,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(1,1,1,1,1,1,1,1),
(0,0,0,1,1,0,0,0),
(1,1,1,1,1,1,1,1),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(1,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,1,1,1,1,1),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0)),
(
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1)),
(
(1,1,1,1,0,0,0,0),
(1,1,1,1,0,0,0,0),
(1,1,1,1,0,0,0,0),
(1,1,1,1,0,0,0,0),
(1,1,1,1,0,0,0,0),
(1,1,1,1,0,0,0,0),
(1,1,1,1,0,0,0,0),
(1,1,1,1,0,0,0,0)),
(
(0,0,0,0,1,1,1,1),
(0,0,0,0,1,1,1,1),
(0,0,0,0,1,1,1,1),
(0,0,0,0,1,1,1,1),
(0,0,0,0,1,1,1,1),
(0,0,0,0,1,1,1,1),
(0,0,0,0,1,1,1,1),
(0,0,0,0,1,1,1,1)),
(
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1),
(1,1,1,1,1,1,1,1),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,0,1,1,0),
(1,1,0,1,1,1,0,0),
(1,1,0,0,1,0,0,0),
(1,1,0,1,1,1,0,0),
(0,1,1,1,0,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,1,1,1,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,1,0),
(0,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,1,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,1,1,0),
(1,1,0,1,1,0,0,0),
(1,1,0,1,1,0,0,0),
(1,1,0,1,1,0,0,0),
(0,1,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,1,1,0,0,1,1,0),
(0,1,1,0,0,1,1,0),
(0,1,1,0,0,1,1,0),
(0,1,1,0,0,1,1,0),
(0,1,1,1,1,1,0,0),
(0,1,1,0,0,0,0,0),
(1,1,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,1,1,1,0,1,1,0),
(1,1,0,1,1,1,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(1,1,1,1,1,1,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(1,1,1,1,1,1,0,0)),
(
(0,0,1,1,1,0,0,0),
(0,1,1,0,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,1,1,1,1,1,0),
(1,1,0,0,0,1,1,0),
(0,1,1,0,1,1,0,0),
(0,0,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,1,0,0,0),
(0,1,1,0,1,1,0,0),
(1,1,0,0,0,1,1,0),
(1,1,0,0,0,1,1,0),
(0,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(1,1,1,0,1,1,1,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,1,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,1,1,1,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,1,1,1,0),
(1,1,0,1,1,0,1,1),
(1,1,0,1,1,0,1,1),
(0,1,1,1,1,1,1,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,1,1,0),
(0,0,0,0,1,1,0,0),
(0,1,1,1,1,1,1,0),
(1,1,0,1,1,0,1,1),
(1,1,0,1,1,0,1,1),
(0,1,1,1,1,1,1,0),
(0,1,1,0,0,0,0,0),
(1,1,0,0,0,0,0,0)),
(
(0,0,1,1,1,0,0,0),
(0,1,1,0,0,0,0,0),
(1,1,0,0,0,0,0,0),
(1,1,1,1,1,0,0,0),
(1,1,0,0,0,0,0,0),
(0,1,1,0,0,0,0,0),
(0,0,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,1,0,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(1,1,0,0,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,1,1,1,0),
(0,0,0,1,1,0,1,1),
(0,0,0,1,1,0,1,1),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0)),
(
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(1,1,0,1,1,0,0,0),
(1,1,0,1,1,0,0,0),
(0,1,1,1,0,0,0,0)),
(
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0),
(1,1,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,1,1,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,1,1,1,0,1,1,0),
(1,1,0,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,1,1,1,0,1,1,0),
(1,1,0,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,1,1,1,0,0,0),
(0,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(0,0,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,1,1,1,1),
(0,0,0,0,1,1,0,0),
(0,0,0,0,1,1,0,0),
(0,0,0,0,1,1,0,0),
(1,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(0,0,1,1,1,1,0,0),
(0,0,0,1,1,1,0,0)),
(
(0,1,1,1,1,0,0,0),
(0,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(0,1,1,0,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,1,1,1,0,0,0,0),
(0,0,0,1,1,0,0,0),
(0,0,1,1,0,0,0,0),
(0,1,1,0,0,0,0,0),
(0,1,1,1,1,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,1,1,1,1,0,0),
(0,0,1,1,1,1,0,0),
(0,0,1,1,1,1,0,0),
(0,0,1,1,1,1,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0)),
(
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0),
(0,0,0,0,0,0,0,0))
);





var
  CurrentColor:     Word;
  CurrentBkColor: Word;
  CurrentX : smallint;   { viewport relative }
  CurrentY : smallint;   { viewport relative }

  ClipPixels: Boolean;  { Should cliiping be enabled }


  CurrentWriteMode: smallint;


  _GraphResult : smallint;


  LineInfo : LineSettingsType;
  FillSettings: FillSettingsType;

  { information for Text Output routines }
  CurrentTextInfo : TextSettingsType;
  CurrentXRatio, CurrentYRatio: graph_float;
  installedfonts: longint;  { Number of installed fonts }


  StartXViewPort: smallint; { absolute }
  StartYViewPort: smallint; { absolute }
  ViewWidth : smallint;
  ViewHeight: smallint;


  IsGraphMode : Boolean; { Indicates if we are in graph mode or not }


  ArcCall: ArcCoordsType;   { Information on the last call to Arc or Ellipse }


var

  { ******************** HARDWARE INFORMATION ********************* }
  { Should be set in InitGraph once only.                           }
  IntCurrentMode : smallint;
  IntCurrentDriver : smallint;       { Currently loaded driver          }
  IntCurrentNewDriver: smallint;
  XAspect : word;
  YAspect : word;
  MaxX : smallint;       { Maximum resolution - ABSOLUTE }
  MaxY : smallint;       { Maximum resolution - ABSOLUTE }
  MaxColor : Longint;
  PaletteSize : longint; { Maximum palette entry we can set, usually equal}
                         { maxcolor.                                      }
  HardwarePages : byte;  { maximum number of hardware visual pages        }
  DriverName: String;
  DirectColor : Boolean ; { Is it a direct color mode? }
  ModeList : PModeInfo;
  newModeList: TNewModeInfo;
  DirectVideo : Boolean;  { Direct access to video memory? }




{--------------------------------------------------------------------------}
{                                                                          }
{                    LINE AND LINE RELATED ROUTINES                        }
{                                                                          }
{--------------------------------------------------------------------------}

  {
    This file is part of the Free Pascal run time library.
    Copyright (c) 1999-2000 by the Free Pascal development team

    This include implements the different clipping algorithms

    See the file COPYING.FPC, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}
const
  LEFT   = 1;    { Left window   }
  RIGHT  = 2;    { Right window  }
  BOTTOM = 4;    { Bottom window }
  TOP    = 8;    { Top window    }
                 { 0 = in window }







  function LineClipped(var x1, y1,x2,y2: smallint; xmin, ymin,
      xmax, ymax:smallint): boolean;
  {********************************************************}
  { Function LineClipped()                                 }
  {--------------------------------------------------------}
  { This routine clips the line coordinates to the         }
  { min. and max. values of the window. Returns TRUE if    }
  { the ENTIRE line was clipped.  Updated                  }
  { clipped line endpoints are also returned.              }
  { This algorithm is the classic Cohen-Sutherland line    }
  { clipping algorithm.                                    }
  {--------------------------------------------------------}
  var
   code1, code2: longint;
   code: longint;
   newx,newy: smallint;
   done:boolean;


    function outcode(x,y:smallint): longint;
    {********************************************************}
    { Function OutCode()                                     }
    {--------------------------------------------------------}
    { This routine determines if the specified end point     }
    { of a line lies within the visible window, if not it    }
    { determines in which window the point is.               }
    {--------------------------------------------------------}

    var
     code: longint;
    begin
      code := 0;
      if (x<xmin) then
        code:=code or LEFT
      else if (x>xmax) then
        code:=code or RIGHT;
      if (y>ymax) then
        code:=code or BOTTOM
      else if (y<ymin) then
        code:=code or TOP;

      outcode:=code;
    end;

  begin
    done:=false;
    code1:= OutCode(x1,y1);
    code2:= OutCode(x2,y2);

    while not done do
     begin
       { Accept trivially }
       { both points are in window }
       if ((code1=0) and (code2=0)) then
         begin
           done:=TRUE;
           LineClipped:=FALSE;
               exit;
             end
       else
       { Reject trivially }
       { Neither points are in window }
       if (code1 and code2) <> 0 then
         begin
           done:=true;
           LineClipped:=TRUE;
           exit;
         end
       else
          begin
            { Some points are partially out of the window }
            { find the new end point of the lines...      }
            if code1 = 0 then
             code:=code2
            else
             code:=code1;
            if (code and LEFT) <> 0 then
              begin
                newy:=y1+((y2-y1)*(xmin-x1)) div (x2-x1);
                newx:=xmin;
              end
            else
            if (code and RIGHT) <> 0 then
              begin
                newy:=y1+((y2-y1)*(xmax-x1)) div (x2-x1);
                newx:=xmax;
              end
            else
            if (code and BOTTOM) <> 0 then
              begin
                newx:=x1+((x2-x1)*(ymax-y1)) div (y2-y1);
                newy:=ymax;
              end
            else
            if (code and TOP) <> 0 then
              begin
                newx:=x1+((x2-x1)*(ymin-y1)) div (y2-y1);
                newy:=ymin;
              end;
           if (code1 = code) then
            begin
              x1 := newx;  y1:= newy;
              code1:=outcode(x1,y1)
            end
               else
            begin
              x2:= newx; y2:= newy;
              code2:=outcode(x2,y2);
            end
         end;
      end;
  LineClipped:=FALSE;
end;



  procedure HLineDefault(x,x2,y: smallint); {$ifndef fpc}far;{$endif fpc}

   var
    xtmp: smallint;
   Begin

    { must we swap the values? }
    if x >= x2 then
      Begin
        xtmp := x2;
        x2 := x;
        x:= xtmp;
      end;
    { First convert to global coordinates }
    X   := X + StartXViewPort;
    X2  := X2 + StartXViewPort;
    Y   := Y + StartYViewPort;
    if ClipPixels then
      Begin
         if LineClipped(x,y,x2,y,StartXViewPort,StartYViewPort,
                StartXViewPort+ViewWidth, StartYViewPort+ViewHeight) then
            exit;
      end;
    for x:= x to x2 do
      DirectPutPixel(X,Y);
   end;


  procedure VLineDefault(x,y,y2: smallint); {$ifndef fpc}far;{$endif fpc}

   var
    ytmp: smallint;
  Begin
    { must we swap the values? }
    if y >= y2 then
     Begin
       ytmp := y2;
       y2 := y;
       y:= ytmp;
     end;
    { First convert to global coordinates }
    X   := X + StartXViewPort;
    Y2  := Y2 + StartYViewPort;
    Y   := Y + StartYViewPort;
    if ClipPixels then
      Begin
         if LineClipped(x,y,x,y2,StartXViewPort,StartYViewPort,
                StartXViewPort+ViewWidth, StartYViewPort+ViewHeight) then
            exit;
      end;
    for y := y to y2 do Directputpixel(x,y)
  End;

  Procedure DirectPutPixelClip(x,y: smallint);
  { for thickwidth lines, because they may call DirectPutPixel for coords }
  { outside the current viewport (bug found by CEC)                       }
  Begin
    If (Not ClipPixels) Or
       ((X >= StartXViewPort) And (X <= (StartXViewPort + ViewWidth)) And
        (Y >= StartYViewPort) And (Y <= (StartYViewPort + ViewHeight))) then
      Begin
        DirectPutPixel(x,y)
      End
  End;

  procedure LineDefault(X1, Y1, X2, Y2: smallint); {$ifndef fpc}far;{$endif fpc}

  var X, Y :           smallint;
      deltax, deltay : smallint;
      d, dinc1, dinc2: smallint;
      xinc1          : smallint;
      xinc2          : smallint;
      yinc1          : smallint;
      yinc2          : smallint;
      i              : smallint;
      Flag           : Boolean; { determines pixel direction in thick lines }
      NumPixels      : smallint;
      PixelCount     : smallint;
      OldCurrentColor: Word;
      swtmp          : smallint;
      TmpNumPixels   : smallint;
 begin
{******************************************}
{  SOLID LINES                             }
{******************************************}
  if lineinfo.LineStyle = SolidLn then
    Begin
       { we separate normal and thick width for speed }
       { and because it would not be 100% compatible  }
       { with the TP graph unit otherwise             }
       if y1 = y2 then
        Begin
     {******************************************}
     {  SOLID LINES HORIZONTAL                  }
     {******************************************}
          if lineinfo.Thickness=NormWidth then
            hline(x1,x2,y2)
          else
            begin
               { thick width }
               hline(x1,x2,y2-1);
               hline(x1,x2,y2);
               hline(x2,x2,y2+1);
            end;
        end
    else
    if x1 = x2 then
        Begin
     {******************************************}
     {  SOLID LINES VERTICAL                    }
     {******************************************}
          if lineinfo.Thickness=NormWidth then
            vline(x1,y1,y2)
          else
            begin
            { thick width }
              vline(x1-1,y1,y2);
              vline(x1,y1,y2);
              vline(x1+1,y1,y2);
            end;
        end
    else
    begin
     { Convert to global coordinates. }
     x1 := x1 + StartXViewPort;
     x2 := x2 + StartXViewPort;
     y1 := y1 + StartYViewPort;
     y2 := y2 + StartYViewPort;
     { if fully clipped then exit... }
     if ClipPixels then
       begin
       if LineClipped(x1,y1,x2,y2,StartXViewPort, StartYViewPort,
           StartXViewPort+ViewWidth, StartYViewPort+ViewHeight) then
              exit;
       end;
     {******************************************}
     {  SLOPED SOLID LINES                      }
     {******************************************}
           oldCurrentColor :=
           CurrentColor;
           { Calculate deltax and deltay for initialisation }
           deltax := abs(x2 - x1);
           deltay := abs(y2 - y1);

          { Initialize all vars based on which is the independent variable }
          if deltax >= deltay then
            begin

             Flag := FALSE;
             { x is independent variable }
             numpixels := deltax + 1;
             d := (2 * deltay) - deltax;
             dinc1 := deltay Shl 1;
             dinc2 := (deltay - deltax) shl 1;
             xinc1 := 1;
             xinc2 := 1;
             yinc1 := 0;
             yinc2 := 1;
            end
          else
            begin

             Flag := TRUE;
             { y is independent variable }
             numpixels := deltay + 1;
             d := (2 * deltax) - deltay;
             dinc1 := deltax Shl 1;
             dinc2 := (deltax - deltay) shl 1;
             xinc1 := 0;
             xinc2 := 1;
             yinc1 := 1;
             yinc2 := 1;
            end;

         { Make sure x and y move in the right directions }
         if x1 > x2 then
           begin
            xinc1 := - xinc1;
            xinc2 := - xinc2;
           end;
         if y1 > y2 then
          begin
           yinc1 := - yinc1;
           yinc2 := - yinc2;
          end;

         { Start drawing at <x1, y1> }
         x := x1;
         y := y1;


         If LineInfo.Thickness=NormWidth then

          Begin

            { Draw the pixels }
            for i := 1 to numpixels do
              begin
                DirectPutPixel(x, y);
                if d < 0 then
                  begin
                   d := d + dinc1;
                   x := x + xinc1;
                   y := y + yinc1;
                  end
                else
                  begin
                   d := d + dinc2;
                   x := x + xinc2;
                   y := y + yinc2;
                  end;
                  CurrentColor := OldCurrentColor;
             end;
          end
        else
         { Thick width lines }
          begin
            { Draw the pixels }
             for i := 1 to numpixels do
               begin
                { all depending on the slope, we can determine         }
                { in what direction the extra width pixels will be put }
                If Flag then
                  Begin
                    DirectPutPixelClip(x-1,y);
                    DirectPutPixelClip(x,y);
                    DirectPutPixelClip(x+1,y);
                  end
                else
                  Begin
                    DirectPutPixelClip(x, y-1);
                    DirectPutPixelClip(x, y);
                    DirectPutPixelClip(x, y+1);
                  end;
                if d < 0 then
                  begin
                    d := d + dinc1;
                    x := x + xinc1;
                    y := y + yinc1;
                  end
                else
                  begin
                    d := d + dinc2;
                    x := x + xinc2;
                    y := y + yinc2;
                  end;
                CurrentColor := OldCurrentColor;
               end;
          end;
        end;
  end
   else
{******************************************}
{  begin patterned lines                   }
{******************************************}
    Begin
      { Convert to global coordinates. }
      x1 := x1 + StartXViewPort;
      x2 := x2 + StartXViewPort;
      y1 := y1 + StartYViewPort;
      y2 := y2 + StartYViewPort;
      { if fully clipped then exit... }
      if ClipPixels then
       begin
       if LineClipped(x1,y1,x2,y2,StartXViewPort, StartYViewPort,
           StartXViewPort+ViewWidth, StartYViewPort+ViewHeight) then
              exit;
       end;

      OldCurrentColor := CurrentColor;
      PixelCount:=0;
      if y1 = y2 then
            Begin
             { Check if we must swap }
         if x1 >= x2 then
               Begin
                 swtmp := x1;
                 x1 := x2;
                 x2 := swtmp;
               end;
         if LineInfo.Thickness = NormWidth then
              Begin
               for PixelCount:=x1 to x2 do
                     { optimization: PixelCount mod 16 }
                     if LinePatterns[PixelCount and 15] = TRUE then
                      begin
                        DirectPutPixel(PixelCount,y2);
                      end;
              end
             else
              Begin
               for i:=-1 to 1 do
                     Begin
                       for PixelCount:=x1 to x2 do
                         { Optimization from Thomas - mod 16 = and 15 }
                         {this optimization has been performed by the compiler
                          for while as well (JM)}
                         if LinePatterns[PixelCount and 15] = TRUE then
                           begin
                                 DirectPutPixelClip(PixelCount,y2+i);
                           end;
                     end;
              end;
        end
      else
      if x1 = x2 then
           Begin
            { Check if we must swap }
            if y1 >= y2 then
              Begin
                swtmp := y1;
                y1 := y2;
                y2 := swtmp;
              end;
            if LineInfo.Thickness = NormWidth then
              Begin
                for PixelCount:=y1 to y2 do
                    { compare if we should plot a pixel here , compare }
                    { with predefined line patterns...                 }
                    if LinePatterns[PixelCount and 15] = TRUE then
                      begin
                        DirectPutPixel(x1,PixelCount);
                      end;
              end
            else
              Begin
                for i:=-1 to 1 do
                     Begin
                       for PixelCount:=y1 to y2 do
                       { compare if we should plot a pixel here , compare }
                       { with predefined line patterns...                 }
                         if LinePatterns[PixelCount and 15] = TRUE then
                           begin
                             DirectPutPixelClip(x1+i,PixelCount);
                           end;
                     end;
              end;
           end
      else
           Begin
             oldCurrentColor := CurrentColor;
             { Calculate deltax and deltay for initialisation }
             deltax := abs(x2 - x1);
             deltay := abs(y2 - y1);

             { Initialize all vars based on which is the independent variable }
             if deltax >= deltay then
               begin

                 Flag := FALSE;
                 { x is independent variable }
                 numpixels := deltax + 1;
                 d := (2 * deltay) - deltax;
                 dinc1 := deltay Shl 1;
                 dinc2 := (deltay - deltax) shl 1;
                 xinc1 := 1;
                 xinc2 := 1;
                 yinc1 := 0;
                 yinc2 := 1;
              end
            else
              begin

                Flag := TRUE;
                { y is independent variable }
                numpixels := deltay + 1;
                d := (2 * deltax) - deltay;
                dinc1 := deltax Shl 1;
                dinc2 := (deltax - deltay) shl 1;
                xinc1 := 0;
                xinc2 := 1;
                yinc1 := 1;
                yinc2 := 1;
              end;

            { Make sure x and y move in the right directions }
            if x1 > x2 then
              begin
                xinc1 := - xinc1;
                xinc2 := - xinc2;
              end;
            if y1 > y2 then
              begin
                yinc1 := - yinc1;
                yinc2 := - yinc2;
              end;

            { Start drawing at <x1, y1> }
            x := x1;
            y := y1;

            If LineInfo.Thickness=ThickWidth then

             Begin
               TmpNumPixels := NumPixels-1;
               { Draw the pixels }
               for i := 0 to TmpNumPixels do
                 begin
                     { all depending on the slope, we can determine         }
                     { in what direction the extra width pixels will be put }
                       If Flag then
                          Begin
                            { compare if we should plot a pixel here , compare }
                            { with predefined line patterns...                 }
                            if LinePatterns[i and 15] = TRUE then
                              begin
                                DirectPutPixelClip(x-1,y);
                                DirectPutPixelClip(x,y);
                                DirectPutPixelClip(x+1,y);
                              end;
                          end
                       else
                          Begin
                            { compare if we should plot a pixel here , compare }
                            { with predefined line patterns...                 }
                            if LinePatterns[i and 15] = TRUE then
                             begin
                               DirectPutPixelClip(x,y-1);
                               DirectPutPixelClip(x,y);
                               DirectPutPixelClip(x,y+1);
                             end;
                          end;
                   if d < 0 then
                         begin
                           d := d + dinc1;
                           x := x + xinc1;
                           y := y + yinc1;
                         end
                   else
                         begin
                   d := d + dinc2;
                   x := x + xinc2;
                   y := y + yinc2;
                         end;
                end;
            end
           else
            Begin
             { instead of putting in loop , substract by one now }
             TmpNumPixels := NumPixels-1;
            { NormWidth }
             for i := 0 to TmpNumPixels do
             begin
                  if LinePatterns[i and 15] = TRUE then
                    begin
                      DirectPutPixel(x,y);
                    end;
             if d < 0 then
                 begin
                   d := d + dinc1;
                   x := x + xinc1;
                   y := y + yinc1;
                 end
             else
                 begin
                   d := d + dinc2;
                   x := x + xinc2;
                   y := y + yinc2;
                 end;
             end;
            end
        end;
{******************************************}
{  end patterned lines                     }
{******************************************}
       { restore color }
       CurrentColor:=OldCurrentColor;
   end;
 end;  { Line }


  {********************************************************}
  { Procedure DummyPatternLine()                           }
  {--------------------------------------------------------}
  { This is suimply an procedure that does nothing which   }
  { can be passed as a patternlineproc for non-filled      }
  { ellipses                                               }
  {********************************************************}
  Procedure DummyPatternLine(x1, x2, y: smallint); {$ifdef tp} far; {$endif tp}
  begin
  end;


  {********************************************************}
  { Procedure InternalEllipse()                            }
  {--------------------------------------------------------}
  { This routine first calculates all points required to   }
  { draw a circle to the screen, and stores the points     }
  { to display in a buffer before plotting them. The       }
  { aspect ratio of the screen is taken into account when  }
  { calculating the values.                                }
  {--------------------------------------------------------}
  { INPUTS: X,Y : Center coordinates of Ellipse.           }
  {  XRadius - X-Axis radius of ellipse.                   }
  {  YRadius - Y-Axis radius of ellipse.                   }
  {  stAngle, EndAngle: Start angle and end angles of the  }
  {  ellipse (used for partial ellipses and circles)       }
  {  pl: procedure which either draws a patternline (for   }
  {      FillEllipse) or does nothing (arc etc)            }
  {--------------------------------------------------------}
  { NOTE: -                                                }
  {       -                                                }
  {********************************************************}

  Procedure InternalEllipseDefault(X,Y: smallint;XRadius: word;
    YRadius:word; stAngle,EndAngle: word; pl: PatternLineProc); {$ifndef fpc}far;{$endif fpc}
   Const ConvFac = Pi/180.0;

   var
    j, Delta, DeltaEnd: graph_float;
    NumOfPixels: longint;
    TempTerm: graph_float;
    xtemp, ytemp, xp, yp, xm, ym, xnext, ynext,
      plxpyp, plxmyp, plxpym, plxmym: smallint;
    BackupColor, TmpAngle, OldLineWidth: word;
  Begin
   If LineInfo.ThickNess = ThickWidth Then
    { first draw the two outer ellipses using normwidth and no filling (JM) }
     Begin
       OldLineWidth := LineInfo.Thickness;
       LineInfo.Thickness := NormWidth;
       InternalEllipseDefault(x,y,XRadius,YRadius,StAngle,EndAngle,
                              {$ifdef fpc}@{$endif fpc}DummyPatternLine);
       InternalEllipseDefault(x,y,XRadius+1,YRadius+1,StAngle,EndAngle,
                              {$ifdef fpc}@{$endif fpc}DummyPatternLine);
       If (XRadius > 0) and (YRadius > 0) Then
         { draw the smallest ellipse last, since that one will use the }
         { original pl, so it could possibly draw patternlines (JM)    }
         Begin
           Dec(XRadius);
           Dec(YRadius);
         End
       Else Exit;
       { restore line thickness }
       LineInfo.Thickness := OldLineWidth;
     End;
   If xradius = 0 then inc(xradius);
   if yradius = 0 then inc(yradius);
   { check for an ellipse with negligable x and y radius }
   If (xradius <= 1) and (yradius <= 1) then
     begin
       putpixel(x,y,CurrentColor);
       ArcCall.X := X;
       ArcCall.Y := Y;
       ArcCall.XStart := X;
       ArcCall.YStart := Y;
       ArcCall.XEnd := X;
       ArcCall.YEnd := Y;
       exit;
     end;
   { check if valid angles }
   stangle := stAngle mod 361;
   EndAngle := EndAngle mod 361;
   { if impossible angles then swap them! }
   if Endangle < StAngle then
     Begin
       TmpAngle:=EndAngle;
       EndAngle:=StAngle;
       Stangle:=TmpAngle;
     end;
   { approximate the number of pixels required by using the circumference }
   { equation of an ellipse.                                              }
   { Changed this formula a it (trial and error), but the net result is that }
   { less pixels have to be calculated now                                   }
   NumOfPixels:=Round(Sqrt(3)*sqrt(sqr(XRadius)+sqr(YRadius)));
   { Calculate the angle precision required }
   Delta := 90.0 / NumOfPixels;
   { for restoring after PatternLine }
   BackupColor := CurrentColor;
   { removed from inner loop to make faster }
   { store some arccall info }
   ArcCall.X := X;
   ArcCall.Y := Y;
   TempTerm := (StAngle)*ConvFac;
   ArcCall.XStart := round(XRadius*Cos(TempTerm)) + X;
   ArcCall.YStart := round(YRadius*Sin(TempTerm+Pi)) + Y;
   TempTerm := (EndAngle)*ConvFac;
   ArcCall.XEnd := round(XRadius*Cos(TempTerm)) + X;
   ArcCall.YEnd := round(YRadius*Sin(TempTerm+Pi)) + Y;
   { Always just go over the first 90 degrees. Could be optimized a   }
   { bit if StAngle and EndAngle lie in the same quadrant, left as an }
   { exercise for the reader :) (JM)                                  }
   j := 0;
   { calculate stop position, go 1 further than 90 because otherwise }
   { 1 pixel is sometimes not drawn (JM)                             }
   DeltaEnd := 91;
   { Calculate points }
   xnext := XRadius;
   ynext := 0;
   Repeat
     xtemp := xnext;
     ytemp := ynext;
     { this is used by both sin and cos }
     TempTerm := (j+Delta)*ConvFac;
     { Calculate points }
     xnext := round(XRadius*Cos(TempTerm));
     ynext := round(YRadius*Sin(TempTerm+Pi));

     xp := x + xtemp;
     xm := x - xtemp;
     yp := y + ytemp;
     ym := y - ytemp;
     plxpyp := maxsmallint;
     plxmyp := -maxsmallint-1;
     plxpym := maxsmallint;
     plxmym := -maxsmallint-1;
     If (j >= StAngle) and (j <= EndAngle) then
       begin
         plxpyp := xp;
         PutPixel(xp,yp,CurrentColor);
       end;
     If ((180-j) >= StAngle) and ((180-j) <= EndAngle) then
       begin
         plxmyp := xm;
         PutPixel(xm,yp,CurrentColor);
       end;
     If ((j+180) >= StAngle) and ((j+180) <= EndAngle) then
       begin
         plxmym := xm;
         PutPixel(xm,ym,CurrentColor);
       end;
     If ((360-j) >= StAngle) and ((360-j) <= EndAngle) then
       begin
         plxpym := xp;
         PutPixel(xp,ym,CurrentColor);
       end;
     If (ynext <> ytemp) and
        (xp - xm >= 1) then
       begin
         CurrentColor := FillSettings.Color;
         pl(plxmyp+1,plxpyp-1,yp);
         pl(plxmym+1,plxpym-1,ym);
         CurrentColor := BackupColor;
       end;
     j:=j+Delta;
   Until j > (DeltaEnd);
  end;


  procedure PatternLineDefault(x1,x2,y: smallint); {$ifndef fpc}far;{$endif fpc}
  {********************************************************}
  { Draws a horizontal patterned line according to the     }
  { current Fill Settings.                                 }
  {********************************************************}
  { Important notes:                                       }
  {  - CurrentColor must be set correctly before entering  }
  {    this routine.                                       }
  {********************************************************}
   var
    NrIterations: smallint;
    i           : smallint;
    j           : smallint;
    TmpFillPattern : byte;
    OldWriteMode : word;
    OldCurrentColor : word;
   begin
     { convert to global coordinates ... }
     x1 := x1 + StartXViewPort;
     x2 := x2 + StartXViewPort;
     y  := y + StartYViewPort;
     { if line was fully clipped then exit...}
     if LineClipped(x1,y,x2,y,StartXViewPort,StartYViewPort,
        StartXViewPort+ViewWidth, StartYViewPort+ViewHeight) then
         exit;

     OldWriteMode := CurrentWriteMode;
     CurrentWriteMode := NormalPut;


     { Get the current pattern }
     TmpFillPattern := FillPatternTable
       [FillSettings.Pattern][(y and $7)+1];

     Case TmpFillPattern Of
       0:
         begin
           OldCurrentColor := CurrentColor;
           CurrentColor := CurrentBkColor;
  { hline converts the coordinates to global ones, but that has been done }
  { already here!!! Convert them back to local ones... (JM)                }
           HLine(x1-StartXViewPort,x2-StartXViewPort,y-StartYViewPort);
           CurrentColor := OldCurrentColor;
         end;
      $ff:
        begin
          OldCurrentColor := CurrentColor;
          CurrentColor := FillSettings.Color;
          HLine(x1-StartXViewPort,x2-StartXViewPort,y-StartYViewPort);
          CurrentColor := OldCurrentColor;
        end;
       else
         begin
           { number of times to go throuh the 8x8 pattern }
           NrIterations := abs(x2 - x1+8) div 8;
           For i:= 0 to NrIterations do
             Begin
               for j:=0 to 7 do
                    Begin
                    { x1 mod 8 }
                    if RevBitArray[x1 and 7] and TmpFillPattern <> 0 then
                      begin
                        OldCurrentColor := CurrentColor;
                        CurrentColor := FillSettings.Color;
                        DirectPutpixel(x1,y);
                        CurrentColor := OldCurrentColor;
                      end
                    else
                      begin
                        { According to the TP graph manual, we overwrite everything }
                        { which is filled up - checked against VGA and CGA drivers  }
                        { of TP.                                                    }
                        OldCurrentColor := CurrentColor;
                        CurrentColor := CurrentBkColor;
                        DirectPutPixel(x1,y);
                        CurrentColor := OldCurrentColor;
                      end;
                    Inc(x1);
                    if x1 > x2 then
                     begin
                       CurrentWriteMode := OldWriteMode;
                       exit;
                     end;
                   end;
             end;
          end;
     End;
     CurrentWriteMode := OldWriteMode;
   end;


  procedure LineRel(Dx, Dy: smallint);

   Begin
     Line(CurrentX, CurrentY, CurrentX + Dx, CurrentY + Dy);
     CurrentX := CurrentX + Dx;
     CurrentY := CurrentY + Dy;
   end;


  procedure LineTo(x,y : smallint);

   Begin
     Line(CurrentX, CurrentY, X, Y);
     CurrentX := X;
     CurrentY := Y;
   end;




  procedure Rectangle(x1,y1,x2,y2:smallint);
   var
     tmp: smallint;
   begin
     if x1 > x2 then
     begin
       tmp := x1;
       x1 := x2;
       x2 := tmp;
     end;
     if y1 > y2 then
     begin
       tmp := y1;
       y1 := y2;
       y2 := tmp;
     end;
     { Do not draw the end points }
     Line(x1,y1,x2-1,y1);
     Line(x1,y1+1,x1,y2);
     Line(x2,y1,x2,y2-1);
     Line(x1+1,y2,x2,y2);
   end;


  procedure GetLineSettings(var ActiveLineInfo : LineSettingsType);

   begin
    Activelineinfo:=Lineinfo;
   end;


  procedure SetLineStyle(LineStyle: word; Pattern: word; Thickness: word);

   var
    i: byte;
    j: byte;

   Begin
    if (LineStyle > UserBitLn) or ((Thickness <> Normwidth) and (Thickness <> ThickWidth)) then
      _GraphResult := grError
    else
      begin
       LineInfo.Thickness := Thickness;
       LineInfo.LineStyle := LineStyle;
       case LineStyle of
            UserBitLn: Lineinfo.Pattern := pattern;
            SolidLn:   Lineinfo.Pattern  := $ffff;  { ------- }
            DashedLn : Lineinfo.Pattern := $F8F8;   { -- -- --}
            DottedLn:  LineInfo.Pattern := $CCCC;   { - - - - }
            CenterLn: LineInfo.Pattern :=  $FC78;   { -- - -- }
       end; { end case }
       { setup pattern styles }
       j:=16;
       for i:=0 to 15 do
        Begin
         dec(j);
         { bitwise mask for each bit in the word }
         if (word($01 shl i) AND LineInfo.Pattern) <> 0 then
               LinePatterns[j]:=TRUE
             else
               LinePatterns[j]:=FALSE;
        end;
      end;
   end;




{--------------------------------------------------------------------------}
{                                                                          }
{                    VIEWPORT RELATED ROUTINES                             }
{                                                                          }
{--------------------------------------------------------------------------}


Procedure ClearViewPortDefault; {$ifndef fpc}far;{$endif fpc}
var
 j: smallint;
 OldWriteMode, OldCurColor: word;
 LineSets : LineSettingsType;
Begin
  { CP is always RELATIVE coordinates }
  CurrentX := 0;
  CurrentY := 0;

  { Save all old settings }
  OldCurColor := CurrentColor;
  CurrentColor:=CurrentBkColor;
  OldWriteMode:=CurrentWriteMode;
  CurrentWriteMode:=NormalPut;
  GetLineSettings(LineSets);
  { reset to normal line style...}
  SetLineStyle(SolidLn, 0, NormWidth);
  { routines are relative here...}
  { ViewHeight is Height-1 !     }
  for J:=0 to ViewHeight do
       HLine(0, ViewWidth , J);

  { restore old settings...}
  SetLineStyle(LineSets.LineStyle, LineSets.Pattern, LineSets.Thickness);
  CurrentColor := OldCurColor;
  CurrentWriteMode := OldWriteMode;
end;


Procedure SetViewPort(X1, Y1, X2, Y2: smallint; Clip: Boolean);
Begin
  if (X1 > GetMaxX) or (X2 > GetMaxX) or (X1 > X2) or (X1 < 0) then
  Begin
{$ifdef logging}
    logln('invalid setviewport parameters: ('
      +strf(x1)+','+strf(y1)+'), ('+strf(x2)+','+strf(y2)+')');
    logln('maxx = '+strf(getmaxx)+', maxy = '+strf(getmaxy));
{$endif logging}
    _GraphResult := grError;
    exit;
  end;
  if (Y1 > GetMaxY) or (Y2 > GetMaxY) or (Y1 > Y2) or (Y1 < 0) then
  Begin
{$ifdef logging}
    logln('invalid setviewport parameters: ('
      +strf(x1)+','+strf(y1)+'), ('+strf(x2)+','+strf(y2)+')');
    logln('maxx = '+strf(getmaxx)+', maxy = '+strf(getmaxy));
{$endif logging}
    _GraphResult := grError;
    exit;
  end;
  { CP is always RELATIVE coordinates }
  CurrentX := 0;
  CurrentY := 0;
  StartXViewPort := X1;
  StartYViewPort := Y1;
  ViewWidth :=  X2-X1;
  ViewHeight:=  Y2-Y1;
  ClipPixels := Clip;
end;


procedure GetViewSettings(var viewport : ViewPortType);
begin
  ViewPort.X1 := StartXViewPort;
  ViewPort.Y1 := StartYViewPort;
  ViewPort.X2 := ViewWidth + StartXViewPort;
  ViewPort.Y2 := ViewHeight + StartYViewPort;
  ViewPort.Clip := ClipPixels;
end;

procedure ClearDevice;
var
  ViewPort: ViewPortType;
begin
  { Reset the CP }
  CurrentX := 0;
  CurrentY := 0;
  { save viewport }
  ViewPort.X1 :=  StartXviewPort;
  ViewPort.X2 :=  ViewWidth - StartXViewPort;
  ViewPort.Y1 :=  StartYViewPort;
  ViewPort.Y2 :=  ViewHeight - StartYViewPort;
  ViewPort.Clip := ClipPixels;
  { put viewport to full screen }
  StartXViewPort := 0;
  ViewHeight := MaxY;
  StartYViewPort := 0;
  ViewWidth := MaxX;
  ClipPixels := TRUE;
  ClearViewPort;
  { restore old viewport }
  StartXViewPort := ViewPort.X1;
  ViewWidth := ViewPort.X2-ViewPort.X1;
  StartYViewPort := ViewPort.Y1;
  ViewHeight := ViewPort.Y2-ViewPort.Y1;
  ClipPixels := ViewPort.Clip;
end;



{--------------------------------------------------------------------------}
{                                                                          }
{                      BITMAP PUT/GET ROUTINES                             }
{                                                                          }
{--------------------------------------------------------------------------}


  Procedure GetScanlineDefault (X1, X2, Y : smallint; Var Data); {$ifndef fpc}far;{$endif fpc}
  {**********************************************************}
  { Procedure GetScanLine()                                  }
  {----------------------------------------------------------}
  { Returns the full scanline of the video line of the Y     }
  { coordinate. The values are returned in a WORD array      }
  { each WORD representing a pixel of the specified scanline }
  { note: we only need the pixels inside the ViewPort! (JM)  }
  { note2: extended so you can specify start and end X coord }
  {   so it is usable for GetImage too (JM)                  }
  {**********************************************************}


  Var
    x : smallint;
  Begin
     For x:=X1 to X2 Do
       WordArray(Data)[x-x1]:=GetPixel(x, y);
  End;



Function DefaultImageSize(X1,Y1,X2,Y2: smallint): longint; {$ifndef fpc}far;{$endif fpc}
Begin
  { each pixel uses two bytes, to enable modes with colors up to 64K }
  { to work.                                                         }
  DefaultImageSize := 12 + (((X2-X1+1)*(Y2-Y1+1))*2);
end;

Procedure DefaultPutImage(X,Y: smallint; var Bitmap; BitBlt: Word); {$ifndef fpc}far;{$endif fpc}
type
  pt = array[0..$fffffff] of word;
  ptw = array[0..2] of longint;
var
  k: longint;
  oldCurrentColor: word;
  oldCurrentWriteMode, i, j, y1, x1, deltaX, deltaX1, deltaY: smallint;
Begin
{$ifdef logging}
  LogLn('putImage at ('+strf(x)+','+strf(y)+') with width '+strf(ptw(Bitmap)[0])+
    ' and height '+strf(ptw(Bitmap)[1]));
  deltaY := 0;
{$endif logging}
  inc(x,startXViewPort);
  inc(y,startYViewPort);
  { width/height are 1-based, coordinates are zero based }
  x1 := ptw(Bitmap)[0]+x-1; { get width and adjust end coordinate accordingly }
  y1 := ptw(Bitmap)[1]+y-1; { get height and adjust end coordinate accordingly }

  deltaX := 0;
  deltaX1 := 0;
  k := 3 * sizeOf(Longint) div sizeOf(Word); { Three reserved longs at start of bitmap }
 { check which part of the image is in the viewport }
  if clipPixels then
    begin
      if y < startYViewPort then
        begin
          deltaY := startYViewPort - y;
          inc(k,(x1-x+1)*deltaY);
          y := startYViewPort;
         end;
      if y1 > startYViewPort+viewHeight then
        y1 := startYViewPort+viewHeight;
      if x < startXViewPort then
        begin
          deltaX := startXViewPort-x;
          x := startXViewPort;
        end;
      if x1 > startXViewPort + viewWidth then
        begin
          deltaX1 := x1 - (startXViewPort + viewWidth);
          x1 := startXViewPort + viewWidth;
        end;
    end;
{$ifdef logging}
  LogLn('deltax: '+strf(deltax)+', deltax1: '+strf(deltax1)+',deltay: '+strf(deltay));
{$endif logging}
  oldCurrentColor := currentColor;
  oldCurrentWriteMode := currentWriteMode;
  currentWriteMode := bitBlt;
  for j:=Y to Y1 do
   Begin
     inc(k,deltaX);
     for i:=X to X1 do
      begin
        currentColor := pt(bitmap)[k];
        directPutPixel(i,j);
        inc(k);
     end;
     inc(k,deltaX1);
   end;
  currentWriteMode := oldCurrentWriteMode;
  currentColor := oldCurrentColor;
end;

Procedure DefaultGetImage(X1,Y1,X2,Y2: smallint; Var Bitmap); {$ifndef fpc}far;{$endif fpc}
type
  pt = array[0..$fffffff] of word;
  ptw = array[0..2] of longint;
var
  i,j: smallint;
  k: longint;
Begin
  k:= 3 * Sizeof(longint) div sizeof(word); { Three reserved longs at start of bitmap }
  i := x2 - x1 + 1;
  for j:=Y1 to Y2 do
   Begin
     GetScanLine(x1,x2,j,pt(Bitmap)[k]);
     inc(k,i);
   end;
   ptw(Bitmap)[0] := X2-X1+1;   { First longint  is width  }
   ptw(Bitmap)[1] := Y2-Y1+1;   { Second longint is height }
   ptw(bitmap)[2] := 0;       { Third longint is reserved}
end;






  Procedure GetArcCoords(var ArcCoords: ArcCoordsType);
   Begin
     ArcCoords.X := ArcCall.X;
     ArcCoords.Y := ArcCall.Y;
     ArcCoords.XStart := ArcCall.XStart;
     ArcCoords.YStart := ArcCall.YStart;
     ArcCoords.XEnd := ArcCall.XEnd;
     ArcCoords.YEnd := ArcCall.YEnd;
   end;


  procedure SetVisualPageDefault(page : word); {$ifndef fpc}far;{$endif fpc}
   begin
   end;


  procedure SetActivePageDefault(page : word); {$ifndef fpc}far;{$endif fpc}
   begin
   end;

  procedure DirectPutPixelDefault(X,Y: smallint);
   begin
     Writeln(stderr,'Error: Not in graphics mode (use InitGraph and test GraphResult afterwards)');
     Halt(1);
   end;

  function GetPixelDefault(X,Y: smallint): word;
   begin
     Writeln(stderr,'Error: Not in graphics mode (use InitGraph and test GraphResult afterwards)');
     Halt(1);
     exit(0); { avoid warning }
   end;

  procedure PutPixelDefault(X,Y: smallint; Color: Word);
   begin
     Writeln(stderr,'Error: Not in graphics mode (use InitGraph and test GraphResult afterwards)');
     Halt(1);
   end;

  procedure SetRGBPaletteDefault(ColorNum, RedValue, GreenValue, BlueValue: smallint);
   begin
     Writeln(stderr,'Error: Not in graphics mode (use InitGraph and test GraphResult afterwards)');
     Halt(1);
   end;

  procedure GetRGBPaletteDefault(ColorNum: smallint; var
            RedValue, GreenValue, BlueValue: smallint);
   begin
     Writeln(stderr,'Error: Not in graphics mode (use InitGraph and test GraphResult afterwards)');
     Halt(1);
   end;


  procedure OutTextXYDefault(x,y : smallint;const TextString : string);forward;
  procedure CircleDefault(X, Y: smallint; Radius:Word);forward;
  procedure SetBkColorDefault(ColorNum: Word);forward;
  function GetBkColorDefault: Word;forward;

{

    This file is part of the Free Pascal run time library.
    Copyright (c) 1999-2000 by the Free Pascal development team

    This include implements the different palette manipulation
    routines.

    See the file COPYING.FPC, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}

CONST
 { This is the standard VGA palette for the first 256 colors }
 { at 6 bpp for each RG and B component.                     }
 DefaultColors: Array[0..255] of RGBRec = (
  (Red:   0;Green:   0;Blue:   0),
  (Red:   0;Green:   0;Blue: 168),
  (Red:   0;Green: 168;Blue:   0),
  (Red:   0;Green: 168;Blue: 168),
  (Red: 168;Green:   0;Blue:   0),
  (Red: 168;Green:   0;Blue: 168),
  (Red: 168;Green:  84;Blue:   0),
  (Red: 168;Green: 168;Blue: 168),
  (Red:  84;Green:  84;Blue:  84),
  (Red:  84;Green:  84;Blue: 252),
  (Red:  84;Green: 252;Blue:  84),
  (Red:  84;Green: 252;Blue: 252),
  (Red: 252;Green:  84;Blue:  84),
  (Red: 252;Green:  84;Blue: 252),
  (Red: 252;Green: 252;Blue:  84),
  (Red: 252;Green: 252;Blue: 252),
  (Red:   0;Green:   0;Blue:   0),
  (Red:  20;Green:  20;Blue:  20),
  (Red:  32;Green:  32;Blue:  32),
  (Red:  44;Green:  44;Blue:  44),
  (Red:  56;Green:  56;Blue:  56),
  (Red:  68;Green:  68;Blue:  68),
  (Red:  80;Green:  80;Blue:  80),
  (Red:  96;Green:  96;Blue:  96),
  (Red: 112;Green: 112;Blue: 112),
  (Red: 128;Green: 128;Blue: 128),
  (Red: 144;Green: 144;Blue: 144),
  (Red: 160;Green: 160;Blue: 160),
  (Red: 180;Green: 180;Blue: 180),
  (Red: 200;Green: 200;Blue: 200),
  (Red: 224;Green: 224;Blue: 224),
  (Red: 252;Green: 252;Blue: 252),
  (Red:   0;Green:   0;Blue: 252),
  (Red:  64;Green:   0;Blue: 252),
  (Red: 124;Green:   0;Blue: 252),
  (Red: 188;Green:   0;Blue: 252),
  (Red: 252;Green:   0;Blue: 252),
  (Red: 252;Green:   0;Blue: 188),
  (Red: 252;Green:   0;Blue: 124),
  (Red: 252;Green:   0;Blue:  64),
  (Red: 252;Green:   0;Blue:   0),
  (Red: 252;Green:  64;Blue:   0),
  (Red: 252;Green: 124;Blue:   0),
  (Red: 252;Green: 188;Blue:   0),
  (Red: 252;Green: 252;Blue:   0),
  (Red: 188;Green: 252;Blue:   0),
  (Red: 124;Green: 252;Blue:   0),
  (Red:  64;Green: 252;Blue:   0),
  (Red:   0;Green: 252;Blue:   0),
  (Red:   0;Green: 252;Blue:  64),
  (Red:   0;Green: 252;Blue: 124),
  (Red:   0;Green: 252;Blue: 188),
  (Red:   0;Green: 252;Blue: 252),
  (Red:   0;Green: 188;Blue: 252),
  (Red:   0;Green: 124;Blue: 252),
  (Red:   0;Green:  64;Blue: 252),
  (Red: 124;Green: 124;Blue: 252),
  (Red: 156;Green: 124;Blue: 252),
  (Red: 188;Green: 124;Blue: 252),
  (Red: 220;Green: 124;Blue: 252),
  (Red: 252;Green: 124;Blue: 252),
  (Red: 252;Green: 124;Blue: 220),
  (Red: 252;Green: 124;Blue: 188),
  (Red: 252;Green: 124;Blue: 156),
  (Red: 252;Green: 124;Blue: 124),
  (Red: 252;Green: 156;Blue: 124),
  (Red: 252;Green: 188;Blue: 124),
  (Red: 252;Green: 220;Blue: 124),
  (Red: 252;Green: 252;Blue: 124),
  (Red: 220;Green: 252;Blue: 124),
  (Red: 188;Green: 252;Blue: 124),
  (Red: 156;Green: 252;Blue: 124),
  (Red: 124;Green: 252;Blue: 124),
  (Red: 124;Green: 252;Blue: 156),
  (Red: 124;Green: 252;Blue: 188),
  (Red: 124;Green: 252;Blue: 220),
  (Red: 124;Green: 252;Blue: 252),
  (Red: 124;Green: 220;Blue: 252),
  (Red: 124;Green: 188;Blue: 252),
  (Red: 124;Green: 156;Blue: 252),
  (Red: 180;Green: 180;Blue: 252),
  (Red: 196;Green: 180;Blue: 252),
  (Red: 216;Green: 180;Blue: 252),
  (Red: 232;Green: 180;Blue: 252),
  (Red: 252;Green: 180;Blue: 252),
  (Red: 252;Green: 180;Blue: 232),
  (Red: 252;Green: 180;Blue: 216),
  (Red: 252;Green: 180;Blue: 196),
  (Red: 252;Green: 180;Blue: 180),
  (Red: 252;Green: 196;Blue: 180),
  (Red: 252;Green: 216;Blue: 180),
  (Red: 252;Green: 232;Blue: 180),
  (Red: 252;Green: 252;Blue: 180),
  (Red: 232;Green: 252;Blue: 180),
  (Red: 216;Green: 252;Blue: 180),
  (Red: 196;Green: 252;Blue: 180),
  (Red: 180;Green: 252;Blue: 180),
  (Red: 180;Green: 252;Blue: 196),
  (Red: 180;Green: 252;Blue: 216),
  (Red: 180;Green: 252;Blue: 232),
  (Red: 180;Green: 252;Blue: 252),
  (Red: 180;Green: 232;Blue: 252),
  (Red: 180;Green: 216;Blue: 252),
  (Red: 180;Green: 196;Blue: 252),
  (Red:   0;Green:   0;Blue: 112),
  (Red:  28;Green:   0;Blue: 112),
  (Red:  56;Green:   0;Blue: 112),
  (Red:  84;Green:   0;Blue: 112),
  (Red: 112;Green:   0;Blue: 112),
  (Red: 112;Green:   0;Blue:  84),
  (Red: 112;Green:   0;Blue:  56),
  (Red: 112;Green:   0;Blue:  28),
  (Red: 112;Green:   0;Blue:   0),
  (Red: 112;Green:  28;Blue:   0),
  (Red: 112;Green:  56;Blue:   0),
  (Red: 112;Green:  84;Blue:   0),
  (Red: 112;Green: 112;Blue:   0),
  (Red:  84;Green: 112;Blue:   0),
  (Red:  56;Green: 112;Blue:   0),
  (Red:  28;Green: 112;Blue:   0),
  (Red:   0;Green: 112;Blue:   0),
  (Red:   0;Green: 112;Blue:  28),
  (Red:   0;Green: 112;Blue:  56),
  (Red:   0;Green: 112;Blue:  84),
  (Red:   0;Green: 112;Blue: 112),
  (Red:   0;Green:  84;Blue: 112),
  (Red:   0;Green:  56;Blue: 112),
  (Red:   0;Green:  28;Blue: 112),
  (Red:  56;Green:  56;Blue: 112),
  (Red:  68;Green:  56;Blue: 112),
  (Red:  84;Green:  56;Blue: 112),
  (Red:  96;Green:  56;Blue: 112),
  (Red: 112;Green:  56;Blue: 112),
  (Red: 112;Green:  56;Blue:  96),
  (Red: 112;Green:  56;Blue:  84),
  (Red: 112;Green:  56;Blue:  68),
  (Red: 112;Green:  56;Blue:  56),
  (Red: 112;Green:  68;Blue:  56),
  (Red: 112;Green:  84;Blue:  56),
  (Red: 112;Green:  96;Blue:  56),
  (Red: 112;Green: 112;Blue:  56),
  (Red:  96;Green: 112;Blue:  56),
  (Red:  84;Green: 112;Blue:  56),
  (Red:  68;Green: 112;Blue:  56),
  (Red:  56;Green: 112;Blue:  56),
  (Red:  56;Green: 112;Blue:  68),
  (Red:  56;Green: 112;Blue:  84),
  (Red:  56;Green: 112;Blue:  96),
  (Red:  56;Green: 112;Blue: 112),
  (Red:  56;Green:  96;Blue: 112),
  (Red:  56;Green:  84;Blue: 112),
  (Red:  56;Green:  68;Blue: 112),
  (Red:  80;Green:  80;Blue: 112),
  (Red:  88;Green:  80;Blue: 112),
  (Red:  96;Green:  80;Blue: 112),
  (Red: 104;Green:  80;Blue: 112),
  (Red: 112;Green:  80;Blue: 112),
  (Red: 112;Green:  80;Blue: 104),
  (Red: 112;Green:  80;Blue:  96),
  (Red: 112;Green:  80;Blue:  88),
  (Red: 112;Green:  80;Blue:  80),
  (Red: 112;Green:  88;Blue:  80),
  (Red: 112;Green:  96;Blue:  80),
  (Red: 112;Green: 104;Blue:  80),
  (Red: 112;Green: 112;Blue:  80),
  (Red: 104;Green: 112;Blue:  80),
  (Red:  96;Green: 112;Blue:  80),
  (Red:  88;Green: 112;Blue:  80),
  (Red:  80;Green: 112;Blue:  80),
  (Red:  80;Green: 112;Blue:  88),
  (Red:  80;Green: 112;Blue:  96),
  (Red:  80;Green: 112;Blue: 104),
  (Red:  80;Green: 112;Blue: 112),
  (Red:  80;Green: 104;Blue: 112),
  (Red:  80;Green:  96;Blue: 112),
  (Red:  80;Green:  88;Blue: 112),
  (Red:   0;Green:   0;Blue:  64),
  (Red:  16;Green:   0;Blue:  64),
  (Red:  32;Green:   0;Blue:  64),
  (Red:  48;Green:   0;Blue:  64),
  (Red:  64;Green:   0;Blue:  64),
  (Red:  64;Green:   0;Blue:  48),
  (Red:  64;Green:   0;Blue:  32),
  (Red:  64;Green:   0;Blue:  16),
  (Red:  64;Green:   0;Blue:   0),
  (Red:  64;Green:  16;Blue:   0),
  (Red:  64;Green:  32;Blue:   0),
  (Red:  64;Green:  48;Blue:   0),
  (Red:  64;Green:  64;Blue:   0),
  (Red:  48;Green:  64;Blue:   0),
  (Red:  32;Green:  64;Blue:   0),
  (Red:  16;Green:  64;Blue:   0),
  (Red:   0;Green:  64;Blue:   0),
  (Red:   0;Green:  64;Blue:  16),
  (Red:   0;Green:  64;Blue:  32),
  (Red:   0;Green:  64;Blue:  48),
  (Red:   0;Green:  64;Blue:  64),
  (Red:   0;Green:  48;Blue:  64),
  (Red:   0;Green:  32;Blue:  64),
  (Red:   0;Green:  16;Blue:  64),
  (Red:  32;Green:  32;Blue:  64),
  (Red:  40;Green:  32;Blue:  64),
  (Red:  48;Green:  32;Blue:  64),
  (Red:  56;Green:  32;Blue:  64),
  (Red:  64;Green:  32;Blue:  64),
  (Red:  64;Green:  32;Blue:  56),
  (Red:  64;Green:  32;Blue:  48),
  (Red:  64;Green:  32;Blue:  40),
  (Red:  64;Green:  32;Blue:  32),
  (Red:  64;Green:  40;Blue:  32),
  (Red:  64;Green:  48;Blue:  32),
  (Red:  64;Green:  56;Blue:  32),
  (Red:  64;Green:  64;Blue:  32),
  (Red:  56;Green:  64;Blue:  32),
  (Red:  48;Green:  64;Blue:  32),
  (Red:  40;Green:  64;Blue:  32),
  (Red:  32;Green:  64;Blue:  32),
  (Red:  32;Green:  64;Blue:  40),
  (Red:  32;Green:  64;Blue:  48),
  (Red:  32;Green:  64;Blue:  56),
  (Red:  32;Green:  64;Blue:  64),
  (Red:  32;Green:  56;Blue:  64),
  (Red:  32;Green:  48;Blue:  64),
  (Red:  32;Green:  40;Blue:  64),
  (Red:  44;Green:  44;Blue:  64),
  (Red:  48;Green:  44;Blue:  64),
  (Red:  52;Green:  44;Blue:  64),
  (Red:  60;Green:  44;Blue:  64),
  (Red:  64;Green:  44;Blue:  64),
  (Red:  64;Green:  44;Blue:  60),
  (Red:  64;Green:  44;Blue:  52),
  (Red:  64;Green:  44;Blue:  48),
  (Red:  64;Green:  44;Blue:  44),
  (Red:  64;Green:  48;Blue:  44),
  (Red:  64;Green:  52;Blue:  44),
  (Red:  64;Green:  60;Blue:  44),
  (Red:  64;Green:  64;Blue:  44),
  (Red:  60;Green:  64;Blue:  44),
  (Red:  52;Green:  64;Blue:  44),
  (Red:  48;Green:  64;Blue:  44),
  (Red:  44;Green:  64;Blue:  44),
  (Red:  44;Green:  64;Blue:  48),
  (Red:  44;Green:  64;Blue:  52),
  (Red:  44;Green:  64;Blue:  60),
  (Red:  44;Green:  64;Blue:  64),
  (Red:  44;Green:  60;Blue:  64),
  (Red:  44;Green:  52;Blue:  64),
  (Red:  44;Green:  48;Blue:  64),
  (Red:   0;Green:   0;Blue:   0),
  (Red:   0;Green:   0;Blue:   0),
  (Red:   0;Green:   0;Blue:   0),
  (Red:   0;Green:   0;Blue:   0),
  (Red:   0;Green:   0;Blue:   0),
  (Red:   0;Green:   0;Blue:   0),
  (Red:   0;Green:   0;Blue:   0),
  (Red:   0;Green:   0;Blue:   0));

  procedure SetAllPaletteDefault(const Palette:PaletteType);
   var
    i: longint;
    Size: longint;
   begin
     { palette routines do not work in DirectColor mode }
     if DirectColor then
       begin
         _GraphResult := grError;
         exit;
       end;
     Size:=Palette.Size;  { number of entries...}
     { first determine if we are not trying to }
     { change too much colors...               }
     if Palette.Size > PaletteSize then
      begin
        _GraphResult := grError;
        exit;
      end;
     Dec(Size); { Color arrays are indexed according to zero }
     for i:=0 to Size do
      begin
        { skip if RGB values are -1 , as stated in the TP manual }
        if (Palette.Colors[i].Red <> -1) and (Palette.Colors[i].Green <> -1)
           and (Palette.Colors[i].Blue <> -1) then
              SetRGBPalette(i,
                Palette.Colors[i].Red,
                Palette.Colors[i].Green,
                Palette.Colors[i].Blue);
      end;
   end;

  {********************************************************}
  { Procedure SetPalette()                                 }
  {--------------------------------------------------------}
  { This routine changes the colorNum to the default       }
  { palette entry specified in the second parameter.       }
  { For example, SetPalette(0, Lightcyan) makes the        }
  { 0th palette entry to the default Light Cyan Color .    }
  {********************************************************}
   Procedure SetPalette(ColorNum: word; Color: shortint);
    begin
     { palette routines do not work in DirectColor mode }
     if DirectColor then
       begin
         _GraphResult := grError;
         exit;
       end;
      { Check if we can actually change that palette color }
      if ColorNum > PaletteSize then
        Begin
          _GraphResult := grError;
          exit;
        end
      else
      { Check if the max. default color is reached...}
      { no, this disables palette setting for 256 color modes! (JM) }
{      if Color > EGAWhite then
        begin
          _GraphResult := grError;
          exit;
        end;}
      SetRGBPalette(ColorNum,
          DefaultColors[Color].Red,
          DefaultColors[Color].Green,
          DefaultColors[Color].Blue);
    end;


    procedure GetPalette(var Palette: PaletteType);
      var
        i: longint;
        size : longint;
      begin
        { palette routines do not work in DirectColor mode }
        if DirectColor then
         begin
           _GraphResult := grError;
           exit;
         end;
        Palette.Size := PaletteSize;
        { index at zero }
        size := PaletteSize - 1;
        for i:=0 to size do
          GetRGBPalette(i,
             Palette.Colors[i].Red,
             Palette.Colors[i].Green,
             Palette.Colors[i].Blue);
      end;

    function GetPaletteSize: smallint;
     begin
       GetPaletteSize := PaletteSize;
     end;

    procedure GetDefaultPalette(var Palette: PaletteType);
      begin
        System.move(DefaultColors, Palette.Colors, sizeof(DefaultColors));
        { The default palette always has 256 entries, but in reality }
        { it depends on the number of colors possible.               }
        Palette.Size := PaletteSize;
        if PaletteSize > 256 then Palette.Size := 256;
      end;



  Procedure DefaultHooks;
  {********************************************************}
  { Procedure DefaultHooks()                               }
  {--------------------------------------------------------}
  { Resets all hookable routine either to nil for those    }
  { which need overrides, and others to defaults.          }
  { This is called each time SetGraphMode() is called.     }
  {********************************************************}
  Begin
    { All default hooks procedures }

    { required...}
    DirectPutPixel := {$ifdef fpc}@{$endif}DirectPutPixelDefault;
    PutPixel := {$ifdef fpc}@{$endif}PutPixelDefault;
    GetPixel := {$ifdef fpc}@{$endif}GetPixelDefault;
    SetRGBPalette := {$ifdef fpc}@{$endif}SetRGBPaletteDefault;
    GetRGBPalette := {$ifdef fpc}@{$endif}GetRGBPaletteDefault;
    { optional...}
    SetAllPalette := {$ifdef fpc}@{$endif}SetAllPaletteDefault;
    SetActivePage := {$ifdef fpc}@{$endif}SetActivePageDefault;
    SetVisualPage := {$ifdef fpc}@{$endif}SetVisualPageDefault;
    ClearViewPort := {$ifdef fpc}@{$endif}ClearViewportDefault;
    PutImage := {$ifdef fpc}@{$endif}DefaultPutImage;
    GetImage := {$ifdef fpc}@{$endif}DefaultGetImage;
    ImageSize := {$ifdef fpc}@{$endif}DefaultImageSize;
    GraphFreeMemPtr := nil;
    GraphGetMemPtr := nil;
    GetScanLine := {$ifdef fpc}@{$endif}GetScanLineDefault;
    Line := {$ifdef fpc}@{$endif}LineDefault;
    InternalEllipse := {$ifdef fpc}@{$endif}InternalEllipseDefault;
    PatternLine := {$ifdef fpc}@{$endif}PatternLineDefault;
    HLine := {$ifdef fpc}@{$endif}HLineDefault;
    VLine := {$ifdef fpc}@{$endif}VLineDefault;
    OuttextXY := {$ifdef fpc}@{$endif}OuttextXYDefault;
    Circle := {$ifdef fpc}@{$endif}CircleDefault;
    SetBkColor := {$ifdef fpc}@{$endif}SetBkColorDefault;
    GetBkColor := {$ifdef fpc}@{$endif}GetBkColorDefault;
  end;

  Procedure InitVars;
  {********************************************************}
  { Procedure InitVars()                                   }
  {--------------------------------------------------------}
  { Resets all internal variables, and resets all          }
  { overridable routines.                                  }
  {********************************************************}
   Begin
    DirectVideo := TRUE;  { By default use fastest access possible }
    ArcCall.X := 0;
    ArcCall.Y := 0;
    ArcCall.XStart := 0;
    ArcCall.YStart := 0;
    ArcCall.XEnd := 0;
    ArcCall.YEnd := 0;
    { Reset to default values }
    IntCurrentMode := 0;
    IntCurrentDriver := 0;
    IntCurrentNewDriver := 0;
    XAspect := 0;
    YAspect := 0;
    MaxX := 0;
    MaxY := 0;
    MaxColor := 0;
    PaletteSize := 0;
    DirectColor := FALSE;
    HardwarePages := 0;
    if hardwarepages=0 then; { remove note }
    DefaultHooks;
  end;

  {

    This file is part of the Free Pascal run time library.
    Copyright (c) 1999-2000 by the Free Pascal development team

    This include implements video mode management.

    See the file COPYING.FPC, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}

{-----------------------------------------------------------------------}
{                          Internal routines                            }
{-----------------------------------------------------------------------}

 procedure res2Mode(x, y, maxColor: longint; var driver,mode: smallInt);
 var
   l: longint;
 begin
   case maxColor of
     2: driver := D1bit;
     4: driver := D2bit;
     16: driver := D4bit;
     64: driver := D6bit;
     256: driver := D8bit;
     4096: driver := D12bit;
     32768: driver := D15bit;
     65536: driver := D16bit;
{    not yet supported
     65536*256: driver := D24bit;
     65536*65536: driver := D32bit;}
     else
       begin
         driver := maxsmallint;
         exit;
       end;
   end;
   { Check whether this is known/predefined mode }
   for l := lowNewMode to highNewMode do
     if (resolutions[l].x = x) and
        (resolutions[l].y = y) then
       begin
         { Found! }
         mode := l;
         exit;
       end;
   { Not Found }
   mode := maxsmallint;
 end;

function mode2res(modeNr: smallInt; var x,y: longint): boolean;
begin
  if (modeNr < lowNewMode) or
     (modeNr > highNewMode) then
    begin
      mode2res := false;
      exit;
    end;
  mode2res := true;
  x := resolutions[modeNr].x;
  y := resolutions[modeNr].y;
end;


 procedure addmode(const mode: TModeInfo);
  {********************************************************}
  { Procedure AddMode()                                    }
  {--------------------------------------------------------}
  { This routine adds <mode> to the list of recognized     }
  { modes. Duplicates are allowed.                         }
  {********************************************************}
  var
    i,driverNr, modeNr: smallint;
    prev: PModeInfo;
    list: PModeInfo;
    newlst : PModeInfo;
  begin
    res2Mode(mode.maxx+1,mode.maxy+1,mode.maxColor,driverNr,ModeNr);
    { bitdepth supported? }
    if (driverNr <> maxsmallint) then
     begin
      { Yes, add the mode }
      if not assigned(newModeList.modeinfo[driverNr]) then
        begin
{$ifdef logging}
          logln('Adding resolution '+strf(modenr)+' for drivernr '+strf(drivernr)+
                ' ('+strf(mode.maxx)+'x'+strf(mode.maxy)+')');
{$endif logging}
          new(newModeList.modeinfo[driverNr]);
          newModeList.modeinfo[driverNr]^ := mode;
          newModeList.modeinfo[driverNr]^.next:=nil;
        end
      else
        begin
          prev := nil;
          list := newModeList.modeinfo[driverNr];
          { sort first by x resolution, then by yresolution }
          while assigned(list) and
                ((list^.maxx < mode.maxx) or
                 ((list^.maxx = mode.maxx) and
                  (list^.maxy < mode.maxy))) do
            begin
              prev := list;
              list := list^.next;
            end;
          { mode already exists? -> replace (assume later added modes are }
          { better)                                                       }
          if assigned(list) and
             (list^.maxx = mode.maxx) and
             (list^.maxy = mode.maxy) then
            begin
{$ifdef logging}
              logln('replacing resolution '+strf(modenr)+' for drivernr '+strf(drivernr)+
                ' ('+strf(mode.maxx)+'x'+strf(mode.maxy)+')');
{$endif logging}
              { save/restore next, drivernr and drivermode in list }
              prev := list^.next;
              list^ := mode;
              list^.next := prev;
            end
          else
            begin
              new(newLst);
              { Increase the number of modes for this driver }
              newLst^ := mode;
{$ifdef logging}
              logln('Adding resolution '+strf(modenr)+' for drivernr '+strf(drivernr)+
                ' ('+strf(mode.maxx)+'x'+strf(mode.maxy)+')');
{$endif logging}
              newLst^.next := list;
              if assigned(prev) then
                prev^.next := newLst
              else
                newModeList.modeinfo[driverNr] := newLst;
            end;
        end;
       { renumber internmodenumber }
       list := newModeList.modeinfo[driverNr];
       i:=0;
       while assigned(list) do
        begin
          inc(i);
          list^.internmodenumber:=i;
          list:=list^.next;
        end;
       newModeList.loHiModeNr[driverNr].lo:=1;
       newModeList.loHiModeNr[driverNr].hi:=i;
     end;
    { TP-like mode stuff }
    if not assigned(ModeList) then
      begin
        new(ModeList);
        System.move(mode, ModeList^, sizeof(Mode));
      end
    else
      begin
        list := ModeList;
        { go to the end of the list }
        while assigned(list^.next) do
          list:=list^.next;
        new(NewLst);
        list^.next := NewLst;
        System.move(mode, NewLst^, sizeof(Mode));
      end;
  end;


  procedure initmode(var mode: TModeInfo);
  {********************************************************}
  { Procedure InitMode()                                   }
  {--------------------------------------------------------}
  { This routine initialized the mode to default values.   }
  {********************************************************}
   begin
     FillChar(mode,sizeof(Mode),#0);
   end;


   function searchmode(ReqDriver : smallint; var reqmode: smallint): PModeInfo;
  {********************************************************}
  { Procedure SearchMode()                                 }
  {--------------------------------------------------------}
  { This routine searches the list of recognized modes,    }
  { and tries to find the <reqmode> in the <reqdriver>     }
  { return nil if not found, otherwise returns the found   }
  { structure.                                             }
  { note: if reqmode = -32768, the first mode available    }
  {   for reqdriver is returned (JM)                       }
  {       if reqmode = -32767, the last mode available     }
  {   for reqdriver is returned (JM)                       }
  {********************************************************}
    var
     list, lastModeInfo: PModeInfo;
     x,y: longint;
    begin
{$ifdef logging}
      LogLn('Searching for driver '+strf(reqdriver)+' and mode '+strf(reqmode));
{$endif logging}
      if (reqDriver >= lowNewDriver) and
         (reqDriver <= highNewDriver) then
        begin
          case reqMode of
            -32768:
              begin
                reqMode := newModeList.loHiModeNr[reqDriver].lo;
                searchMode := newModeList.modeinfo[reqDriver];
              end;
            -32767:
              begin
                reqMode := newModeList.loHiModeNr[reqDriver].hi;
                searchMode := nil;
                { Are there any modes available for this driver? }
                if reqMode <> -1 then
                  begin
                    list := newModeList.modeinfo[reqDriver];
                    while assigned(list^.next) do
                      list := list^.next;
                    searchMode := list;
                  end;
              end;
            else
              begin
                list := newModeList.modeinfo[reqDriver];
                searchMode := nil;
                if not assigned(list) then
                  exit;
                if mode2res(reqMode,x,y) then
                  begin
                    x := pred(x);
                    y := pred(y);
                    while assigned(list) and
                          ((list^.maxx < x) or
                           ((list^.maxx = x) and
                            (list^.maxy < y))) do
                      list := list^.next;
                    if not assigned(list) or
                       (list^.maxx <> x) or
                       (list^.maxy <> y) then
                      list := nil;
                    searchmode := list;
                  end
                else
                  begin
                    while assigned(list) and
                          (list^.internModeNumber <> reqMode) do
                      list := list^.next;
                    searchMode := list;
                  end;

              end;
          end;
          exit;
        end;
       searchmode := nil;
       list := ModeList;
       If assigned(list) then
         lastModeInfo := list;
       { go to the end of the list }
       while assigned(list) do
         begin
{$ifdef logging}
           Log('Found driver '+strf(list^.DriverNumber)+
               ' and mode $'+hexstr(list^.ModeNumber,4)+'...');
{$endif logging}
             if ((list^.DriverNumber = ReqDriver) and
                 ((list^.ModeNumber = ReqMode) or
                  { search for lowest mode }
                  (reqMode = -32768))) or
                 { search for highest mode }
                ((reqMode = -32767) and
                 (lastModeInfo^.driverNumber = reqDriver) and
                 ((list^.driverNumber <> lastModeInfo^.driverNumber) or
                   not(assigned(list^.next)))) then
               begin
{$ifdef logging}
                 LogLn('Accepted!');
{$endif logging}
                 searchmode := list;
                 If reqMode = -32768 then
                   reqMode := list^.ModeNumber
                 else if reqMode = -32767 then
                   begin
                     reqMode := lastModeInfo^.ModeNumber;
                     searchMode := lastModeInfo;
                   end;
                 exit;
               end;
{$ifdef logging}
             LogLn('Rejected.');
{$endif logging}
             lastModeInfo := list;
             list:=list^.next;
          end;
    end;


{-----------------------------------------------------------------------}
{                          External routines                            }
{-----------------------------------------------------------------------}

   function GetModeName(ModeNumber: smallint): string;
  {********************************************************}
  { Function GetModeName()                                 }
  {--------------------------------------------------------}
  { Checks  the known video list, and returns ModeName     }
  { string. On error returns an empty string.              }
  {********************************************************}
    var
     mode: PModeInfo;
    begin
      mode:=nil;
      GetModeName:='';
      { only search in the current driver modes ... }
      if (ModeNumber >= lowNewMode) and (ModeNumber <= highNewMode) then
          mode:=SearchMode(IntCurrentNewDriver,ModeNumber)
      else
          mode:=SearchMode(IntCurrentDriver,ModeNumber);
      if assigned(mode) then
          GetModeName:=Mode^.ModeName
      else
         _GraphResult := grInvalidMode;
    end;

   function GetGraphMode: smallint;
     begin
      GetGraphMode := IntCurrentMode;
     end;

   function GetMaxMode: smallint;
   { I know , i know, this routine is very slow, and it would }
   { be much easier to sort the linked list of possible modes }
   { instead of doing this, but I'm lazy!! And anyways, the   }
   { speed of the routine here is not that important....      }
    var
     i: smallint;
     mode: PModeInfo;
    begin
      mode:=nil;
      i:=0;
      repeat
        inc(i);
        { mode 0 always exists... }
        { start search at 1..     }
        mode:=SearchMode(IntCurrentNewDriver,i);
      until not assigned(mode);
      GetMaxMode:=i;
    end;


    procedure GetModeRange(GraphDriver: smallint; var LoMode,
      HiMode: smallint);
      var
       mode : PModeInfo;
     begin
       {$ifdef logging}
       LogLn('GetModeRange : Enter ('+strf(GraphDriver)+')');
       {$endif}
       HiMode:=-1;
       mode := nil;
       { First search if the graphics driver is supported ..  }
       { since mode zero is always supported.. if that driver }
       { is supported it should return something...           }

       { not true, e.g. VESA doesn't have a mode 0. Changed so}
       { -32768 means "return lowest mode in second parameter }
       { also, under VESA some modes may not be supported     }
       { (e.g. $108 here) while some with a higher number can }
       { be supported ($112 and onward), so I also added that }
       { -32767 means "return highest mode in second parameter}
       { This whole system should be overhauled though to work}
       { without such hacks (JM)                              }
       loMode := -32768;
       mode := SearchMode(GraphDriver, loMode);
       { driver not supported...}
       if not assigned(mode) then
         begin
           loMode := -1;
           exit;
         end;
       {$ifdef logging}
       LogLn('GetModeRange : Mode '+strf(lomode)+' found');
       {$endif}
       { now it exists... find highest available mode... }
       hiMode := -32767;
       mode:=SearchMode(GraphDriver,hiMode);
     end;


  procedure SetGraphMode(mode: smallint);
    var
     modeinfo: PModeInfo;
     usedDriver: Integer;
     dummyNewModeNr: smallint;
    begin
      { check if the mode exists... }
      { Depending on the modenumber, we search using the old or new }
      { graphdriver number (because once we entered graphmode,      }
      { getgraphmode() returns the old mode number and              }
      { both setgraphmode(getgraphmode) and setgraphmode(mAAAxBBB)  }
      { have to work (JM)                                           }
      case mode of
        detectMode:
          begin
            mode := -32767;
            usedDriver := IntcurrentNewDriver;
            modeInfo := searchmode(IntcurrentNewDriver,mode);
          end;
        lowNewMode..highNewMode:
          begin
            usedDriver := IntcurrentNewDriver;
            modeInfo := searchmode(IntcurrentNewDriver,mode);
          end;
        else
          begin
            usedDriver := IntcurrentDriver;
            modeinfo := searchmode(IntcurrentDriver,mode);
          end;
      end;
      if not assigned(modeinfo) then
        begin
{$ifdef logging}
          LogLn('Mode setting failed in setgraphmode pos 1');
{$endif logging}
          _GraphResult := grInvalidMode;
          exit;
       end;
    { reset all hooks...}
    DefaultHooks;

      { required hooks - returns error if no hooks to these }
      { routines.                                           }
      if assigned(modeinfo^.DirectPutPixel) then
         DirectPutPixel := modeinfo^.DirectPutPixel
      else
        begin
{$ifdef logging}
          LogLn('Mode setting failed in setgraphmode pos 2');
{$endif logging}
         DefaultHooks;
         _Graphresult := grInvalidMode;
         exit;
       end;

      if assigned(modeinfo^.PutPixel) then
         PutPixel := modeinfo^.PutPixel
      else
        begin
{$ifdef logging}
          LogLn('Mode setting failed in setgraphmode pos 3');
{$endif logging}
         DefaultHooks;
         _Graphresult := grInvalidMode;
         exit;
       end;

      if assigned(modeinfo^.GetPixel) then
         GetPixel := modeinfo^.GetPixel
      else
        begin
{$ifdef logging}
          LogLn('Mode setting failed in setgraphmode pos 4');
{$endif logging}
         DefaultHooks;
         _Graphresult := grInvalidMode;
         exit;
       end;

     if assigned(modeinfo^.SetRGBPalette) then
         SetRGBPalette := modeinfo^.SetRGBPalette
     else
        begin
{$ifdef logging}
          LogLn('Mode setting failed in setgraphmode pos 5');
{$endif logging}
         DefaultHooks;
         _Graphresult := grInvalidMode;
         exit;
       end;

     if assigned(modeinfo^.GetRGBPalette) then
         GetRGBPalette := modeinfo^.GetRGBPalette
     else
        begin
{$ifdef logging}
          LogLn('Mode setting failed in setgraphmode pos 6');
{$endif logging}
         DefaultHooks;
         _Graphresult := grInvalidMode;
         exit;
       end;

      { optional hooks. }
      if assigned(modeinfo^.SetAllPalette) then
        SetAllPalette := modeinfo^.SetAllPalette;

      if assigned(modeinfo^.ClearViewPort) then
         ClearViewPort := modeinfo^.ClearViewPort;
      if assigned(modeinfo^.PutImage) then
         PutImage := modeinfo^.PutImage;
      if assigned(modeinfo^.GetImage) then
         GetImage := modeinfo^.GetImage;
      if assigned(modeinfo^.ImageSize) then
         ImageSize := modeinfo^.ImageSize;
      if assigned(modeinfo^.GetScanLine) then
         GetScanLine := modeinfo^.GetScanLine;
      if assigned(modeinfo^.Line) then
         Line := modeinfo^.Line;
      if assigned(modeinfo^.InternalEllipse) then
         InternalEllipse := modeinfo^.InternalEllipse;
      if assigned(modeinfo^.PatternLine) then
         PatternLine := modeinfo^.PatternLine;
      if assigned(modeinfo^.HLine) then
         Hline := modeinfo^.Hline;
      if assigned(modeinfo^.Vline) then
         VLine := modeinfo^.VLine;
      if assigned(modeInfo^.SetVisualPage) then
         SetVisualPage := modeInfo^.SetVisualPage;
      if assigned(modeInfo^.SetActivePage) then
         SetActivePage := modeInfo^.SetActivePage;
      if assigned(modeInfo^.OutTextXY) then
         OutTextXY:=modeInfo^.OutTextXY;
      if assigned(modeInfo^.SetBkColor) then
         SetBkColor:=modeInfo^.SetBkColor;
      if assigned(modeInfo^.GetBkColor) then
         GetBkColor:=modeInfo^.GetBkColor;

      IntCurrentMode := modeinfo^.ModeNumber;
      { fix by nickysn: always set *both* IntCurrentDriver and IntCurrentNewDriver to the
        *correct* old & new style driver numbers. Solves a lot of bugs, e.g. when using InitGraph
	with a new style driver/mode and then later using SetGraphMode(GetGraphMode) }
      IntCurrentDriver := modeinfo^.DriverNumber;
      res2Mode(modeinfo^.maxx+1, modeinfo^.maxy+1, modeinfo^.maxColor, IntCurrentNewDriver, dummyNewModeNr);
{$ifdef logging}
       logln('Entering mode '+strf(intCurrentMode)+' of driver '+strf(intCurrentDriver));
{$endif logging}
      XAspect := modeinfo^.XAspect;
      YAspect := modeinfo^.YAspect;
      MaxX := modeinfo^.MaxX;
      MaxY := modeinfo^.MaxY;
{$ifdef logging}
      logln('maxx = '+strf(maxx)+', maxy = '+strf(maxy));
{$endif logging}
      HardwarePages := modeInfo^.HardwarePages;
      MaxColor := modeinfo^.MaxColor;
      PaletteSize := modeinfo^.PaletteSize;
      { is this a direct color mode? }
      DirectColor := modeinfo^.DirectColor;
      { now actually initialize the video mode...}
      { check first if the routine exists        }
      if not assigned(modeinfo^.InitMode) then
        begin
{$ifdef logging}
          LogLn('Mode setting failed in setgraphmode pos 7');
{$endif logging}
          DefaultHooks;
          _GraphResult := grInvalidMode;
          exit;
        end;
      _GraphResult := grOk;
      modeinfo^.InitMode;
      if _GraphResult <> grOk then
        begin
          DefaultHooks;
          exit;
        end;
      isgraphmode := true;
      { It is very important that this call be made }
      { AFTER the other variables have been setup.  }
      { Since it calls some routines which rely on  }
      { those variables.                            }
      SetActivePage(0);
      SetVisualPage(0);
      SetViewPort(0,0,MaxX,MaxY,TRUE);
      GraphDefaults;
    end;

    procedure RestoreCrtMode;
  {********************************************************}
  { Procedure RestoreCRTMode()                             }
  {--------------------------------------------------------}
  { Returns to the video mode which was set before the     }
  { InitGraph. Hardware state is set to the old values.    }
  {--------------------------------------------------------}
  { NOTE: -                                                }
  {       -                                                }
  {********************************************************}
     begin
       isgraphmode := false;
       RestoreVideoState;
     end;



  function InstallUserDriver(Name: string; AutoDetectPtr: Pointer): smallint;
   begin
     _graphResult := grError;
     InstallUserDriver:=grError;
   end;

  function RegisterBGIDriver(driver: pointer): smallint;

   begin
     _graphResult := grError;
     RegisterBGIDriver:=grError;
   end;



{ ----------------------------------------------------------------- }


  Procedure Arc(X,Y : smallint; StAngle,EndAngle,Radius: word);

{   var
    OldWriteMode: word;}

   Begin
     { Only if we are using thickwidths lines do we accept }
     { XORput write modes.                                 }
{     OldWriteMode := CurrentWriteMode;
     if (LineInfo.Thickness = NormWidth) then
       CurrentWriteMode := NormalPut;}
     InternalEllipse(X,Y,Radius,(longint(Radius)*XAspect) div YAspect,StAngle,Endangle,{$ifdef fpc}@{$endif}DummyPatternLine);
{     CurrentWriteMode := OldWriteMode;}
   end;


 procedure Ellipse(X,Y : smallint; stAngle, EndAngle: word; XRadius,YRadius: word);
  Begin
    InternalEllipse(X,Y,XRadius,YRadius,StAngle,Endangle,{$ifdef fpc}@{$endif}DummyPatternLine);
  end;


 procedure FillEllipse(X, Y: smallint; XRadius, YRadius: Word);
  {********************************************************}
  { Procedure FillEllipse()                                }
  {--------------------------------------------------------}
  { Draws a filled ellipse using (X,Y) as a center point   }
  { and XRadius and YRadius as the horizontal and vertical }
  { axes. The ellipse is filled with the current fill color}
  { and fill style, and is bordered with the current color.}
  {********************************************************}
  begin
    InternalEllipse(X,Y,XRadius,YRadius,0,360,PatternLine)
  end;



 procedure CircleDefault(X, Y: smallint; Radius:Word);
  {********************************************************}
  { Draws a circle centered at X,Y with the given Radius.  }
  {********************************************************}
  { Important notes:                                       }
  {  - Thickwidth circles use the current write mode, while}
  {    normal width circles ALWAYS use CopyPut/NormalPut   }
  {    mode. (Tested against VGA BGI driver -CEC 13/Aug/99 }
  {********************************************************}
  var OriginalArcInfo: ArcCoordsType;
      OldWriteMode: word;

  begin
     if (Radius = 0) then
          Exit;

     if (Radius = 1) then
     begin
      { only normal put mode is supported by a call to PutPixel }
          PutPixel(X, Y, CurrentColor);
          Exit;
     end;

     { save state of arc information }
     { because it is not needed for  }
     { a circle call.                }
     System.move(ArcCall,OriginalArcInfo, sizeof(ArcCall));
     if LineInfo.Thickness = Normwidth then
       begin
             OldWriteMode := CurrentWriteMode;
             CurrentWriteMode := CopyPut;
       end;
     { Adjust for screen aspect ratio }
     InternalEllipse(X,Y,Radius,(longint(Radius)*XAspect) div YAspect,0,360,{$ifdef fpc}@{$endif}DummyPatternLine);
     if LineInfo.Thickness = Normwidth then
         CurrentWriteMode := OldWriteMode;
     { restore arc information }
     System.move(OriginalArcInfo, ArcCall,sizeof(ArcCall));
 end;

 procedure SectorPL(x1,x2,y: smallint); {$ifndef fpc}far;{$endif fpc}
 var plx1, plx2: smallint;
 begin
   If (x1 = -maxsmallint) Then
     If (x2 = maxsmallint-1) Then
       { no ellipse points drawn on this line }
       If (((Y < ArcCall.Y) and (Y > ArcCall.YStart)) or
          ((Y > ArcCall.Y) and (Y < ArcCall.YStart))) Then
         { there is a part of the sector at this y coordinate, but no    }
         { ellips points are plotted on this line, so draw a patternline }
         { between the lines connecting (arccall.x,arccall.y) with       }
         { the start and the end of the arc (JM)                         }
         { use: y-y1=(y2-y1)/(x2-x1)*(x-x1) =>                           }
         { x = (y-y1)/(y2-y1)*(x2-x1)+x1                                 }
         Begin
           plx1 := (y-ArcCall.Y)*(ArcCall.XStart-ArcCall.X)
                   div (ArcCall.YStart-ArcCall.Y)+ArcCall.X;
           plx2 := (y-ArcCall.Y)*(ArcCall.XEnd-ArcCall.X)
                   div (ArcCall.YEnd-ArcCall.Y)+ArcCall.X;
           If plx1 > plx2 then
             begin
               plx1 := plx1 xor plx2;
               plx2 := plx1 xor plx2;
               plx1 := plx1 xor plx2;
             end;
         End
       { otherwise two points which have nothing to do with the sector }
       Else exit
     Else
       { the arc is plotted at the right side, but not at the left side, }
       { fill till the line between (ArcCall.X,ArcCall.Y) and            }
       { (ArcCall.XStart,ArcCall.YStart)                                 }
       Begin
         If (y < ArcCall.Y) then
           begin
             plx1 := (y-ArcCall.Y)*(ArcCall.XEnd-ArcCall.X)
                     div (ArcCall.YEnd-ArcCall.Y)+ArcCall.X
           end
         else if (y > ArcCall.Y) then
           begin
             plx1 := (y-ArcCall.Y)*(ArcCall.XStart-ArcCall.X)
                     div (ArcCall.YStart-ArcCall.Y)+ArcCall.X
             end
         else plx1 := ArcCall.X;
         plx2 := x2;
       End
   Else
     If (x2 = maxsmallint-1) Then
       { the arc is plotted at the left side, but not at the rigth side.   }
       { the right limit can be either the first or second line. Just take }
       { the closest one, but watch out for division by zero!              }
       Begin
         If (y < ArcCall.Y) then
           begin
             plx2 := (y-ArcCall.Y)*(ArcCall.XStart-ArcCall.X)
                     div (ArcCall.YStart-ArcCall.Y)+ArcCall.X
           end
         else if (y > ArcCall.Y) then
           begin
             plx2 := (y-ArcCall.Y)*(ArcCall.XEnd-ArcCall.X)
                     div (ArcCall.YEnd-ArcCall.Y)+ArcCall.X
           end
         else plx2 := ArcCall.X;
         plx1 := x1;
       End
     Else
       { the arc is plotted at both sides }
       Begin
         plx1 := x1;
         plx2 := x2;
       End;
   If plx2 > plx1 then
     Begin
       PatternLine(plx1,plx2,y);
     end;
 end;

 procedure Sector(x, y: smallint; StAngle,EndAngle, XRadius, YRadius: Word);
  begin
     internalellipse(x,y,XRadius, YRadius, StAngle, EndAngle, {$ifdef fpc}@{$endif}SectorPL);
     Line(ArcCall.XStart, ArcCall.YStart, x,y);
     Line(x,y,ArcCall.Xend,ArcCall.YEnd);
  end;



   procedure SetFillStyle(Pattern : word; Color: word);

   begin
     { on invalid input, the current fill setting will be }
     { unchanged.                                         }
     if (Pattern > UserFill) or (Color > GetMaxColor) then
      begin
{$ifdef logging}
           logln('invalid fillstyle parameters');
{$endif logging}
           _GraphResult := grError;
           exit;
      end;
     FillSettings.Color := Color;
     FillSettings.Pattern := Pattern;
   end;


  procedure SetFillPattern(Pattern: FillPatternType; Color: word);
  {********************************************************}
  { Changes the Current FillPattern to a user defined      }
  { pattern and changes also the current fill color.       }
  { The FillPattern is saved in the FillPattern array so   }
  { it can still be used with SetFillStyle(UserFill,Color) }
  {********************************************************}
   var
    i: smallint;

   begin
     if Color > GetMaxColor then
       begin
{$ifdef logging}
            logln('invalid fillpattern parameters');
{$endif logging}
            _GraphResult := grError;
            exit;
       end;

     FillSettings.Color := Color;
     FillSettings.Pattern := UserFill;

     { Save the pattern in the buffer }
     For i:=1 to 8 do
       FillPatternTable[UserFill][i] := Pattern[i];

   end;

  procedure Bar(x1,y1,x2,y2:smallint);
  {********************************************************}
  { Important notes for compatibility with BP:             }
  {     - WriteMode is always CopyPut                      }
  {     - No contour is drawn for the lines                }
  {********************************************************}
  var y               : smallint;
      origcolor       : longint;
      origlinesettings: Linesettingstype;
      origwritemode   : smallint;
   begin
     origlinesettings:=lineinfo;
     origcolor:=CurrentColor;
     if y1>y2 then
       begin
          y:=y1;
          y1:=y2;
          y2:=y;
       end;
     if x1>x2 then
       begin
          y:=x1;
          x1:=x2;
          x2:=y;
       end;

     { Always copy mode for Bars }
     origwritemode := CurrentWriteMode;
     CurrentWriteMode := CopyPut;

     { All lines used are of this style }
     Lineinfo.linestyle:=solidln;
     Lineinfo.thickness:=normwidth;

     case Fillsettings.pattern of
     EmptyFill :
       begin
         Currentcolor:=CurrentBkColor;
         for y:=y1 to y2 do
           Hline(x1,x2,y);
       end;
     SolidFill :
       begin
         CurrentColor:=FillSettings.color;
           for y:=y1 to y2 do
              Hline(x1,x2,y);
       end;
     else
      Begin
        CurrentColor:=FillSettings.color;
        for y:=y1 to y2 do
          patternline(x1,x2,y);
      end;
    end;
    CurrentColor:= Origcolor;
    LineInfo := OrigLineSettings;
    CurrentWriteMode := OrigWritemode;
   end;




procedure bar3D(x1, y1, x2, y2 : smallint;depth : word;top : boolean);
var
 origwritemode : smallint;
 OldX, OldY : smallint;
begin
  if x1 > x2 then
  begin
    OldX := x1;
    x1 := x2;
    x2 := OldX;
  end;
  if y1 > y2 then
  begin
    OldY := y1;
    y1 := y2;
    y2 := OldY;
  end;

  origwritemode := CurrentWriteMode;
  CurrentWriteMode := CopyPut;
  Bar(x1,y1,x2,y2);
  Rectangle(x1,y1,x2,y2);

  { Current CP should not be updated in Bar3D }
  { therefore save it and then restore it on  }
  { exit.                                     }
  OldX := CurrentX;
  OldY := CurrentY;

  if top then begin
    Moveto(x1,y1);
    Lineto(x1+depth,y1-depth);
    Lineto(x2+depth,y1-depth);
    Lineto(x2,y1);
  end;
  if Depth <> 0 then
    Begin
      Moveto(x2+depth,y1-depth);
      Lineto(x2+depth,y2-depth);
      Lineto(x2,y2);
    end;
  { restore CP }
  CurrentX := OldX;
  CurrentY := OldY;
  CurrentWriteMode := origwritemode;
end;



{--------------------------------------------------------------------------}
{                                                                          }
{                       COLOR AND PALETTE ROUTINES                         }
{                                                                          }
{--------------------------------------------------------------------------}


  procedure SetColor(Color: Word);

   Begin
     CurrentColor := Color;
   end;


  function GetColor: Word;

   Begin
     GetColor := CurrentColor;
   end;

  function GetBkColorDefault: Word;

   Begin
     GetBkColorDefault := CurrentBkColor;
   end;


  procedure SetBkColorDefault(ColorNum: Word);
  { Background color means background screen color in this case, and it is  }
  { INDEPENDANT of the viewport settings, so we must clear the whole screen }
  { with the color.                                                         }
   var
     ViewPort: ViewportType;
   Begin
     GetViewSettings(Viewport);
{$ifdef logging}
      logln('calling setviewport from setbkcolor');
{$endif logging}
     SetViewPort(0,0,MaxX,MaxY,FALSE);
{$ifdef logging}
      logln('calling setviewport from setbkcolor done');
{$endif logging}
     CurrentBkColor := ColorNum;
     {ClearViewPort;}
     if not DirectColor and (ColorNum<256) then
      SetRGBPalette(0,
          DefaultColors[ColorNum].Red,
          DefaultColors[ColorNum].Green,
          DefaultColors[ColorNum].Blue);
     SetViewport(ViewPort.X1,Viewport.Y1,Viewport.X2,Viewport.Y2,Viewport.Clip);
   end;


  function GetMaxColor: word;
  { Checked against TP VGA driver - CEC }

   begin
      GetMaxColor:=MaxColor-1; { based on an index of zero so subtract one }
   end;






   Procedure MoveRel(Dx, Dy: smallint);
    Begin
     CurrentX := CurrentX + Dx;
     CurrentY := CurrentY + Dy;
   end;

   Procedure MoveTo(X,Y: smallint);
  {********************************************************}
  { Procedure MoveTo()                                     }
  {--------------------------------------------------------}
  { Moves the current pointer in VIEWPORT relative         }
  { coordinates to the specified X,Y coordinate.           }
  {********************************************************}
    Begin
     CurrentX := X;
     CurrentY := Y;
    end;


function GraphErrorMsg(ErrorCode: smallint): string;
Begin
 GraphErrorMsg:='';
 case ErrorCode of
  grOk,grFileNotFound,grInvalidDriver: exit;
  grNoInitGraph: GraphErrorMsg:='Graphics driver not installed';
  grNotDetected: GraphErrorMsg:='Graphics hardware not detected';
  grNoLoadMem,grNoScanMem,grNoFloodMem: GraphErrorMsg := 'Not enough memory for graphics';
  grNoFontMem: GraphErrorMsg := 'Not enough memory to load font';
  grFontNotFound: GraphErrorMsg:= 'Font file not found';
  grInvalidMode: GraphErrorMsg := 'Invalid graphics mode';
  grError: GraphErrorMsg:='Graphics error';
  grIoError: GraphErrorMsg:='Graphics I/O error';
  grInvalidFont,grInvalidFontNum: GraphErrorMsg := 'Invalid font';
  grInvalidVersion: GraphErrorMsg:='Invalid driver version';
 end;
end;




  Function GetMaxX: smallint;
  { Routine checked against VGA driver - CEC }
   Begin
     GetMaxX := MaxX;
   end;

  Function GetMaxY: smallint;
  { Routine checked against VGA driver - CEC }
   Begin
    GetMaxY := MaxY;
   end;




Function GraphResult: smallint;
Begin
  GraphResult := _GraphResult;
  _GraphResult := grOk;
end;


  Function GetX: smallint;
   Begin
     GetX := CurrentX;
   end;


  Function GetY: smallint;
   Begin
     GetY := CurrentY;
   end;

   Function GetDriverName: string;
    begin
      GetDriverName:=DriverName;
    end;


   procedure graphdefaults;
   { PS: GraphDefaults does not ZERO the ArcCall structure }
   { so a call to GetArcCoords will not change even the    }
   { returned values even if GraphDefaults is called in    }
   { between.                                              }
    var
     i: smallint;
   begin
     lineinfo.linestyle:=solidln;
     lineinfo.thickness:=normwidth;
     { reset line style pattern }
     for i:=0 to 15 do
       LinePatterns[i] := TRUE;

     { By default, according to the TP prog's reference }
     { the default pattern is solid, and the default    }
     { color is the maximum color in the palette.       }
     fillsettings.color:=GetMaxColor;
     fillsettings.pattern:=solidfill;
     { GraphDefaults resets the User Fill pattern to $ff }
     { checked with VGA BGI driver - CEC                 }
     for i:=1 to 8 do
       FillPatternTable[UserFill][i] := $ff;


     CurrentColor:=white;


     ClipPixels := TRUE;
     { Reset the viewport }
     StartXViewPort := 0;
     StartYViewPort := 0;
     ViewWidth := MaxX;
     ViewHeight := MaxY;

     { Reset CP }
     CurrentX := 0;
     CurrentY := 0;

     SetBkColor(Black);

     { normal write mode }
     CurrentWriteMode := CopyPut;

     { set font style }
     CurrentTextInfo.font := DefaultFont;
     CurrentTextInfo.direction:=HorizDir;
     CurrentTextInfo.charsize:=1;
     CurrentTextInfo.horiz:=LeftText;
     CurrentTextInfo.vert:=TopText;

     {fix by nickysn: TP7 doesn't reset aspect ratio in GraphDefaults }
     {XAspect:=10000; YAspect:=10000;}
   end;


  procedure GetAspectRatio(var Xasp,Yasp : word);
  begin
    XAsp:=XAspect;
    YAsp:=YAspect;
  end;

  procedure SetAspectRatio(Xasp, Yasp : word);
  begin
    Xaspect:= XAsp;
    YAspect:= YAsp;
  end;


  procedure SetWriteMode(WriteMode : smallint);
  { TP sets the writemodes according to the following scheme (JM) }
   begin
     Case writemode of
       xorput, andput: CurrentWriteMode := XorPut;
       notput, orput, copyput: CurrentWriteMode := CopyPut;
     End;
   end;


  procedure GetFillSettings(var Fillinfo:Fillsettingstype);
   begin
     Fillinfo:=Fillsettings;
   end;

  procedure GetFillPattern(var FillPattern:FillPatternType);
   begin
     FillPattern:=FillpatternTable[UserFill];
   end;

  procedure DrawPoly(numpoints : word;var polypoints);
    type
      ppointtype = ^pointtype;
      pt = array[0..16000] of pointtype;
    var
      i, j, LastPolygonStart: longint;
      Closing: boolean;
    begin
      if numpoints < 2 then
        begin
          { nickysn: TP7 doesn't return error here }
          {_GraphResult := grError;}
          exit;
        end;
      Closing := false;
      LastPolygonStart := 0;
      for i:=0 to numpoints-2 do
      begin
        { skip an edge after each 'closing' edge }
        if not Closing then
          line(pt(polypoints)[i].x,
               pt(polypoints)[i].y,
               pt(polypoints)[i+1].x,
               pt(polypoints)[i+1].y);

        { check if the current edge is 'closing'. This means that it 'closes'
          the polygon by going back to the first point of the polygon.
          Also, 0-length edges are never considered 'closing'. }
        if ((pt(polypoints)[i+1].x <> pt(polypoints)[i].x) or
            (pt(polypoints)[i+1].y <> pt(polypoints)[i].y)) and
            (LastPolygonStart < i) and
           ((pt(polypoints)[i+1].x = pt(polypoints)[LastPolygonStart].x) and
            (pt(polypoints)[i+1].y = pt(polypoints)[LastPolygonStart].y)) then
        begin
          Closing := true;
          LastPolygonStart := i + 2;
        end
        else
          Closing := false;
      end;
    end;


  procedure PieSlice(X,Y,stangle,endAngle:smallint;Radius: Word);
  begin
    Sector(x,y,stangle,endangle,radius,(longint(Radius)*XAspect) div YAspect);
  end;

{

    This file is part of the Free Pascal run time library.
    Copyright (c) 1999-2000 by Thomas Schatzl and Carl Eric Codere

    This include implements polygon filling and flood filling.

    See the file COPYING.FPC, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}

{ simple descriptive name }
function max(a, b : Longint) : Longint;
begin
  max := b;
  if (a > b) then max := a;
end;

{ here too }
function min(a, b : Longint) : Longint;
begin
  min := b;
  if (a < b) then min := a;
end;

procedure fillpoly(numpoints : Word; var polypoints);

{ disable range check mode }
{$ifopt R+}
{$define OPT_R_WAS_ON}
{$R-}
{$endif}
type
  pedge = ^tedge;
  tedge = packed record
    yMin, yMax, x, dX, dY, frac : Longint;
  end;

  pedgearray = ^tedgearray;
  tedgearray = array[0..0] of tedge;

  ppedgearray = ^tpedgearray;
  tpedgearray = array[0..0] of pedge;

var
  nActive, nNextEdge : Longint;
  p0, p1 : pointtype;
  i, j, gap, x0, x1, y, nEdges : Longint;
  ET : pedgearray;
  GET, AET : ppedgearray;
  t : pedge;

  ptable : ^pointtype;

  LastPolygonStart : Longint;
  Closing, PrevClosing : Boolean;

begin
{ /********************************************************************
  * Add entries to the global edge table.  The global edge table has a
  * bucket for each scan line in the polygon. Each bucket contains all
  * the edges whose yMin == yScanline.  Each bucket contains the yMax,
  * the x coordinate at yMax, and the denominator of the slope (dX)
*/}
  getmem(et, sizeof(tedge) * numpoints);
  getmem(get, sizeof(pedge) * numpoints);
  getmem(aet, sizeof(pedge) * numpoints);

  ptable := @polypoints;

 { check for getmem success }

  nEdges := 0;
  LastPolygonStart := 0;
  Closing := false;
  for i := 0 to (numpoints-1) do begin
    p0 := ptable[i];
    if (i+1) >= numpoints then p1 := ptable[0]
    else p1 := ptable[i+1];
    { save the 'closing' flag for the previous edge }
    PrevClosing := Closing;
    { check if the current edge is 'closing'. This means that it 'closes'
      the polygon by going back to the first point of the polygon.
      Also, 0-length edges are never considered 'closing'. }
    if ((p1.x <> ptable[i].x) or
        (p1.y <> ptable[i].y)) and
        (LastPolygonStart < i) and
       ((p1.x = ptable[LastPolygonStart].x) and
        (p1.y = ptable[LastPolygonStart].y)) then
    begin
      Closing := true;
      LastPolygonStart := i + 2;
    end
    else
      Closing := false;
    { skip current edge if the previous edge was 'closing'. This is TP7 compatible }
    if PrevClosing then
      continue;
   { draw the edges }
{    nickysn: moved after drawing the filled area
    Line(p0.x,p0.y,p1.x,p1.y);}
   { ignore if this is a horizontal edge}
    if (p0.y = p1.y) then continue;
   { swap ptable if necessary to ensure p0 contains yMin}
    if (p0.y > p1.y) then begin
      p0 := p1;
      p1 := ptable[i];
    end;
   { create the new edge }
    et^[nEdges].ymin := p0.y;
    et^[nEdges].ymax := p1.y;
    et^[nEdges].x := p0.x;
    et^[nEdges].dX := p1.x-p0.x;
    et^[nEdges].dy := p1.y-p0.y;
    et^[nEdges].frac := 0;
    get^[nEdges] :=  @et^[nEdges];
    inc(nEdges);
  end;
 { sort the GET on ymin }
  gap := 1;
  while (gap < nEdges) do gap := 3*gap+1;
  gap := gap div 3;
  while (gap > 0) do begin
    for i := gap to (nEdges-1) do begin
      j := i - gap;
      while (j >= 0) do begin
        if (GET^[j]^.ymin <= GET^[j+gap]^.yMin) then break;
        t := GET^[j];
        GET^[j] := GET^[j+gap];
        GET^[j+gap] := t;
        dec(j, gap);
      end;
    end;
    gap := gap div 3;
  end;
  { initialize the active edge table, and set y to first entering edge}
  nActive := 0;
  nNextEdge := 0;

  y := GET^[nNextEdge]^.ymin;
  { Now process the edges using the scan line algorithm.  Active edges
  will be added to the Active Edge Table (AET), and inactive edges will
  be deleted.  X coordinates will be updated with incremental integer
  arithmetic using the slope (dY / dX) of the edges. }
  while (nNextEdge < nEdges) or (nActive <> 0) do begin
    {Move from the ET bucket y to the AET those edges whose yMin == y
    (entering edges) }
    while (nNextEdge < nEdges) and (GET^[nNextEdge]^.ymin = y) do begin
      AET^[nActive] := GET^[nNextEdge];
      inc(nActive);
      inc(nNextEdge);
    end;
    { Remove from the AET those entries for which yMax == y (leaving
    edges) }
    i := 0;
    while (i < nActive) do begin
      if (AET^[i]^.yMax = y) then begin
        dec(nActive);
        System.move(AET^[i+1], AET^[i], (nActive-i)*sizeof(pedge));
      end else
        inc(i);
    end;

    if (y >= 0) then begin
    {Now sort the AET on x.  Since the list is usually quite small,
    the sort is implemented as a simple non-recursive shell sort }

    gap := 1;
    while (gap < nActive) do gap := 3*gap+1;

    gap := gap div 3;
    while (gap > 0) do begin
      for i := gap to (nActive-1) do begin
        j := i - gap;
        while (j >= 0) do begin
          if (AET^[j]^.x <= AET^[j+gap]^.x) then break;
          t := AET^[j];
          AET^[j] := AET^[j+gap];
          AET^[j+gap] := t;
          dec(j, gap);
        end;
      end;
      gap := gap div 3;
    end;

    { Fill in desired pixels values on scan line y by using pairs of x
    coordinates from the AET }
    i := 0;
    while (i < (nActive - 1)) do begin
      x0 := AET^[i]^.x;
      x1 := AET^[i+1]^.x;
      {Left edge adjustment for positive fraction.  0 is interior. }
      if (AET^[i]^.frac >= 0) then inc(x0);
      {Right edge adjustment for negative fraction.  0 is exterior. }
      if (AET^[i+1]^.frac <= 0) then dec(x1);

      x0 := max(x0, 0);
      x1 := min(x1, viewWidth);
      { Draw interior spans}
      if (x1 >= x0) then begin
        PatternLine(x0, x1, y);
      end;

      inc(i, 2);
    end;

    end;

    { Update all the x coordinates.  Edges are scan converted using a
    modified midpoint algorithm (Bresenham's algorithm reduces to the
    midpoint algorithm for two dimensional lines) }
    for i := 0 to (nActive-1) do begin
      t := AET^[i];
      { update the fraction by dX}
      inc(t^.frac, t^.dX);

      if (t^.dX < 0) then
        while ( -(t^.frac) >= t^.dY) do begin
          inc(t^.frac, t^.dY);
          dec(t^.x);
        end
      else
        while (t^.frac >= t^.dY) do begin
          dec(t^.frac, t^.dY);
          inc(t^.x);
        end;
    end;
    inc(y);
    if (y >= ViewHeight) then break;
  end;

  { finally, draw the edges }
  LastPolygonStart := 0;
  Closing := false;
  for i := 0 to (numpoints-1) do begin
    p0 := ptable[i];
    if (i+1) >= numpoints then p1 := ptable[0]
    else p1 := ptable[i+1];
    { save the 'closing' flag for the previous edge }
    PrevClosing := Closing;
    { check if the current edge is 'closing'. This means that it 'closes'
      the polygon by going back to the first point of the polygon.
      Also, 0-length edges are never considered 'closing'. }
    if ((p1.x <> p0.x) or
        (p1.y <> p0.y)) and
        (LastPolygonStart < i) and
       ((p1.x = ptable[LastPolygonStart].x) and
        (p1.y = ptable[LastPolygonStart].y)) then
    begin
      Closing := true;
      LastPolygonStart := i + 2;
    end
    else
      Closing := false;
    { skip edge if the previous edge was 'closing'. This is TP7 compatible }
    if PrevClosing then
      continue;
   { draw the edges }
    Line(p0.x,p0.y,p1.x,p1.y);
  end;

  System.freemem(et, sizeof(tedge) * numpoints);
  System.freemem(get, sizeof(pedge) * numpoints);
  System.freemem(aet, sizeof(pedge) * numpoints);
end;


{ maximum supported Y resultion }
const
  MaxYRes = 2048;
  { changing this to 1 or 2 doesn't improve performance noticably }
  YResDiv = 4;

type
  PFloodLine = ^TFloodLine;
  TFloodLine = record
    next: PFloodLine;
    x1 : smallint;
    x2 : smallint;
    y  : smallint;
  end;

  TDrawnList  = Array[0..(MaxYRes - 1) div YResDiv] of PFloodLine;

var
   DrawnList : TDrawnList;
   Buffer : Record                         { Union for byte and word addressing of buffer }
     ByteIndex : Word;
     WordIndex : Word;
     Case Boolean Of
        False : (Bytes : Array [0..StdBufferSize-1] Of Byte);
        True  : (Words : Array [0..(StdBufferSize DIV 2)-1] Of Word);
     End;

  s1, s2, s3 : PWordArray;                { Three buffers for scanlines                 }


  Procedure PushPoint (x, y : smallint);
  {********************************************************}
  { Adds a  point to the list of points to check if we     }
  { need to draw. Doesn't add the point if there is a      }
  { buffer overflow.                                       }
  {********************************************************}
  Begin
    If Buffer.WordIndex<(StdBufferSize DIV 2)-3 then
     Begin
       Buffer.Words[Buffer.WordIndex]:=x;
       Buffer.Words[Buffer.WordIndex+1]:=y;
       Inc (Buffer.WordIndex,2);
     End
  End;

  Procedure PopPoint (Var x, y : smallint);
  {********************************************************}
  { Removes a point from the list of points to check, if   }
  { we try to access an illegal point, then the routine    }
  { returns -1,-1 as a coordinate pair.                    }
  {********************************************************}
  Begin
   If Buffer.WordIndex>1 then
    Begin
      x:=Buffer.Words[Buffer.WordIndex-2];
      y:=Buffer.Words[Buffer.WordIndex-1];
      Dec (Buffer.WordIndex,2);
    End
   Else
    Begin
      x:=-1;
      y:=-1;
    End;
  End;






  {********************************************************}
  { Procedure AddLinePoints()                              }
  {--------------------------------------------------------}
  { Adds a line segment to the list of lines which will be }
  { drawn to the screen. The line added is on the specified}
  { Y axis, from the x1 to x2 coordinates.                 }
  {********************************************************}
  Procedure AddLinePoints(x1,x2,y: smallint);
   var temp: PFloodLine;
   begin
     new(temp);
     temp^.x1 := x1;
     temp^.x2 := x2;
     temp^.y := y;
     temp^.next := DrawnList[y div YResDiv];
     DrawnList[y div YResDiv] := temp;
   end;

  {********************************************************}
  { Procedure AlreadyDrawn()                               }
  {--------------------------------------------------------}
  { This routine searches through the list of segments     }
  { which will be drawn to the screen, and determines  if  }
  { the specified point (x,y) will already be drawn.       }
  { i.e : Checks if the x,y point lies within a known      }
  { segment which will be drawn to the screen. This makes  }
  { sure that we don't draw some segments two times.       }
  { Return TRUE if the point is already in the segment list}
  { to draw, otherwise returns FALSE.                      }
  {********************************************************}
  Function AlreadyDrawn(x, y: smallint): boolean;
  var
    temp : PFloodLine;
   begin
    AlreadyDrawn := false;
    temp := DrawnList[y div YResDiv];
    while assigned(temp) do
      begin
        if (temp^.y = y) and
           (temp^.x1 <= x) and
           (temp^.x2 >= x) then
          begin
            AlreadyDrawn := true;
            exit;
          end;
        temp := temp^.next;
      end;
   end;

  {********************************************************}
  { Procedure CleanUpDrawnList                             }
  {--------------------------------------------------------}
  { removes all elements from the DrawnList. Doesn't init  }
  { elements of it with NILL                               }
  {********************************************************}
  Procedure CleanUpDrawnList;
  var
    l: longint;
    temp1, temp2: PFloodLine;
  begin
    for l := 0 to high(DrawnList) do
      begin
        temp1 := DrawnList[l];
        while assigned(temp1) do
          begin
            temp2 := temp1;
            temp1 := temp1^.next;
            dispose(temp2);
          end;
      end;
  end;


  Procedure FloodFill (x, y : smallint; Border: word);
  {********************************************************}
  { Procedure FloodFill()                                  }
  {--------------------------------------------------------}
  { This routine fills a region of the screen bounded by   }
  { the <Border> color. It uses the current fillsettings   }
  { for the flood filling. Clipping is supported, and      }
  { coordinates are local/viewport relative.               }
  {********************************************************}
  Var
   stemp: PWordArray;
   Beginx : smallint;
   d, e : Byte;
   Cont : Boolean;
   BackupColor : Word;
   x1, x2, prevy: smallint;
  Begin
    FillChar(DrawnList,sizeof(DrawnList),0);
    { init prevy }
    prevy := 32767;
    { Save current drawing color }
    BackupColor := CurrentColor;
    CurrentColor := FillSettings.Color;
    { MaxX is based on zero index }
    GetMem (s1,(ViewWidth+1)*2);  { A pixel color represents a word }
    GetMem (s2,(ViewWidth+1)*2);  { A pixel color represents a word }
    GetMem (s3,(ViewWidth+1)*2);  { A pixel color represents a word }
    if (not assigned(s1)) or (not assigned(s2)) or (not assigned(s3)) then
      begin
        _GraphResult := grNoFloodMem;
        exit;
      end;
    If (x<0) Or (y<0) Or
       (x>ViewWidth) Or (y>ViewHeight) then Exit;
    { Index of points to check  }
    Buffer.WordIndex:=0;
    PushPoint (x,y);
    While Buffer.WordIndex>0 Do
     Begin
       PopPoint (x,y);
       { Get the complete lines for the following }
       If y <> prevy then
         begin
           If (prevy - y = 1) then
             { previous line was one below the new one, so the previous s2 }
             { = new s1                                                    }
             Begin
               stemp := s3;
               s3 := s1;
               s1 := s2;
               s2 := stemp;
               GetScanline(0,ViewWidth,y-1,s2^);
             End
           Else If (y - prevy = 1) then
             { previous line was one above the new one, so the previous s3 }
             { = new s1                                                    }
             Begin
               stemp := s2;
               s2 := s1;
               s1 := s3;
               s3 := stemp;
               GetScanline(0,ViewWidth,y+1,s3^);
             End
           Else
             begin
               GetScanline(0,ViewWidth,y-1,s2^);
               GetScanline(0,ViewWidth,y,s1^);
               GetScanline(0,ViewWidth,y+1,s3^);
             end;
         end;
       prevy := y;
       { check the current scan line }
       While (s1^[x]<>Border) And (x<=ViewWidth) Do Inc (x);
       d:=0;
       e:=0;
       dec(x);
       Beginx:=x;
       REPEAT
         { check the above line }
         If y<ViewHeight then
           Begin
              Cont:=(s3^[x]<>Border) and (not AlreadyDrawn(x,y+1));
              If (e=0) And Cont then
                Begin
                  PushPoint (x,y+1);
                  e:=1;
                End
              Else
                If (e=1) And Not Cont then e:=0;
           End;
        { check the line below }
        If (y>0) then
          Begin
            Cont:=(s2^[x]<>Border) and (not AlreadyDrawn(x,y-1));
            If (d=0) And Cont then
              Begin
                PushPoint (x,y-1);
                d:=1;
              End
            Else
              If (d=1) And Not Cont then d:=0;
          End;
        Dec (x);
       Until (x<0) Or (s1^[x]=Border);
       { swap the values }
       x1:=x+1;
       x2:=BeginX;
       if x1 > x2 then
         Begin
           x:=x1;
           x1:=x2;
           x2:=x;
         end;
       { Add to the list of drawn lines }
       AddLinePoints(x1,x2,y);
       PatternLine (x1,x2,y);
     End; { end while }

    System.FreeMem (s1,(ViewWidth+1)*2);
    System.FreeMem (s2,(ViewWidth+1)*2);
    System.FreeMem (s3,(ViewWidth+1)*2);
    CleanUpDrawnList;
    CurrentColor := BackUpColor;
  End;

{ restore previous range check mode }
{$ifdef OPT_R_WAS_ON}
{$R+}
{$endif}

{
    This file is part of the Free Pascal run time library.
    Copyright (c) 1999-2000 by the Free Pascal development team

    See the file COPYING.FPC, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}

{***************************************************************************}
{                             Text output routines                          }
{***************************************************************************}

    const
       maxfonts    = 16;   { maximum possible fonts              }
       MaxChars    = 255;  { Maximum nr. of characters in a file }
       Prefix_Size = $80;  { prefix size to skip                 }
       SIGNATURE   = '+';  { Signature of CHR file               }

    type
      { Prefix header of Font file }
{      PFHeader = ^TFHeader;}
      TFHeader = packed record
         header_size: word;    {* Version 2.0 Header Format        *}
         font_name: array[1..4] of char;
         font_size: word;      {* Size in byte of file        *}
         font_major: byte;     {* Driver Version Information    *}
         font_minor: byte;
         min_major: byte;      {* BGI Revision Information         *}
         min_minor: byte;
      end;


      { Font record information }
{      PHeader = ^THeader;}
      THeader = packed record
        Signature:  char;     { signature byte                        }
        Nr_chars:   smallint;  { number of characters in file          }
        Reserved:   byte;
        First_char: byte;     { first character in file               }
        cdefs :     smallint;  { offset to character definitions       }
        scan_flag:  byte;     { TRUE if char is scanable              }
        org_to_cap: shortint;     { Height from origin to top of capitol  }
        org_to_base:shortint;     { Height from origin to baseline        }
        org_to_dec: shortint;     { Height from origin to bot of decender }
        _reserved: array[1..4] of char;
        Unused: byte;
      end;


      TOffsetTable =array[0..MaxChars] of smallint;
      TWidthTable  =array[0..MaxChars] of byte;

      tfontrec = packed record
        name : string[8];
        header : THeader;        { font header   }
        pheader : TFHeader;      { prefix header }
        offsets : TOffsetTable;
        widths : TWidthTable;
        instrlength: longint;    { length of instr, because instr can }
        instr : pchar;           { contain null characters            }
      end;



{      pStroke = ^TStroke;}
      TStroke = packed record
        opcode: byte;
        x: smallint;  { relative x offset character }
        y: smallint;  { relative y offset character }
      end;


      TStrokes = Array[0..1000] of TStroke;

      opcodes = (_END_OF_CHAR, _DO_SCAN, _DRAW := 253, _MOVE := 254 );


    var
       fonts : array[1..maxfonts] of tfontrec;
       Strokes: TStrokes; {* Stroke Data Base           *}
{       Stroke_count: Array[0..MaxChars] of smallint;} {* Stroke Count Table *}

{***************************************************************************}
{                         Internal support routines                         }
{***************************************************************************}

{$ifdef FPC_BIG_ENDIAN}
procedure swap_fheader(var h: tfheader);
(*
      TFHeader = packed record
         header_size: word;    {* Version 2.0 Header Format        *}
         font_name: array[1..4] of char;
         font_size: word;      {* Size in byte of file        *}
         font_major: byte;     {* Driver Version Information    *}
         font_minor: byte;
         min_major: byte;      {* BGI Revision Information         *}
         min_minor: byte;
      end;
*)
begin
  with h do
    begin
      header_size := swap(header_size);
      font_size := swap(font_size);
    end;
end;

procedure swap_header(var h: theader);
(*
      THeader = packed record
        Signature:  char;     { signature byte                        }
        Nr_chars:   smallint;  { number of characters in file          }
        Reserved:   byte;
        First_char: byte;     { first character in file               }
        cdefs :     smallint;  { offset to character definitions       }
        scan_flag:  byte;     { TRUE if char is scanable              }
        org_to_cap: shortint;     { Height from origin to top of capitol  }
        org_to_base:shortint;     { Height from origin to baseline        }
        org_to_dec: shortint;     { Height from origin to bot of decender }
        _reserved: array[1..4] of char;
        Unused: byte;
      end;
*)
begin
  with h do
    begin
      nr_chars := swap(nr_chars);
      cdefs := swap(cdefs);
    end;
end;


procedure swap_offsets(var t: toffsettable; start, len: longint);
(*
      TOffsetTable =array[0..MaxChars] of smallint;
*)
var
  i: longint;
begin
  for i := start to start+len-1 do
    t[i]:=Swap(t[i]);
end;
{$endif FPC_BIG_ENDIAN}


function ConvertString(const OrigString: String): String;
var
  i: Integer;
  ConvResult: String;
begin
  if GraphStringTransTable = nil then
    ConvertString := OrigString
  else
  begin
    SetLength(ConvResult, Length(OrigString));
    for i := 1 to Length(OrigString) do
      ConvResult[i] := GraphStringTransTable^[OrigString[i]];
    ConvertString := ConvResult;
  end;
end;


    function testfont(p : pchar) : boolean;

      begin
         testfont:=(p[0]='P') and
          (p[1]='K') and
          (p[2]=#8) and
          (p[3]=#8);
      end;


    function InstallUserFont(const FontFileName : string) : smallint;

      begin
         _graphresult:=grOk;
         { first check if we do not allocate too many fonts! }
         if installedfonts=maxfonts then
           begin
              _graphresult:=grError;
              InstallUserFont := DefaultFont;
              exit;
           end;
         inc(installedfonts);
         fonts[installedfonts].name:=FontFileName;
         fonts[installedfonts].instr := nil;
         fonts[installedfonts].instrlength := 0;
         InstallUserFont:=installedfonts;
      end;


    function Decode(byte1,byte2: char; var x,y: smallint): smallint;
    { This routines decoes a signle word in a font opcode section  }
    { to a stroke record.                                          }
      var
       b1,b2: shortint;
     Begin
       b1:=shortint(byte1);
       b2:=shortint(byte2);
       { Decode the CHR OPCODE }
       Decode:=byte((shortint(b1 and $80) shr 6)+(shortint(b2 and $80) shr 7));
       { Now get the X,Y coordinates        }
       { bit 0..7 only which are considered }
       { signed values.                     }
{ disable range check mode }
{$ifopt R+}
{$define OPT_R_WAS_ON}
{$R-}
{$endif}
       b1:=b1 and $7f;
       b2:=b2 and $7f;
       { Now if the MSB of these values are set }
       { then the value is signed, therefore we }
       { sign extend it...                      }
       if (b1 and $40)<>0 then b1:=b1 or $80;
       if (b2 and $40)<>0 then b2:=b2 or $80;
       x:=smallint(b1);
       y:=smallint(b2);
{ restore previous range check mode }
{$ifdef OPT_R_WAS_ON}
{$R+}
{$endif}
     end;


    function unpack(buf: pchar; index: smallint; var Stroke: TStrokes): smallint;

     var
      po: TStrokes;
      num_ops: smallint;
      opcode, i, opc: word;
      counter: smallint;
      lindex: smallint;
      jx, jy: smallint;
     begin
       num_ops := 0;
       counter := index;
       lindex :=0;


       while TRUE do    {* For each byte in buffer      *}
         Begin
           Inc(num_ops);  {* Count the operation                *}
           opcode := decode( buf[counter], buf[counter+1] ,jx, jy );
           Inc(counter,2);
           if( opcode = ord(_END_OF_CHAR) ) then break; {* Exit loop at end of char     *}
         end;

       counter:=index;

       for i:=0 to num_ops-1 do    {    /* For each opcode in buffer    */ }
         Begin
           opc := decode(buf[counter], buf[counter+1], po[lindex].x, po[lindex].y);  {* Decode the data field   *}
           inc(counter,2);
           po[lindex].opcode := opc;      {* Save the opcode            *}
           Inc(lindex);
         end;
       Stroke:=po;
       unpack := num_ops;       {* return OPS count             *}
     end;



    procedure GetTextPosition(var xpos,ypos: longint; const TextString: string);
     begin
         if CurrentTextInfo.Font = DefaultFont then
          begin
           if Currenttextinfo.direction=horizdir then
            begin
              case Currenttextinfo.horiz of
                   centertext : XPos:=(textwidth(textstring) shr 1);
                   lefttext   : XPos:=0;
                   righttext  : XPos:=textwidth(textstring);
              end;
              case Currenttextinfo.vert of
                  centertext : YPos:=-(textheight(textstring) shr 1);
                  bottomtext : YPos:=-textheight(textstring);
                  toptext    : YPos:=0;
              end;
            end else
            begin
              case Currenttextinfo.horiz of
                   centertext : XPos:=(textheight(textstring) shr 1);
                   lefttext   : XPos:=textheight(textstring);
                   righttext  : XPos:=textheight(textstring);
              end;
              case Currenttextinfo.vert of
                  centertext : YPos:=(textwidth(textstring) shr 1);
                  bottomtext : YPos:=0;
                  toptext    : YPos:=textwidth(textstring);
              end;
            end;
          end
         else
          begin
            if Currenttextinfo.direction=horizdir then
            begin
              case CurrentTextInfo.horiz of
                   centertext : XPos:=(textwidth(textstring) shr 1);
                   lefttext   : XPos:=0;
                   righttext  : XPos:=textwidth(textstring);
              end;
              case CurrentTextInfo.vert of
                  centertext : YPos:=(textheight(textstring) shr 1);
                  bottomtext : YPos:=0;
                  toptext    : YPos:=textheight(textstring);
              end;
            end else
            begin
              case CurrentTextInfo.horiz of
                   centertext : XPos:=(textheight(textstring) shr 1);
                   lefttext   : XPos:=0;
                   righttext  : XPos:=textheight(textstring);
              end;
              case CurrentTextInfo.vert of
                  centertext : YPos:=(textwidth(textstring) shr 1);
                  bottomtext : YPos:=0;
                  toptext    : YPos:=textwidth(textstring);
              end;
           end;
          end;
     end;

{***************************************************************************}
{                         Exported routines                                 }
{***************************************************************************}


    function RegisterBGIfont(font : pointer) : smallint;

      var
         hp : pchar;
         b : word;
         i: longint;
         Header: THeader;
         counter: longint;
         FontData: pchar;
         FHeader: TFHeader;
      begin
         RegisterBGIfont:=grInvalidFontNum;
         i:=0;
         { Check if the font header is valid first of all }
         if testfont(font) then
           begin
              hp:=pchar(font);
              { Move to EOF in prefix header }
              while (hp[i] <> chr($1a)) do Inc(i);
              System.move(hp[i+1],FHeader,sizeof(FHeader));
              System.move(hp[Prefix_Size],header,sizeof(Header));
{$ifdef FPC_BIG_ENDIAN}
              swap_fheader(fheader);
              swap_header(header);
{$endif FPC_BIG_ENDIAN}
              { check if the font name is already allocated? }
              i:=Prefix_Size+sizeof(Header);
              for b:=1 to installedfonts do
                begin
                   if fonts[b].name=FHeader.Font_name then
                     begin
                        System.move(FHeader,fonts[b].PHeader,sizeof(FHeader));
                        System.move(Header,fonts[b].Header,sizeof(Header));
                        System.move(hp[i],Fonts[b].Offsets[Fonts[b].Header.First_Char],Fonts[b].Header.Nr_chars*sizeof(smallint));
{$ifdef FPC_BIG_ENDIAN}
                        swap_offsets(Fonts[b].Offsets,Fonts[b].Header.First_Char,Fonts[b].Header.Nr_chars);
{$endif FPC_BIG_ENDIAN}
                        Inc(i,Fonts[b].Header.Nr_chars*sizeof(smallint));
                        System.move(hp[i],Fonts[b].Widths[Fonts[b].Header.First_Char],Fonts[b].Header.Nr_chars*sizeof(byte));
                        Inc(i,Fonts[b].Header.Nr_chars*sizeof(byte));
                        counter:=Fonts[b].PHeader.font_size+PREFIX_SIZE-i;
                        { allocate also space for null }
                        GetMem(FontData,Counter+1);
                        System.move(hp[i],FontData^,Counter);
                        { Null terminate the string }
                        FontData[counter+1] := #0;
                        if fonts[b].header.Signature<> SIGNATURE then
                          begin
                            _graphResult:=grInvalidFont;
                            System.Freemem(FontData, Counter+1);
                            exit;
                          end;
                        fonts[b].instr:=FontData;
                        fonts[b].instrlength:=Counter+1;
                        RegisterBGIfont:=b;
                     end;
                end;
           end
         else
           RegisterBGIFont:=grInvalidFont;
      end;



    procedure GetTextSettings(var TextInfo : TextSettingsType);

      begin
         textinfo:=currenttextinfo;
      end;



    function TextHeight(const TextString : string) : word;

      begin
         if Currenttextinfo.font=DefaultFont
            then TextHeight:=8*CurrentTextInfo.CharSize
            else
              TextHeight:=Trunc((fonts[Currenttextinfo.font].header.org_to_cap-
                fonts[Currenttextinfo.font].header.org_to_dec) * CurrentYRatio);
      end;

    function TextWidth(const TextString : string) : word;
      var i,x : smallint;
          c   : byte;
          s   : String;
      begin
         x := 0;
         { if this is the default font ... }
         if Currenttextinfo.font = Defaultfont then
            TextWidth:=length(TextString)*8*CurrentTextInfo.CharSize
         { This is a stroked font ... }
            else begin
               s := ConvertString(TextString);
               for i:=1 to length(s) do
                begin
                   c:=byte(s[i]);
{                   dec(c,fonts[Currenttextinfo.font].header.first_char);}
                   if (c-fonts[Currenttextinfo.font].header.first_char>=
                       fonts[Currenttextinfo.font].header.nr_chars) then
                     continue;
                   x:=x+byte(fonts[Currenttextinfo.font].widths[c]);
               end;
             TextWidth:=round(x * CurrentXRatio) ;
            end;
      end;

    procedure OutTextXYDefault(x,y : smallint;const TextString : string);

      type
       Tpoint = record
         X,Y: smallint;
       end;
      var
         ConvString    : String;
         i,j,k,c       : longint;
         xpos,ypos     : longint;
         counter       : longint;
         cnt1,cnt2     : smallint;
         cnt3,cnt4     : smallint;
         charsize      : word;
         WriteMode     : word;
         curX2, curY2, xpos2, ypos2, x2, y2: graph_float;
         oldvalues     : linesettingstype;
         fontbitmap    : TBitmapChar;
         chr           : char;
         curx2i,cury2i,
         xpos2i,ypos2i : longint;

      begin
         { save current write mode }
         WriteMode := CurrentWriteMode;
         CurrentWriteMode := NormalPut;
         GetTextPosition(xpos,ypos,textstring);
         X:=X-XPos; Y:=Y+YPos;
         XPos:=X; YPos:=Y;

         ConvString := ConvertString(TextString);
         CharSize := CurrentTextInfo.Charsize;
         if Currenttextinfo.font=DefaultFont then
         begin
           c:=length(ConvString);
           if CurrentTextInfo.direction=HorizDir then
           { Horizontal direction }
             begin
                for i:=0 to c-1 do
                  begin
                     xpos:=x+(i*8)*Charsize;
                     { we copy the character bitmap before accessing it }
                     { this improves speed on non optimizing compilers  }
                     { since it is one less address calculation.        }
                     Fontbitmap:=TBitmapChar(DefaultFontData[ConvString[i+1]]);
                     { no scaling }
                     if CharSize = 1 then
                      Begin
                        for j:=0 to 7 do
                           for k:=0 to 7 do
                             if Fontbitmap[j,k]<>0 then
                               PutPixel(xpos+k,j+y,CurrentColor)
                             else if DrawTextBackground then
                               PutPixel(xpos+k,j+y,CurrentBkColor);
                      end
                     else
                      { perform scaling of bitmap font }
                      Begin
                        j:=0;
                        cnt3:=0;

                        while j <= 7 do
                        begin
                          { X-axis scaling }
                          for cnt4 := 0 to charsize-1 do
                           begin
                             k:=0;
                             cnt2 := 0;
                             while k <= 7  do
                                begin
                                  for cnt1 := 0 to charsize-1 do
                                    begin
                                       If FontBitmap[j,k] <> 0 then
                                           PutPixel(xpos+cnt1+cnt2,y+cnt3+cnt4,CurrentColor)
                                       else if DrawTextBackground then
                                           PutPixel(xpos+cnt1+cnt2,y+cnt3+cnt4,CurrentBkColor);
                                    end;
                                    Inc(k);
                                    Inc(cnt2,charsize);
                                end;
                           end;
                          Inc(j);
                          Inc(cnt3,charsize);
                        end;
                      end;
                  end;
             end
           else
           { Vertical direction }
            begin
              for i:=0 to c-1 do
              begin

                chr := ConvString[i+1];
                Fontbitmap:=TBitmapChar(DefaultFontData[chr]);
                ypos := y-(i shl 3)*CharSize;

                { no scaling }
                if CharSize = 1 then
                 Begin
                   for j:=0 to 7 do
                      for k:=0 to 7 do
                        if Fontbitmap[j,k] <> 0 then
                          PutPixel(xpos+j,ypos-k,CurrentColor)
                        else if DrawTextBackground then
                          PutPixel(xpos+j,ypos-k,CurrentBkColor);
                 end
                else
                 { perform scaling of bitmap font }
                 Begin
                   j:=0;
                   cnt3:=0;

                   while j<=7 do
                   begin
                     { X-axis scaling }
                     for cnt4 := 0 to charsize-1 do
                      begin
                        k:=0;
                        cnt2 := 0;
                        while k<=7  do
                           begin
                             for cnt1 := 0 to charsize-1 do
                               begin
                                  If FontBitmap[j,k] <> 0 then
                                    PutPixel(xpos+cnt3-cnt4,ypos+cnt1-cnt2,CurrentColor)
                                  else if DrawTextBackground then
                                    PutPixel(xpos+cnt3-cnt4,ypos+cnt1-cnt2,CurrentBkColor);
                               end;
                               Inc(k);
                               Inc(cnt2,charsize);
                           end;
                      end;
                     Inc(j);
                     Inc(cnt3,charsize);
                   end;
                 end;
              end;
            end;
         end else
         { This is a stroked font which is already loaded into memory }
           begin
              getlinesettings(oldvalues);
              { reset line style to defaults }
              setlinestyle(solidln,oldvalues.pattern,normwidth);
              if Currenttextinfo.direction=vertdir then
                 xpos:=xpos + Textheight(ConvString);
              CurX2:=xpos; xpos2 := curX2; x2 := xpos2;
              CurY2:=ypos; ypos2 := curY2; y2 := ypos2;
{              x:=xpos; y:=ypos;}

              for i:=1 to length(ConvString) do
                begin
                   c:=byte(ConvString[i]);
{                   Stroke_Count[c] := }
                   unpack( fonts[CurrentTextInfo.font].instr,
                     fonts[CurrentTextInfo.font].Offsets[c], Strokes );
                   counter:=0;
                   while true do
                     begin
                         if CurrentTextInfo.direction=VertDir then
                           begin
                             xpos2:=x2-(Strokes[counter].Y*CurrentYRatio);
                             ypos2:=y2-(Strokes[counter].X*CurrentXRatio);
                           end
                         else
                           begin
                             xpos2:=x2+(Strokes[counter].X*CurrentXRatio) ;
                             ypos2:=y2-(Strokes[counter].Y*CurrentYRatio) ;
                           end;
                         case opcodes(Strokes[counter].opcode) of
                           _END_OF_CHAR: break;
                           _DO_SCAN: begin
                                    { Currently unsupported };
                                    end;
                           _MOVE : Begin
                                     CurX2 := XPos2;
                                     CurY2 := YPos2;
                                   end;
                           _DRAW: Begin
                                    curx2i:=trunc(CurX2);
                                    cury2i:=trunc(CurY2);
                                    xpos2i:=trunc(xpos2);
                                    ypos2i:=trunc(ypos2);
                                    { this optimization doesn't matter that much
                                    if (curx2i=xpos2i) then
                                      begin
                                         if (cury2i=ypos2i) then
                                           putpixel(curx2i,cury2i,currentcolor)
                                         else if (cury2i+1=ypos2i) or
                                           (cury2i=ypos2i+1) then
                                            begin
                                               putpixel(curx2i,cury2i,currentcolor);
                                               putpixel(curx2i,ypos2i,currentcolor);
                                            end
                                          else
                                            Line(curx2i,cury2i,xpos2i,ypos2i);
                                      end
                                    else if (cury2i=ypos2i) then
                                      begin
                                         if (curx2i+1=xpos2i) or
                                           (curx2i=xpos2i+1) then
                                            begin
                                               putpixel(curx2i,cury2i,currentcolor);
                                               putpixel(xpos2i,cury2i,currentcolor);
                                            end
                                          else
                                            Line(curx2i,cury2i,xpos2i,ypos2i);
                                      end
                                    else
                                    }
                                    Line(curx2i,cury2i,xpos2i,ypos2i);
                                    CurX2:=xpos2;
                                    CurY2:=ypos2;
                                  end;
                             else
                               Begin
                               end;
                            end;
                        Inc(counter);
                     end; { end while }
                   if Currenttextinfo.direction=VertDir then
                     y2:=y2-(byte(fonts[CurrenttextInfo.font].widths[c])*CurrentXRatio)
                   else
                     x2:=x2+(byte(fonts[Currenttextinfo.font].widths[c])*CurrentXRatio);
                end;
              setlinestyle( oldvalues.linestyle, oldvalues.pattern, oldvalues.thickness);
           end;
        { restore write mode }
        CurrentWriteMode := WriteMode;
      end;


    procedure OutText(const TextString : string);
      var x,y:smallint;
      begin
         { Save CP }
         x:=CurrentX;
         y:=CurrentY;
         OutTextXY(CurrentX,CurrentY,TextString);
         { If the direction is Horizontal and the justification left }
         { then and only then do we update the CP                    }
         if (Currenttextinfo.direction=HorizDir) and
           (Currenttextinfo.horiz=LeftText) then
               inc(x,textwidth(TextString));
         { Update the CP }
         CurrentX := X;
         CurrentY := Y;
      end;





    procedure SetTextJustify(horiz,vert : word);

      begin
         if (horiz<0) or (horiz>2) or
            (vert<0) or (vert>2) then
           begin
              _graphresult:=grError;
              exit;
           end;
         Currenttextinfo.horiz:=horiz;
         Currenttextinfo.vert:=vert;
      end;


    procedure SetTextStyle(font,direction : word;charsize : word);

      var
         f : file;
         Prefix: array[0..Prefix_Size-1] of char; {* File Prefix Holder         *}
         Length, Current: longint;
         FontData: Pchar;
         hp  : pchar;
         i   : longint;
      begin
         if font>installedfonts then
           begin
              _graphresult:=grInvalidFontNum;
              exit;
           end;

         Currenttextinfo.font:=font;
         if (direction<>HorizDir) and (direction<>VertDir) then
           direction:=HorizDir;
         Currenttextinfo.direction:=direction;
         { According to the Turbo Pascal programmer's reference }
         { maximum charsize for bitmapped font is 10            }
         if (CurrentTextInfo.Font = DefaultFont) and (Charsize > 10) then
            Currenttextinfo.charsize:=10
         else if charsize<1 then
            Currenttextinfo.charsize:=1
         else
            Currenttextinfo.charsize:=charsize;

         { This is only valid for stroked fonts }
{$ifdef logging}
         LogLn('(org_to_cap - org_to_dec): ' + strf(
                fonts[Currenttextinfo.font].header.org_to_cap-
                fonts[Currenttextinfo.font].header.org_to_dec));
{$endif logging}
         if (charsize <> usercharsize) then
           Case CharSize of
             1: Begin
                  CurrentXRatio := 0.55;
                  CurrentYRatio := 0.55;
                End;
             2: Begin
                  CurrentXRatio := 0.65;
                  CurrentYRatio := 0.65;
                End;
             3: Begin
                  CurrentXRatio := 0.75;
                  CurrentYRatio := 0.75;
                End;
             4: Begin
                  CurrentXRatio := 1.0;
                  CurrentYRatio := 1.0;
                End;
             5: Begin
                  CurrentXRatio := 1.3;
                  CurrentYRatio := 1.3;
                End;
             6: Begin
                  CurrentXRatio := 1.65;
                  CurrentYRatio := 1.65
                End;
             7: Begin
                  CurrentXRatio := 2.0;
                  CurrentYRatio := 2.0;
                End;
             8: Begin
                  CurrentXRatio := 2.5;
                  CurrentYRatio := 2.5;
                End;
             9: Begin
                  CurrentXRatio := 3.0;
                  CurrentYRatio := 3.0;
                End;
             10: Begin
                   CurrentXRatio := 4.0;
                   CurrentYRatio := 4.0;
                 End
           End;
         { if this is a stroked font then load it if not already loaded }
         { into memory...                                               }
         if (font>DefaultFont) and not assigned(fonts[font].instr) then
           begin
              assign(f,bgipath+fonts[font].name+'.CHR');
{$ifopt I+}
{$define IOCHECK_WAS_ON}
{$i-}
{$endif}
              reset(f,1);
{$ifdef IOCHECK_WAS_ON}
{$i+}
{$endif}
              if ioresult<>0 then
                begin
                   _graphresult:=grFontNotFound;
                   Currenttextinfo.font:=DefaultFont;
                   exit;
                end;
              {* Read in the file prefix        *}
              BlockRead(F, Prefix, Prefix_Size);
              hp:=Prefix;
              i:=0;
              while (hp[i] <> chr($1a)) do Inc(i);
              System.move(hp[i+1],fonts[font].PHeader,sizeof(TFHeader));
              (* Read in the Header file  *)
              BlockRead(F,fonts[font].Header,Sizeof(THeader));
{$ifdef FPC_BIG_ENDIAN}
              swap_fheader(fonts[font].PHeader);
              swap_header(fonts[font].Header);
{$endif FPC_BIG_ENDIAN}
              BlockRead(F,Fonts[font].Offsets[Fonts[font].Header.First_Char],Fonts[font].Header.Nr_chars*sizeof(smallint));
{$ifdef FPC_BIG_ENDIAN}
              swap_offsets(Fonts[font].Offsets,Fonts[font].Header.First_Char,Fonts[font].Header.Nr_chars);
{$endif FPC_BIG_ENDIAN}
              {*        Load the character width table into memory.                     *}
              BlockRead(F,Fonts[font].Widths[Fonts[font].Header.First_Char],Fonts[font].Header.Nr_chars*sizeof(byte));
              {*        Determine the length of the stroke database.                    *}
              current := FilePos( f );          {* Current file location        *}
              Seek( f, FileSize(F));            {* Go to the end of the file    *}
              length := FilePos( f );           {* Get the file length          *}
              Seek( f, current);        {* Restore old file location    *}
              {*        Load the stroke database.                                       *}
              { also allocate space for Null character   }
              Getmem(FontData, Length+1);          {* Create space for font data        *}

              BlockRead(F, FontData^, length-current);        {* Load the stroke data   *}
              FontData[length-current+1] := #0;

             if fonts[font].header.Signature<> SIGNATURE then
             begin
                _graphResult:=grInvalidFont;
                Currenttextinfo.font:=DefaultFont;
                System.Freemem(FontData, Length+1);
                exit;
             end;
             fonts[font].instr:=FontData;
             fonts[font].instrLength:=Length+1;


              if not testfont(Prefix) then
                begin
                   _graphresult:=grInvalidFont;
                   Currenttextinfo.font:=DefaultFont;
                   System.Freemem(FontData, Length+1);
                end;
              close(f);
           end;
      end;

    procedure SetUserCharSize(Multx,Divx,Multy,Divy : word);
      begin
         CurrentXRatio := MultX / DivX;
         CurrentYRatio := MultY / DivY;
      end;



  procedure internDetectGraph(var GraphDriver, GraphMode:smallint;
    calledFromInitGraph: boolean);
  var LoMode, HiMode: smallint;
      CpyMode: smallint;
      CpyDriver: smallint;
  begin
    HiMode := -1;
    LoMode := -1;
    if not calledFromInitGraph or
       (graphDriver < lowNewDriver) or
       (graphDriver > highNewDriver) then
      begin
        { Search lowest supported bitDepth }
        graphdriver := D1bit;
        while (graphDriver <= highNewDriver) and
              (hiMode = -1) do
          begin
            getModeRange(graphDriver,loMode,hiMode);
            inc(graphDriver);
          end;
        dec(graphdriver);
        if hiMode = -1 then
          begin
            _GraphResult := grNotDetected;
            exit;
          end;
        CpyMode := 0;
        repeat
           GetModeRange(GraphDriver,LoMode,HiMode);
           { save the highest mode possible...}
           {$ifdef logging}
           logln('Found driver '+strf(graphdriver)+' with modes '+
                  strf(lomode)+' - '+strf(himode));
           {$endif logging}
           if HiMode <> -1 then
             begin
               CpyMode:=HiMode;
               CpyDriver:=GraphDriver;
             end;
           { go to next driver if it exists...}
           Inc(graphDriver);
        until (graphDriver > highNewDriver);
      end
    else
      begin
        cpyMode := 0;
        getModeRange(graphDriver,loMode,hiMode);
        if hiMode <> -1 then
          begin
            cpyDriver := graphDriver;
            cpyMode := hiMode;
          end;
      end;
    if cpyMode = 0 then
      begin
        _GraphResult := grNotDetected;
        exit;
      end;
    _GraphResult := grOK;
    GraphDriver := CpyDriver;
    GraphMode := CpyMode;
  end;

  procedure detectGraph(var GraphDriver: smallint; var GraphMode:smallint);
  begin
    internDetectGraph(graphDriver,graphMode,false);
  end;

  procedure InitGraph(var GraphDriver:smallint;var GraphMode:smallint;
    const PathToDriver:String);
  const
    dirchar = System.DirectorySeparator;
  begin
    InitVars;
    { path to the fonts (where they will be searched)...}
    bgipath:=PathToDriver;
    if (Length(bgipath) > 0) and (bgipath[length(bgipath)]<>dirchar) then
      bgipath:=bgipath+dirchar;

    if not assigned(SaveVideoState) then
      RunError(216);
    DriverName:=InternalDriverName;   { DOS Graphics driver }

    if (Graphdriver=Detect)
       or (GraphMode = detectMode) then
      begin
        internDetectGraph(GraphDriver,GraphMode,true);
        If _GraphResult = grNotDetected then Exit;

        { _GraphResult is now already set to grOK by DetectGraph }
        IntCurrentDriver := GraphDriver;

        if (graphDriver >= lowNewDriver) and
           (graphDriver <= highNewDriver) then
          IntCurrentNewDriver := GraphDriver
        else IntCurrentNewDriver := -1;

        { Actually set the graph mode...}
        if firstCallOfInitgraph then
          begin
            SaveVideoState;
            firstCallOfInitgraph := false;
          end;
        SetGraphMode(GraphMode);
      end
    else
      begin
        { Search if that graphics modec actually exists...}
        if SearchMode(GraphDriver,GraphMode) = nil then
          begin
            _GraphResult := grInvalidMode;
            exit;
         end
        else
         begin
           _GraphResult := grOK;
           IntCurrentDriver := GraphDriver;

           if (graphDriver >= lowNewDriver) and
              (graphDriver <= highNewDriver) then
             IntCurrentNewDriver := GraphDriver
           else IntCurrentNewDriver := -1;

           if firstCallOfInitgraph then
             begin
               SaveVideoState;
               firstCallOfInitgraph := false;
             end;
           SetGraphMode(GraphMode);
         end;
      end;
  end;


 procedure SetDirectVideo(DirectAccess: boolean);
  begin
    DirectVideo := DirectAccess;
  end;

 function GetDirectVideo: boolean;
  begin
    GetDirectVideo := DirectVideo;
  end;

 procedure GraphExitProc; {$ifndef fpc} far; {$endif fpc}
 { deallocates all memory allocated by the graph unit }
  var
    list: PModeInfo;
    tmp : PModeInfo;
    c: longint;
  begin
   { restore old exitproc! }
   exitproc := exitsave;
   if IsGraphMode and ((errorcode<>0) or (erroraddr<>nil)) then
     CloseGraph;
   { release memory allocated for fonts }
   for c := 1 to installedfonts do
     with fonts[c] Do
       If assigned(instr) Then
         System.Freemem(instr,instrlength);
   { release memory allocated for modelist }
   list := ModeList;
   while assigned(list) do
     begin
       tmp := list;
       list:=list^.next;
       dispose(tmp);
     end;
   for c := lowNewDriver to highNewDriver do
     begin
       list := newModeList.modeinfo[c];
       while assigned(list) do
         begin
           tmp := list;
           list:=list^.next;
           dispose(tmp);
         end;
     end;
{$IFDEF DPMI}
  { We had copied the buffer of mode information }
  { and allocated it dynamically... now free it  }
  { Warning: if GetVESAInfo returned false, this buffer is not allocated! (JM)}
   If hasVesa then
     Dispose(VESAInfo.ModeList);
{$ENDIF}
  end;


procedure InitializeGraph;
begin
{$ifdef logging}
 assign(debuglog,'grlog.txt');
 rewrite(debuglog);
 close(debuglog);
{$endif logging}
 isgraphmode := false;
 ModeList := nil;
 fillChar(newModeList.modeinfo,sizeof(newModeList.modeinfo),#0);
 { lo and hi modenumber are -1 currently (no modes supported) }
 fillChar(newModeList.loHiModeNr,sizeof(newModeList.loHiModeNr),#255);
 SaveVideoState := nil;
 RestoreVideoState := nil;
 { This must be called at startup... because GetGraphMode may }
 { be called even when not in graph mode.                     }
{$ifdef logging}
 LogLn('Calling QueryAdapterInfo...');
{$endif logging}
 QueryAdapterInfo;
 { Install standard fonts }
 { This is done BEFORE startup... }
 InstalledFonts := 0;
 InstallUserFont('TRIP');
 InstallUserFont('LITT');
 InstallUserFont('SANS');
 InstallUserFont('GOTH');
 InstallUserFont('SCRI');
 InstallUserFont('SIMP');
 InstallUserFont('TSCR');
 InstallUserFont('LCOM');
 InstallUserFont('EURO');
 InstallUserFont('BOLD');
 { This installs an exit procedure which cleans up the mode list...}
 ExitSave := ExitProc;
 ExitProc := @GraphExitProc;
{$if defined(win32) and not defined(sdlgraph)}
 charmessagehandler:=nil;
{$endif win32 and not sdlgraph}
end;



{ used to create a file containing all calls to WM_PAINT
  WARNING this probably creates HUGE files PM }
{ $define DEBUG_WM_PAINT}
var
   savedscreen : hbitmap;
   graphrunning : boolean;
   graphdrawing : tcriticalsection;
   pens : array[0..15] of HPEN;
{$ifdef DEBUG_WM_PAINT}
   graphdebug : text;

const
   wm_paint_count : longint = 0;
var
{$endif DEBUG_WM_PAINT}
   oldbitmap : hgdiobj;
   pal : ^rgbrec;
//   SavePtr : pointer; { we don't use that pointer }
   MessageThreadHandle : Handle;
   MessageThreadID : DWord;

function GetPaletteEntry(r,g,b : word) : word;

  var
     dist,i,index,currentdist : longint;

  begin
     dist:=$7fffffff;
     index:=0;
     for i:=0 to maxcolors do
       begin
          currentdist:=abs(r-pal[i].red)+abs(g-pal[i].green)+
            abs(b-pal[i].blue);
          if currentdist<dist then
            begin
               index:=i;
               dist:=currentdist;
               if dist=0 then
                 break;
            end;
       end;
     GetPaletteEntry:=index;
  end;

procedure PutPixel16Win32GUI(x,y : smallint;pixel : word);

  var
     c : colorref;

  begin
    x:=x+startxviewport;
    y:=y+startyviewport;
    { convert to absolute coordinates and then verify clipping...}
    if clippixels then
      begin
         if (x<startxviewport) or (x>(startxviewport+viewwidth)) or
           (y<StartyViewPort) or (y>(startyviewport+viewheight)) then
           exit;
      end;
    if graphrunning then
      begin
         c:=RGB(pal[pixel].red,pal[pixel].green,pal[pixel].blue);
         EnterCriticalSection(graphdrawing);
         if drawtobitmap then
           SetPixelV(bitmapdc,x,y,c);
         if drawtoscreen then
           SetPixelV(windc,x,y,c);
         LeaveCriticalSection(graphdrawing);
      end;
  end;

function GetPixel16Win32GUI(x,y : smallint) : word;

  var
     c : COLORREF;

  begin
    x:=x+startxviewport;
    y:=y+startyviewport;
    { convert to absolute coordinates and then verify clipping...}
    if clippixels then
      begin
         if (x<startxviewport) or (x>(startxviewport+viewwidth)) or
           (y<StartyViewPort) or (y>(startyviewport+viewheight)) then
           exit;
      end;
    if graphrunning then
      begin
         EnterCriticalSection(graphdrawing);
         c:=Windows.GetPixel(bitmapdc,x,y);
         LeaveCriticalSection(graphdrawing);
         GetPixel16Win32GUI:=GetPaletteEntry(GetRValue(c),GetGValue(c),GetBValue(c));
      end
    else
      begin
        _graphresult:=grerror;
        exit;
      end;
  end;

procedure DirectPutPixel16Win32GUI(x,y : smallint);

  var
     col : longint;
     c,c2 : COLORREF;

  begin
    if graphrunning then
      begin
         EnterCriticalSection(graphdrawing);
         col:=CurrentColor;
         case currentwritemode of
           XorPut:
             Begin
                c2:=Windows.GetPixel(windc,x,y);
                c:=RGB(pal[col].red,pal[col].green,pal[col].blue) xor c2;
                if drawtobitmap then
                  SetPixelV(bitmapdc,x,y,c);
                if drawtoscreen then
                  SetPixelV(windc,x,y,c);
             End;
           AndPut:
             Begin
                c2:=Windows.GetPixel(windc,x,y);
                c:=RGB(pal[col].red,pal[col].green,pal[col].blue) and c2;
                if drawtobitmap then
                  SetPixelV(bitmapdc,x,y,c);
                if drawtoscreen then
                  SetPixelV(windc,x,y,c);
             End;
           OrPut:
             Begin
                c2:=Windows.GetPixel(windc,x,y);
                c:=RGB(pal[col].red,pal[col].green,pal[col].blue) or c2;
                if drawtobitmap then
                  SetPixelV(bitmapdc,x,y,c);
                if drawtoscreen then
                  SetPixelV(windc,x,y,c);
             End
           else
             Begin
                If CurrentWriteMode<>NotPut Then
                  col:=CurrentColor
                Else col := Not(CurrentColor);
                c:=RGB(pal[col].red,pal[col].green,pal[col].blue);
                if drawtobitmap then
                  SetPixelV(bitmapdc,x,y,c);
                if drawtoscreen then
                  SetPixelV(windc,x,y,c);
             End
         end;
         LeaveCriticalSection(graphdrawing);
      end;
  end;

var
   bitmapfontverticalcache : array[0..255] of HBITMAP;
   bitmapfonthorizoncache : array[0..255] of HBITMAP;

procedure OutTextXYWin32GUI(x,y : smallint;const TextString : string);

  type
   Tpoint = record
     X,Y: smallint;
   end;
  var
     i,j,k,c       : longint;
     xpos,ypos     : longint;
     counter       : longint;
     cnt1,cnt2     : smallint;
     cnt3,cnt4     : smallint;
     charsize      : word;
     WriteMode     : word;
     curX2, curY2, xpos2, ypos2, x2, y2: graph_float;
     oldvalues     : linesettingstype;
     fontbitmap    : TBitmapChar;
     chr           : char;
     curx2i,cury2i,
     xpos2i,ypos2i : longint;
     charbitmap,oldcharbitmap : HBITMAP;
     chardc : HDC;
     color : longint;
     brushwin,oldbrushwin,brushbitmap,oldbrushbitmap : HBRUSH;
     bitmaprgn,winrgn : HRGN;

  begin
     { save current write mode }
     WriteMode := CurrentWriteMode;
     CurrentWriteMode := NormalPut;
     GetTextPosition(xpos,ypos,textstring);
     X:=X-XPos; Y:=Y+YPos;
     XPos:=X; YPos:=Y;
     CharSize := CurrentTextInfo.Charsize;
     if Currenttextinfo.font=DefaultFont then
     begin
       if CurrentTextInfo.direction=HorizDir then
       { Horizontal direction }
         begin
            if (x>viewwidth) or (y>viewheight) or
              (x<0) or (y<0) then
              begin
                 CurrentWriteMode:=WriteMode;
                 exit;
              end;
            EnterCriticalSection(graphdrawing);
            c:=length(textstring);
            chardc:=CreateCompatibleDC(windc);
            if currentcolor<>white then
              begin
                 color:=RGB(pal[currentcolor].red,pal[currentcolor].green,
                   pal[currentcolor].blue);

                 if drawtoscreen then
                   begin
                      brushwin:=CreateSolidBrush(color);
                      oldbrushwin:=SelectObject(windc,brushwin);
                   end;

                 if drawtobitmap then
                   begin
                      brushbitmap:=CreateSolidBrush(color);
                      oldbrushbitmap:=SelectObject(bitmapdc,brushbitmap);
                   end;
              end;
            inc(x,startxviewport);
            inc(y,startyviewport);

            { let windows do the clipping }
            if drawtobitmap then
              begin
                 bitmaprgn:=CreateRectRgn(startxviewport,startyviewport,
                   startxviewport+viewwidth+1,startyviewport+viewheight+1);
                 SelectClipRgn(bitmapdc,bitmaprgn);
              end;

            if drawtoscreen then
              begin
                 winrgn:=CreateRectRgn(startxviewport,startyviewport,
                   startxviewport+viewwidth+1,startyviewport+viewheight+1);
                 SelectClipRgn(windc,winrgn);
              end;

            for i:=0 to c-1 do
              begin
                 xpos:=x+(i*8)*Charsize;
                 if bitmapfonthorizoncache[byte(textstring[i+1])]=0 then
                   begin
                      charbitmap:=CreateCompatibleBitmap(windc,8,8);
                      if charbitmap=0 then
                        writeln('Bitmap konnte nicht erzeugt werden!');
                      oldcharbitmap:=SelectObject(chardc,charbitmap);
                      Fontbitmap:=TBitmapChar(DefaultFontData[textstring[i+1]]);

                      for j:=0 to 7 do
                         for k:=0 to 7 do
                           if Fontbitmap[j,k]<>0 then
                             SetPixelV(chardc,k,j,$ffffff)
                           else
                             SetPixelV(chardc,k,j,0);
                      bitmapfonthorizoncache[byte(textstring[i+1])]:=charbitmap;
                      SelectObject(chardc,oldcharbitmap);
                   end;
                 oldcharbitmap:=SelectObject(chardc,bitmapfonthorizoncache[byte(textstring[i+1])]);
                 if CharSize=1 then
                   begin
                      if currentcolor=white then
                        begin
                           if drawtoscreen then
                             BitBlt(windc,xpos,y,8,8,chardc,0,0,SRCPAINT);
                           if drawtobitmap then
                             BitBlt(bitmapdc,xpos,y,8,8,chardc,0,0,SRCPAINT);
                        end
                      else
                        begin
                           { could we do this with one pattern operation ?? }
                           { we would need something like DSnaSPao }
                           if drawtoscreen then
                             begin
                                // ROP $00220326=DSna
                                BitBlt(windc,xpos,y,8,8,chardc,0,0,$00220326);
                                // ROP $00EA02E9 = DPSao
                                BitBlt(windc,xpos,y,8,8,chardc,0,0,$00EA02E9);
                             end;

                           if drawtobitmap then
                             begin
                                BitBlt(bitmapdc,xpos,y,8,8,chardc,0,0,$00220326);
                                BitBlt(bitmapdc,xpos,y,8,8,chardc,0,0,$00EA02E9);
                             end;
                        end;
                   end
                 else
                   begin
                      if currentcolor=white then
                        begin
                           if drawtoscreen then
                             StretchBlt(windc,xpos,y,8*charsize,8*charsize,chardc,0,0,8,8,SRCPAINT);
                           if drawtobitmap then
                             StretchBlt(bitmapdc,xpos,y,8*charsize,8*charsize,chardc,0,0,8,8,SRCPAINT);
                        end
                      else
                        begin
                           { could we do this with one pattern operation ?? }
                           { we would need something like DSnaSPao }
                           if drawtoscreen then
                             begin
                                // ROP $00220326=DSna
                                StretchBlt(windc,xpos,y,8*charsize,8*charsize,chardc,0,0,8,8,$00220326);
                                // ROP $00EA02E9 = DPSao
                                StretchBlt(windc,xpos,y,8*charsize,8*charsize,chardc,0,0,8,8,$00EA02E9);
                             end;
                           if drawtobitmap then
                             begin
                                StretchBlt(bitmapdc,xpos,y,8*charsize,8*charsize,chardc,0,0,8,8,$00220326);
                                StretchBlt(bitmapdc,xpos,y,8*charsize,8*charsize,chardc,0,0,8,8,$00EA02E9);
                             end;
                        end;
                   end;
                 SelectObject(chardc,oldcharbitmap);
              end;
           if currentcolor<>white then
             begin
                 if drawtoscreen then
                   begin
                      SelectObject(windc,oldbrushwin);
                      DeleteObject(brushwin);
                   end;

                 if drawtobitmap then
                   begin
                      SelectObject(bitmapdc,oldbrushbitmap);
                      DeleteObject(brushbitmap);
                   end;
             end;
           { release clip regions }
           if drawtobitmap then
             begin
               SelectClipRgn(bitmapdc,0);
               DeleteObject(bitmaprgn);
             end;
           if drawtoscreen then
             begin
                SelectClipRgn(windc,0);
                DeleteObject(winrgn);
             end;
           DeleteDC(chardc);
           LeaveCriticalSection(graphdrawing);
         end
       else
       { Vertical direction }
         begin
            if (x>viewwidth) or (y>viewheight) or
              (x<0) or (y<0) then
              begin
                 CurrentWriteMode:=WriteMode;
                 exit;
              end;
            EnterCriticalSection(graphdrawing);
            c:=length(textstring);
            chardc:=CreateCompatibleDC(windc);
            if currentcolor<>white then
              begin
                 color:=RGB(pal[currentcolor].red,pal[currentcolor].green,
                   pal[currentcolor].blue);

                 if drawtoscreen then
                   begin
                      brushwin:=CreateSolidBrush(color);
                      oldbrushwin:=SelectObject(windc,brushwin);
                   end;

                 if drawtobitmap then
                   begin
                      brushbitmap:=CreateSolidBrush(color);
                      oldbrushbitmap:=SelectObject(bitmapdc,brushbitmap);
                   end;
              end;
            inc(x,startxviewport);
            inc(y,startyviewport);
            { let windows do the clipping }
            if drawtoscreen then
              begin
                 winrgn:=CreateRectRgn(startxviewport,startyviewport,
                   startxviewport+viewwidth+1,startyviewport+viewheight+1);
                 SelectClipRgn(windc,winrgn);
              end;

            if drawtobitmap then
              begin
                 bitmaprgn:=CreateRectRgn(startxviewport,startyviewport,
                   startxviewport+viewwidth+1,startyviewport+viewheight+1);
                 SelectClipRgn(bitmapdc,bitmaprgn);
              end;
            for i:=0 to c-1 do
              begin
                 ypos:=y+1-((i+1)*8)*CharSize;
                 if bitmapfontverticalcache[byte(textstring[i+1])]=0 then
                   begin
                      charbitmap:=CreateCompatibleBitmap(windc,8,8);
                      if charbitmap=0 then
                        writeln('Bitmap konnte nicht erzeugt werden!');
                      oldcharbitmap:=SelectObject(chardc,charbitmap);
                      Fontbitmap:=TBitmapChar(DefaultFontData[textstring[i+1]]);

                      for j:=0 to 7 do
                         for k:=0 to 7 do
                           if Fontbitmap[j,k]<>0 then
                             SetPixelV(chardc,j,7-k,$ffffff)
                           else
                             SetPixelV(chardc,j,7-k,0);
                      bitmapfontverticalcache[byte(textstring[i+1])]:=charbitmap;
                      SelectObject(chardc,oldcharbitmap);
                   end;
                 oldcharbitmap:=SelectObject(chardc,bitmapfontverticalcache[byte(textstring[i+1])]);
                 if CharSize=1 then
                   begin
                      if currentcolor=white then
                        begin
                           if drawtoscreen then
                             BitBlt(windc,x,ypos,8,8,chardc,0,0,SRCPAINT);
                           if drawtobitmap then
                             BitBlt(bitmapdc,x,ypos,8,8,chardc,0,0,SRCPAINT);
                        end
                      else
                        begin
                           { could we do this with one pattern operation ?? }
                           { we would need something like DSnaSPao }
                           if drawtoscreen then
                             begin
                                // ROP $00220326=DSna
                                BitBlt(windc,x,ypos,8,8,chardc,0,0,$00220326);
                                // ROP $00EA02E9 = DPSao
                                BitBlt(windc,x,ypos,8,8,chardc,0,0,$00EA02E9);
                             end;
                           if drawtobitmap then
                             begin
                                BitBlt(bitmapdc,x,ypos,8,8,chardc,0,0,$00220326);
                                BitBlt(bitmapdc,x,ypos,8,8,chardc,0,0,$00EA02E9);
                             end;
                        end;
                   end
                 else
                   begin
                      if currentcolor=white then
                        begin
                           if drawtoscreen then
                             StretchBlt(windc,x,ypos,8*charsize,8*charsize,chardc,0,0,8,8,SRCPAINT);
                           if drawtobitmap then
                             StretchBlt(bitmapdc,x,ypos,8*charsize,8*charsize,chardc,0,0,8,8,SRCPAINT);
                        end
                      else
                        begin
                           { could we do this with one pattern operation ?? }
                           { we would need something like DSnaSPao }
                           if drawtoscreen then
                             begin
                                // ROP $00220326=DSna
                                StretchBlt(windc,x,ypos,8*charsize,8*charsize,chardc,0,0,8,8,$00220326);
                                // ROP $00EA02E9 = DPSao
                                StretchBlt(windc,x,ypos,8*charsize,8*charsize,chardc,0,0,8,8,$00EA02E9);
                             end;
                           if drawtobitmap then
                             begin
                                StretchBlt(bitmapdc,x,ypos,8*charsize,8*charsize,chardc,0,0,8,8,$00220326);
                                StretchBlt(bitmapdc,x,ypos,8*charsize,8*charsize,chardc,0,0,8,8,$00EA02E9);
                             end;
                        end;
                   end;
                 SelectObject(chardc,oldcharbitmap);
              end;
           if currentcolor<>white then
             begin
                if drawtoscreen then
                  begin
                     SelectObject(windc,oldbrushwin);
                     DeleteObject(brushwin);
                  end;

                if drawtobitmap then
                  begin
                     SelectObject(bitmapdc,oldbrushbitmap);
                     DeleteObject(brushbitmap);
                  end;
             end;
           { release clip regions }
           if drawtoscreen then
             begin
                SelectClipRgn(windc,0);
                DeleteObject(winrgn);
             end;
           if drawtobitmap then
             begin
                SelectClipRgn(bitmapdc,0);
                DeleteObject(bitmaprgn);
             end;
           DeleteDC(chardc);
           LeaveCriticalSection(graphdrawing);
        end;
     end else
     { This is a stroked font which is already loaded into memory }
       begin
          getlinesettings(oldvalues);
          { reset line style to defaults }
          setlinestyle(solidln,oldvalues.pattern,normwidth);
          if Currenttextinfo.direction=vertdir then
             xpos:=xpos + Textheight(textstring);
          CurX2:=xpos; xpos2 := curX2; x2 := xpos2;
          CurY2:=ypos; ypos2 := curY2; y2 := ypos2;
{              x:=xpos; y:=ypos;}

          for i:=1 to length(textstring) do
            begin
               c:=byte(textstring[i]);
{                   Stroke_Count[c] := }
               unpack( fonts[CurrentTextInfo.font].instr,
                 fonts[CurrentTextInfo.font].Offsets[c], Strokes );
               counter:=0;
               while true do
                 begin
                     if CurrentTextInfo.direction=VertDir then
                       begin
                         xpos2:=x2-(Strokes[counter].Y*CurrentYRatio);
                         ypos2:=y2-(Strokes[counter].X*CurrentXRatio);
                       end
                     else
                       begin
                         xpos2:=x2+(Strokes[counter].X*CurrentXRatio) ;
                         ypos2:=y2-(Strokes[counter].Y*CurrentYRatio) ;
                       end;
                     case opcodes(Strokes[counter].opcode) of
                       _END_OF_CHAR: break;
                       _DO_SCAN: begin
                                { Currently unsupported };
                                end;
                       _MOVE : Begin
                                 CurX2 := XPos2;
                                 CurY2 := YPos2;
                               end;
                       _DRAW: Begin
                                curx2i:=trunc(CurX2);
                                cury2i:=trunc(CurY2);
                                xpos2i:=trunc(xpos2);
                                ypos2i:=trunc(ypos2);
                                { this optimization doesn't matter that much
                                if (curx2i=xpos2i) then
                                  begin
                                     if (cury2i=ypos2i) then
                                       putpixel(curx2i,cury2i,currentcolor)
                                     else if (cury2i+1=ypos2i) or
                                       (cury2i=ypos2i+1) then
                                        begin
                                           putpixel(curx2i,cury2i,currentcolor);
                                           putpixel(curx2i,ypos2i,currentcolor);
                                        end
                                      else
                                        Line(curx2i,cury2i,xpos2i,ypos2i);
                                  end
                                else if (cury2i=ypos2i) then
                                  begin
                                     if (curx2i+1=xpos2i) or
                                       (curx2i=xpos2i+1) then
                                        begin
                                           putpixel(curx2i,cury2i,currentcolor);
                                           putpixel(xpos2i,cury2i,currentcolor);
                                        end
                                      else
                                        Line(curx2i,cury2i,xpos2i,ypos2i);
                                  end
                                else
                                }
                                Line(curx2i,cury2i,xpos2i,ypos2i);
                                CurX2:=xpos2;
                                CurY2:=ypos2;
                              end;
                         else
                           Begin
                           end;
                        end;
                    Inc(counter);
                 end; { end while }
               if Currenttextinfo.direction=VertDir then
                 y2:=y2-(byte(fonts[CurrenttextInfo.font].widths[c])*CurrentXRatio)
               else
                 x2:=x2+(byte(fonts[Currenttextinfo.font].widths[c])*CurrentXRatio);
            end;
          setlinestyle( oldvalues.linestyle, oldvalues.pattern, oldvalues.thickness);
       end;
    { restore write mode }
    CurrentWriteMode := WriteMode;
  end;

procedure HLine16Win32GUI(x,x2,y: smallint);

   var
      c,c2 : COLORREF;
      col,i : longint;
      oldpen,pen : HPEN;

   Begin
      if graphrunning then
        begin
           { must we swap the values? }
           if x>x2 then
             Begin
               x:=x xor x2;
               x2:=x xor x2;
               x:=x xor x2;
             end;
           if ClipPixels then
             begin
                if (x>ViewWidth) or (y<0) or (y>ViewHeight) or (x2<0) then
                  exit;
                if x<0 then
                  x:=0;
                if x2>ViewWidth then
                  x2:=ViewWidth;
             end;
           X:=X+StartXViewPort;
           X2:=X2+StartXViewPort;
           Y:=Y+StartYViewPort;
           Case CurrentWriteMode of
             AndPut:
               Begin
                  EnterCriticalSection(graphdrawing);
                  col:=CurrentColor;
                  for i:=x to x2 do
                    begin
                       c2:=Windows.GetPixel(windc,i,y);
                       c:=RGB(pal[col].red,pal[col].green,pal[col].blue) and c2;
                       if drawtobitmap then
                         SetPixelV(bitmapdc,i,y,c);

                       if drawtoscreen then
                         SetPixelV(windc,i,y,c);
                    end;
                  LeaveCriticalSection(graphdrawing);
               End;
             XorPut:
               Begin
                  EnterCriticalSection(graphdrawing);
                  col:=CurrentColor;
                  for i:=x to x2 do
                    begin
                       c2:=Windows.GetPixel(windc,i,y);
                       c:=RGB(pal[col].red,pal[col].green,pal[col].blue) xor c2;

                       if drawtobitmap then
                         SetPixelV(bitmapdc,i,y,c);

                       if drawtoscreen then
                         SetPixelV(windc,i,y,c);
                    end;
                  LeaveCriticalSection(graphdrawing);
               End;
             OrPut:
               Begin
                  EnterCriticalSection(graphdrawing);
                  col:=CurrentColor;
                  for i:=x to x2 do
                    begin
                       c2:=Windows.GetPixel(windc,i,y);
                       c:=RGB(pal[col].red,pal[col].green,pal[col].blue) or c2;

                       if drawtobitmap then
                         SetPixelV(bitmapdc,i,y,c);

                       if drawtoscreen then
                         SetPixelV(windc,i,y,c);
                    end;
                  LeaveCriticalSection(graphdrawing);
               End
             Else
               Begin
                  If CurrentWriteMode<>NotPut Then
                    col:=CurrentColor
                  Else col:=Not(CurrentColor);
                  EnterCriticalSection(graphdrawing);
                  if x2-x<=2 then
                    begin
                       c:=RGB(pal[col].red,pal[col].green,pal[col].blue);
                       for x := x to x2 do
                         begin
                            if drawtobitmap then
                              SetPixelV(bitmapdc,x,y,c);
                            if drawtoscreen then
                              SetPixelV(windc,x,y,c);
                         end;
                    end
                  else
                    begin
                       if (col>=0) and (col<=high(pens)) then
                         begin
                            if pens[col]=0 then
                              begin
                                 c:=RGB(pal[col].red,pal[col].green,pal[col].blue);
                                 pens[col]:=CreatePen(PS_SOLID,1,c);
                              end;
                            pen:=pens[col];
                         end
                       else
                         begin
                            c:=RGB(pal[col].red,pal[col].green,pal[col].blue);
                            pen:=CreatePen(PS_SOLID,1,c);
                         end;

                       if drawtobitmap then
                         begin
                            oldpen:=SelectObject(bitmapdc,pen);
                            Windows.MoveToEx(bitmapdc,x,y,nil);
                            Windows.LineTo(bitmapdc,x2+1,y);
                            SelectObject(bitmapdc,oldpen);
                         end;

                       if drawtoscreen then
                         begin
                            oldpen:=SelectObject(windc,pen);
                            Windows.MoveToEx(windc,x,y,nil);
                            Windows.LineTo(windc,x2+1,y);
                            SelectObject(windc,oldpen);
                         end;

                       if (col<0) or (col>high(pens)) then
                         DeleteObject(pen);
                    end;
                   LeaveCriticalSection(graphdrawing);
               End;
           End;
        end;
   end;

procedure VLine16Win32GUI(x,y,y2: smallint); {$ifndef fpc}far;{$endif fpc}

 var
  ytmp: smallint;
  col,c : longint;
  oldpen,pen : HPEN;

Begin
  { must we swap the values? }
  if y >= y2 then
   Begin
     ytmp := y2;
     y2 := y;
     y:= ytmp;
   end;
 if ClipPixels then
   begin
      if (x>ViewWidth) or (x<0) or (y>ViewHeight) or (y2<0) then
        exit;
      if y<0 then
        y:=0;
      if y2>ViewHeight then
        y2:=ViewHeight;
   end;
  { First convert to global coordinates }
  X   := X + StartXViewPort;
  Y2  := Y2 + StartYViewPort;
  Y   := Y + StartYViewPort;
  if currentwritemode=normalput then
    begin
       col:=CurrentColor;
       EnterCriticalSection(graphdrawing);
       if y2-y<=2 then
         begin
            c:=RGB(pal[col].red,pal[col].green,pal[col].blue);
            for y := y to y2 do
              begin
                 if drawtobitmap then
                   SetPixelV(bitmapdc,x,y,c);
                 if drawtoscreen then
                   SetPixelV(windc,x,y,c);
              end;
         end
       else
         begin
            if (col>=0) and (col<=high(pens)) then
              begin
                 if pens[col]=0 then
                   begin
                      c:=RGB(pal[col].red,pal[col].green,pal[col].blue);
                      pens[col]:=CreatePen(PS_SOLID,1,c);
                   end;
                 pen:=pens[col];
              end
            else
              begin
                 c:=RGB(pal[col].red,pal[col].green,pal[col].blue);
                 pen:=CreatePen(PS_SOLID,1,c);
              end;

            if drawtobitmap then
              begin
                 oldpen:=SelectObject(bitmapdc,pen);
                 Windows.MoveToEx(bitmapdc,x,y,nil);
                 Windows.LineTo(bitmapdc,x,y2+1);
                 SelectObject(bitmapdc,oldpen);
              end;

            if drawtoscreen then
              begin
                 oldpen:=SelectObject(windc,pen);
                 Windows.MoveToEx(windc,x,y,nil);
                 Windows.LineTo(windc,x,y2+1);
                 SelectObject(windc,oldpen);
              end;
            if (col<0) or (col>high(pens)) then
              DeleteObject(pen);
         end;
       LeaveCriticalSection(graphdrawing);
    end
  else
    for y := y to y2 do Directputpixel(x,y)
End;

procedure Circle16Win32GUI(X, Y: smallint; Radius:Word);

  var
     bitmaprgn,winrgn : HRGN;
     col,c : longint;
     oldpen,pen : HPEN;
     OriginalArcInfo: ArcCoordsType;
     OldWriteMode: word;

  begin
     if (Radius = 0) then
          Exit;

     if (Radius = 1) then
       begin
          { only normal put mode is supported by a call to PutPixel }
          PutPixel(X, Y, CurrentColor);
          Exit;
       end;

     if (Radius = 2) then
       begin
          { only normal put mode is supported by a call to PutPixel }
          PutPixel(X-1, Y, CurrentColor);
          PutPixel(X+1, Y, CurrentColor);
          PutPixel(X, Y-1, CurrentColor);
          PutPixel(X, Y+1, CurrentColor);
          Exit;
       end;

     if LineInfo.Thickness = Normwidth then
       begin
          EnterCriticalSection(graphdrawing);
          { let windows do the clipping }
          if drawtobitmap then
            begin
               bitmaprgn:=CreateRectRgn(startxviewport,startyviewport,
                 startxviewport+viewwidth+1,startyviewport+viewheight+1);
               SelectClipRgn(bitmapdc,bitmaprgn);
            end;

          if drawtoscreen then
            begin
               winrgn:=CreateRectRgn(startxviewport,startyviewport,
                 startxviewport+viewwidth+1,startyviewport+viewheight+1);
               SelectClipRgn(windc,winrgn);
            end;

          inc(x,StartXViewPort);
          inc(y,StartYViewPort);
          col:=CurrentColor;

          if (col>=0) and (col<=high(pens)) then
            begin
               if pens[col]=0 then
                 begin
                    c:=RGB(pal[col].red,pal[col].green,pal[col].blue);
                    pens[col]:=CreatePen(PS_SOLID,1,c);
                 end;
               pen:=pens[col];
            end
          else
            begin
               c:=RGB(pal[col].red,pal[col].green,pal[col].blue);
               pen:=CreatePen(PS_SOLID,1,c);
            end;

          if drawtobitmap then
            begin
               oldpen:=SelectObject(bitmapdc,pen);
               windows.arc(bitmapdc,x-radius,y-radius,x+radius,y+radius,
                 x,y-radius,x,y-radius);
               SelectObject(bitmapdc,oldpen);
            end;

          if drawtoscreen then
            begin
               oldpen:=SelectObject(windc,pen);
               windows.arc(windc,x-radius,y-radius,x+radius,y+radius,
                 x,y-radius,x,y-radius);
               SelectObject(windc,oldpen);
            end;

          if (col<0) or (col>high(pens)) then
            DeleteObject(pen);
          { release clip regions }
          if drawtoscreen then
            begin
               SelectClipRgn(windc,0);
               DeleteObject(winrgn);
            end;
          if drawtobitmap then
            begin
               SelectClipRgn(bitmapdc,0);
               DeleteObject(bitmaprgn);
            end;
          LeaveCriticalSection(graphdrawing);
       end
     else
       begin
          { save state of arc information }
          { because it is not needed for  }
          { a circle call.                }
          move(ArcCall,OriginalArcInfo, sizeof(ArcCall));
          InternalEllipse(X,Y,Radius,Radius,0,360,{$ifdef fpc}@{$endif}DummyPatternLine);
          { restore arc information }
          move(OriginalArcInfo, ArcCall,sizeof(ArcCall));
       end;
 end;

{
Procedure PutImageWin32GUI(X,Y: smallint; var Bitmap; BitBlt: Word); {$ifndef fpc}far;{$endif fpc}
type
  pt = array[0..$fffffff] of word;
  ptw = array[0..2] of longint;
var
  k: longint;
  oldCurrentColor: word;
  oldCurrentWriteMode, i, j, y1, x1, deltaX, deltaX1, deltaY: smallint;
Begin
{$ifdef logging}
  LogLn('putImage at ('+strf(x)+','+strf(y)+') with width '+strf(ptw(Bitmap)[0])+
    ' and height '+strf(ptw(Bitmap)[1]));
  deltaY := 0;
{$endif logging}
  inc(x,startXViewPort);
  inc(y,startYViewPort);
  x1 := ptw(Bitmap)[0]+x; { get width and adjust end coordinate accordingly }
  y1 := ptw(Bitmap)[1]+y; { get height and adjust end coordinate accordingly }

  deltaX := 0;
  deltaX1 := 0;
  k := 3 * sizeOf(Longint) div sizeOf(Word); { Three reserved longs at start of bitmap }
 { check which part of the image is in the viewport }
  if clipPixels then
    begin
      if y < startYViewPort then
        begin
          deltaY := startYViewPort - y;
          inc(k,(x1-x+1)*deltaY);
          y := startYViewPort;
         end;
      if y1 > startYViewPort+viewHeight then
        y1 := startYViewPort+viewHeight;
      if x < startXViewPort then
        begin
          deltaX := startXViewPort-x;
          x := startXViewPort;
        end;
      if x1 > startXViewPort + viewWidth then
        begin
          deltaX1 := x1 - (startXViewPort + viewWidth);
          x1 := startXViewPort + viewWidth;
        end;
    end;
{$ifdef logging}
  LogLn('deltax: '+strf(deltax)+', deltax1: '+strf(deltax1)+',deltay: '+strf(deltay));
{$endif logging}
  case bitBlt of
  end;
  oldCurrentColor := currentColor;
  oldCurrentWriteMode := currentWriteMode;
  currentWriteMode := bitBlt;
  for j:=Y to Y1 do
   Begin
     inc(k,deltaX);
     for i:=X to X1 do
      begin
        currentColor := pt(bitmap)[k];
        directPutPixel(i,j);
        inc(k);
     end;
     inc(k,deltaX1);
   end;
  currentWriteMode := oldCurrentWriteMode;
  currentColor := oldCurrentColor;
end;
}
procedure SetRGBPaletteWin32GUI(colorNum,redValue,greenvalue,
      bluevalue : smallint);

  begin
     if directcolor or (colornum<0) or (colornum>=maxcolor) then
       begin
         _graphresult:=grerror;
         exit;
       end;
     pal[colorNum].red:=redValue;
     pal[colorNum].green:=greenValue;
     pal[colorNum].blue:=blueValue;
     if (colorNum>=0) and (colorNum<=high(pens)) and (pens[colorNum]<>0) then
       begin
          DeleteObject(pens[colorNum]);
          pens[colorNum]:=0;
       end;
  end;

procedure GetRGBPaletteWin32GUI(colorNum : smallint;
      var redValue,greenvalue,bluevalue : smallint);

  begin
     if directcolor or (colornum<0) or (colornum>=maxcolor) then
       begin
         _graphresult:=grerror;
         exit;
       end;
     redValue:=pal[colorNum].red;
     greenValue:=pal[colorNum].green;
     blueValue:=pal[colorNum].blue;
  end;

procedure savestate;

  begin
  end;


procedure restorestate;

  begin
  end;

function WindowProcGraph(Window: HWnd; AMessage:UInt; WParam : WParam;
                    LParam: LParam): Longint; stdcall;

  var
     dc : hdc;
     ps : paintstruct;
     r : rect;
     oldbrush : hbrush;
     oldpen : hpen;
     i : longint;

begin
  WindowProcGraph := 0;

  case AMessage of
    wm_lbuttondown,
    wm_rbuttondown,
    wm_mbuttondown,
    wm_lbuttonup,
    wm_rbuttonup,
    wm_mbuttonup,
    wm_lbuttondblclk,
    wm_rbuttondblclk,
    wm_mbuttondblclk:
    {
    This leads to problem, i.e. the menu etc doesn't work any longer
    wm_nclbuttondown,
    wm_ncrbuttondown,
    wm_ncmbuttondown,
    wm_nclbuttonup,
    wm_ncrbuttonup,
    wm_ncmbuttonup,
    wm_nclbuttondblclk,
    wm_ncrbuttondblclk,
    wm_ncmbuttondblclk:
    }
      begin
         if assigned(mousemessagehandler) then
           WindowProcGraph:=mousemessagehandler(window,amessage,wparam,lparam);
      end;
    wm_notify:
      begin
         if assigned(notifymessagehandler) then
           WindowProcGraph:=notifymessagehandler(window,amessage,wparam,lparam);
      end;
    wm_command:
      if assigned(commandmessagehandler) then
        WindowProcGraph:=commandmessagehandler(window,amessage,wparam,lparam);
    wm_keydown,
    wm_keyup,
    wm_char:
      begin
         if assigned(charmessagehandler) then
           WindowProcGraph:=charmessagehandler(window,amessage,wparam,lparam);
      end;
    wm_paint:
      begin
{$ifdef DEBUG_WM_PAINT}
         inc(wm_paint_count);
{$endif DEBUG_WM_PAINT}
{$ifdef DEBUGCHILDS}
         writeln('Start child painting');
{$endif DEBUGCHILDS}
         if not GetUpdateRect(Window,@r,false) then
           exit;
         EnterCriticalSection(graphdrawing);
         graphrunning:=true;
         dc:=BeginPaint(Window,@ps);
{$ifdef DEBUG_WM_PAINT}
         Writeln(graphdebug,'WM_PAINT in ((',r.left,',',r.top,
           '),(',r.right,',',r.bottom,'))');
{$endif def DEBUG_WM_PAINT}
         if graphrunning then
           {BitBlt(dc,0,0,maxx+1,maxy+1,bitmapdc,0,0,SRCCOPY);}
           BitBlt(dc,r.left,r.top,r.right-r.left+1,r.bottom-r.top+1,bitmapdc,r.left,r.top,SRCCOPY);
         EndPaint(Window,ps);
         LeaveCriticalSection(graphdrawing);
         Exit;
      end;
    wm_create:
      begin
{$ifdef DEBUG_WM_PAINT}
         assign(graphdebug,'wingraph.log');
         rewrite(graphdebug);
{$endif DEBUG_WM_PAINT}
{$ifdef DEBUGCHILDS}
         writeln('Creating window (HWND: ',window,')... ');
{$endif DEBUGCHILDS}
         GraphWindow:=window;
         EnterCriticalSection(graphdrawing);
         dc:=GetDC(window);
{$ifdef DEBUGCHILDS}
         writeln('Window DC: ',dc);
{$endif DEBUGCHILDS}
         bitmapdc:=CreateCompatibleDC(dc);
         savedscreen:=CreateCompatibleBitmap(dc,maxx+1,maxy+1);
         ReleaseDC(window,dc);
         oldbitmap:=SelectObject(bitmapdc,savedscreen);
         windc:=GetDC(window);
         // clear everything
         oldpen:=SelectObject(bitmapdc,GetStockObject(BLACK_PEN));
         oldbrush:=SelectObject(bitmapdc,GetStockObject(BLACK_BRUSH));
         Windows.Rectangle(bitmapdc,0,0,maxx,maxy);
         SelectObject(bitmapdc,oldpen);
         SelectObject(bitmapdc,oldbrush);
         // ... the window too
         oldpen:=SelectObject(windc,GetStockObject(BLACK_PEN));
         oldbrush:=SelectObject(windc,GetStockObject(BLACK_BRUSH));
         Windows.Rectangle(windc,0,0,maxx,maxy);
         SelectObject(windc,oldpen);
         SelectObject(windc,oldbrush);
         // clear font cache
         fillchar(bitmapfonthorizoncache,sizeof(bitmapfonthorizoncache),0);
         fillchar(bitmapfontverticalcache,sizeof(bitmapfontverticalcache),0);

         // clear predefined pens
         fillchar(pens,sizeof(pens),0);
         if assigned(OnGraphWindowCreation) then
           OnGraphWindowCreation;
         LeaveCriticalSection(graphdrawing);
{$ifdef DEBUGCHILDS}
         writeln('done');
         GetClientRect(window,@r);
         writeln('Window size: ',r.right,',',r.bottom);
{$endif DEBUGCHILDS}
      end;
    wm_Destroy:
      begin
         EnterCriticalSection(graphdrawing);
         graphrunning:=false;
         ReleaseDC(GraphWindow,windc);
         SelectObject(bitmapdc,oldbitmap);
         DeleteObject(savedscreen);
         DeleteDC(bitmapdc);
         // release font cache
         for i:=0 to 255 do
           if bitmapfonthorizoncache[i]<>0 then
             DeleteObject(bitmapfonthorizoncache[i]);
         for i:=0 to 255 do
           if bitmapfontverticalcache[i]<>0 then
             DeleteObject(bitmapfontverticalcache[i]);

         for i:=0 to high(pens) do
           if pens[i]<>0 then
             DeleteObject(pens[i]);

         LeaveCriticalSection(graphdrawing);
{$ifdef DEBUG_WM_PAINT}
         close(graphdebug);
{$endif DEBUG_WM_PAINT}
         PostQuitMessage(0);
         Exit;
      end
    else
      WindowProcGraph := DefWindowProc(Window, AMessage, WParam, LParam);
  end;
end;

function WindowProcParent(Window: HWnd; AMessage:UInt; WParam : WParam;
                    LParam: LParam): Longint; stdcall;

begin
  WindowProcParent := 0;
  case AMessage of
    wm_keydown,
    wm_keyup,
    wm_char:
      begin
         if assigned(charmessagehandler) then
           WindowProcParent:=charmessagehandler(window,amessage,wparam,lparam);
      end;
    wm_notify:
      begin
         if assigned(notifymessagehandler) then
           WindowProcParent:=notifymessagehandler(window,amessage,wparam,lparam);
      end;
    wm_command:
      if assigned(commandmessagehandler) then
        WindowProcParent:=commandmessagehandler(window,amessage,wparam,lparam);
    else
      WindowProcParent := DefWindowProc(Window, AMessage, WParam, LParam);
  end;
end;

function WinRegister: Boolean;
var
  WindowClass: WndClass;
begin
  WindowClass.Style := graphwindowstyle;
  WindowClass.lpfnWndProc := WndProc(@WindowProcGraph);
  WindowClass.cbClsExtra := 0;
  WindowClass.cbWndExtra := 0;
  WindowClass.hInstance := system.MainInstance;
  if icon<>0 then
    WindowClass.hIcon := icon
  else
    WindowClass.hIcon := LoadIcon(0, idi_Application);
  WindowClass.hCursor := LoadCursor(0, idc_Arrow);
  WindowClass.hbrBackground := GetStockObject(BLACK_BRUSH);
  if menu<>0 then
    WindowClass.lpszMenuName := MAKEINTRESOURCE(menu)
  else
    WindowClass.lpszMenuName := nil;
  WindowClass.lpszClassName := 'FPCGraphWindow';

  winregister:=RegisterClass(WindowClass) <> 0;
end;

function WinRegisterWithChild: Boolean;
var
  WindowClass: WndClass;
begin
  WindowClass.Style := graphwindowstyle;
  WindowClass.lpfnWndProc := WndProc(@WindowProcParent);
  WindowClass.cbClsExtra := 0;
  WindowClass.cbWndExtra := 0;
  WindowClass.hInstance := system.MainInstance;
  if icon<>0 then
    WindowClass.hIcon := icon
  else
    WindowClass.hIcon := LoadIcon(0, idi_Application);
  WindowClass.hCursor := LoadCursor(0, idc_Arrow);
  WindowClass.hbrBackground := GetStockObject(BLACK_BRUSH);
  if menu<>0 then
    WindowClass.lpszMenuName := MAKEINTRESOURCE(menu)
  else
    WindowClass.lpszMenuName := nil;
  WindowClass.lpszClassName := 'FPCGraphWindowMain';

  WinRegisterWithChild:=RegisterClass(WindowClass) <> 0;
{$ifdef DEBUGCHILDS}
  writeln('Main window successfully registered: WinRegisterWithChild is ',WinRegisterWithChild);
{$endif DEBUGCHILDS}
  if WinRegisterWithChild then
    begin
       WindowClass.Style := CS_HREDRAW or CS_VREDRAW;
       WindowClass.lpfnWndProc := WndProc(@WindowProcGraph);
       WindowClass.cbClsExtra := 0;
       WindowClass.cbWndExtra := 0;
       WindowClass.hInstance := system.MainInstance;
       WindowClass.hIcon := 0;
       WindowClass.hCursor := LoadCursor(0, idc_Arrow);
       WindowClass.hbrBackground := GetStockObject(BLACK_BRUSH);
       WindowClass.lpszMenuName := nil;
       WindowClass.lpszClassName := 'FPCGraphWindowChild';
       WinRegisterWithChild:=RegisterClass(WindowClass)<>0;
{$ifdef DEBUGCHILDS}
       writeln('Child window registered: WinRegisterWithChild is ',WinRegisterWithChild);
{$endif DEBUGCHILDS}
    end;
end;

var
   // here we can force the creation of a maximized window }
   extrastyle : cardinal;

 { Create the Window Class }
function WinCreate : HWnd;
var
  hWindow: HWnd;
begin
  WinCreate:=0;
  if UseChildWindow then
    begin
       ParentWindow:=CreateWindow('FPCGraphWindowMain', windowtitle,
                  WS_OVERLAPPEDWINDOW or WS_CLIPCHILDREN or extrastyle, longint(CW_USEDEFAULT), 0,
                  maxx+ChildOffset.Left+ChildOffset.Right+1+
                    2*GetSystemMetrics(SM_CXFRAME),
                  maxy+ChildOffset.Top+ChildOffset.Bottom+1+
                    2*GetSystemMetrics(SM_CYFRAME)+
                  GetSystemMetrics(SM_CYCAPTION),
                  0, 0, system.MainInstance, nil);
       if ParentWindow<>0 then
         begin
            ShowWindow(ParentWindow, SW_SHOW);
            UpdateWindow(ParentWindow);
         end
       else
         exit;
       hWindow:=CreateWindow('FPCGraphWindowChild',nil,
                  WS_CHILD, ChildOffset.Left,ChildOffset.Top,
                  maxx+1,maxy+1,
                  ParentWindow, 0, system.MainInstance, nil);
       if hwindow<>0 then
         begin
            ShowWindow(hwindow, SW_SHOW);
            UpdateWindow(hwindow);
         end
       else
         exit;
       WinCreate:=hWindow;
    end
  else
    begin
       hWindow:=CreateWindow('FPCGraphWindow', windowtitle,
                  ws_OverlappedWindow or extrastyle, longint(CW_USEDEFAULT), 0,
                  maxx+1+2*GetSystemMetrics(SM_CXFRAME),
                  maxy+1+2*GetSystemMetrics(SM_CYFRAME)+
                  GetSystemMetrics(SM_CYCAPTION),
                  0, 0, system.MainInstance, nil);
       if hWindow <> 0 then
         begin
            ShowWindow(hWindow, SW_SHOW);
            UpdateWindow(hWindow);
            WinCreate:=hWindow;
         end;
    end;
end;

const
   winregistered : boolean = false;

function MessageHandleThread(p : pointer) : DWord;StdCall;

  var
     AMessage: Msg;

  begin
     if not(winregistered) then
       begin
          if UseChildWindow then
            begin
               if not(WinRegisterWithChild) then
                 begin
                    MessageBox(0, 'Window registration failed', nil, mb_Ok);
                    ExitThread(1);
                 end;
            end
          else
            begin
               if not(WinRegister) then
                 begin
                    MessageBox(0, 'Window registration failed', nil, mb_Ok);
                    ExitThread(1);
                 end;
            end;
          winregistered:=true;
       end;
     GraphWindow:=WinCreate;
     if longint(GraphWindow) = 0 then begin
       MessageBox(0, 'Window creation failed', nil, mb_Ok);
       ExitThread(1);
     end;
     while longint(GetMessage(@AMessage, 0, 0, 0))=longint(true) do
       begin
          TranslateMessage(AMessage);
          DispatchMessage(AMessage);
       end;
     MessageHandleThread:=0;
  end;

procedure InitWin32GUI16colors;

  var
     threadexitcode : longint;
  begin
     getmem(pal,sizeof(RGBrec)*maxcolor);
     move(DefaultColors,pal^,sizeof(RGBrec)*maxcolor);
     if (IntCurrentMode=mMaximizedWindow16) or
       (IntCurrentMode=mMaximizedWindow256) or
       (IntCurrentMode=mMaximizedWindow32k) or
       (IntCurrentMode=mMaximizedWindow64k) or
       (IntCurrentMode=mMaximizedWindow16M) then
       extrastyle:=ws_maximize
     else
       extrastyle:=0;
     { start graph subsystem }
     InitializeCriticalSection(graphdrawing);
     graphrunning:=false;
     MessageThreadHandle:=CreateThread(nil,0,@MessageHandleThread,
       nil,0,MessageThreadID);
     repeat
       GetExitCodeThread(MessageThreadHandle,@threadexitcode);
     until graphrunning or (threadexitcode<>STILL_ACTIVE);
     if threadexitcode<>STILL_ACTIVE then
        _graphresult := grerror;
  end;

procedure CloseGraph;

  begin
     If not isgraphmode then
       begin
         _graphresult := grnoinitgraph;
         exit
       end;
     if UseChildWindow then
       begin
          { if the child window isn't destroyed }
          { the main window can't be closed     }
          { I don't know any other way (FK)     }
          PostMessage(GraphWindow,wm_destroy,0,0);
          PostMessage(ParentWindow,wm_destroy,0,0)
       end
     else
       PostMessage(GraphWindow,wm_destroy,0,0);

     PostThreadMessage(MessageThreadHandle,wm_quit,0,0);
     WaitForSingleObject(MessageThreadHandle,Infinite);
     CloseHandle(MessageThreadHandle);
     DeleteCriticalSection(graphdrawing);
     freemem(pal,sizeof(RGBrec)*maxcolor);

     MessageThreadID := 0;
     MessageThreadHandle := 0;
     isgraphmode := false;
  end;

procedure LineWin32GUI(X1, Y1, X2, Y2: smallint); {$ifndef fpc}far;{$endif fpc}

  var X, Y :           smallint;
      deltax, deltay : smallint;
      d, dinc1, dinc2: smallint;
      xinc1          : smallint;
      xinc2          : smallint;
      yinc1          : smallint;
      yinc2          : smallint;
      i              : smallint;
      Flag           : Boolean; { determines pixel direction in thick lines }
      NumPixels      : smallint;
      PixelCount     : smallint;
      OldCurrentColor: Word;
      swtmp          : smallint;
      TmpNumPixels   : smallint;
      col : longint;
      pen,oldpen : hpen;

 begin
    if graphrunning then
      begin
         {******************************************}
         {  SOLID LINES                             }
         {******************************************}
         if lineinfo.LineStyle = SolidLn then
           Begin
              { Convert to global coordinates. }
              x1 := x1 + StartXViewPort;
              x2 := x2 + StartXViewPort;
              y1 := y1 + StartYViewPort;
              y2 := y2 + StartYViewPort;
              { if fully clipped then exit... }
              if ClipPixels then
                begin
                   if LineClipped(x1,y1,x2,y2,StartXViewPort, StartYViewPort,
                     StartXViewPort+ViewWidth, StartYViewPort+ViewHeight) then
                       exit;
                  If LineInfo.Thickness=NormWidth then
                   Begin
                      EnterCriticalSection(graphdrawing);
                      {
                      if currentwritemode<>normalput then
                        begin
                           case currentwritemode of
                              XORPut:
                                begin
                                   SetROP2(windc,R2_XORPEN);
                                   SetROP2(bitmapdc,R2_XORPEN);
                                end;
                              AndPut:
                                begin
                                   SetROP2(windc,R2_MASKPEN);
                                   SetROP2(bitmapdc,R2_MASKPEN);
                                end;
                              OrPut:
                                begin
                                   SetROP2(windc,R2_MERGEPEN);
                                   SetROP2(bitmapdc,R2_MERGEPEN);
                                end;
                           end;
                        end;
                      }
                      col:=RGB(pal[CurrentColor].red,pal[CurrentColor].green,pal[CurrentColor].blue);
                      pen:=CreatePen(PS_SOLID,1,col);
                      if pen=0 then
                        writeln('Pen konnte nicht erzeugt werden!');

                      oldpen:=SelectObject(windc,pen);
                      MoveToEx(windc,x1,y1,nil);
                      Windows.LineTo(windc,x2,y2);
                      SetPixel(windc,x2,y2,col);
                      SelectObject(windc,oldpen);

                      oldpen:=SelectObject(bitmapdc,pen);
                      MoveToEx(bitmapdc,x1,y1,nil);
                      Windows.LineTo(bitmapdc,x2,y2);
                      SetPixel(bitmapdc,x2,y2,col);
                      SelectObject(bitmapdc,oldpen);

                      DeleteObject(pen);
                      {
                      if currentwritemode<>normalput then
                        begin
                           SetROP2(windc,R2_COPYPEN);
                           SetROP2(bitmapdc,R2_COPYPEN);
                        end;
                      }
                      LeaveCriticalSection(graphdrawing);
                   end
                 else
                  { Thick width lines }
                   begin
                     { Draw the pixels }
                      for i := 1 to numpixels do
                        begin
                         { all depending on the slope, we can determine         }
                         { in what direction the extra width pixels will be put }
                         If Flag then
                           Begin
                             DirectPutPixelClip(x-1,y);
                             DirectPutPixelClip(x,y);
                             DirectPutPixelClip(x+1,y);
                           end
                         else
                           Begin
                             DirectPutPixelClip(x, y-1);
                             DirectPutPixelClip(x, y);
                             DirectPutPixelClip(x, y+1);
                           end;
                         if d < 0 then
                           begin
                             d := d + dinc1;
                             x := x + xinc1;
                             y := y + yinc1;
                           end
                         else
                           begin
                             d := d + dinc2;
                             x := x + xinc2;
                             y := y + yinc2;
                           end;
                        end;
                   end;
                 end;
           end
          else
       {******************************************}
       {  begin patterned lines                   }
       {******************************************}
           Begin
             { Convert to global coordinates. }
             x1 := x1 + StartXViewPort;
             x2 := x2 + StartXViewPort;
             y1 := y1 + StartYViewPort;
             y2 := y2 + StartYViewPort;
             { if fully clipped then exit... }
             if ClipPixels then
              begin
              if LineClipped(x1,y1,x2,y2,StartXViewPort, StartYViewPort,
                  StartXViewPort+ViewWidth, StartYViewPort+ViewHeight) then
                     exit;
              end;

             OldCurrentColor := CurrentColor;
             PixelCount:=0;
             if y1 = y2 then
                   Begin
                    { Check if we must swap }
                if x1 >= x2 then
                      Begin
                        swtmp := x1;
                        x1 := x2;
                        x2 := swtmp;
                      end;
                if LineInfo.Thickness = NormWidth then
                     Begin
                      for PixelCount:=x1 to x2 do
                            { optimization: PixelCount mod 16 }
                            if LinePatterns[PixelCount and 15] = TRUE then
                             begin
                               DirectPutPixel(PixelCount,y2);
                             end;
                     end
                    else
                     Begin
                      for i:=-1 to 1 do
                            Begin
                              for PixelCount:=x1 to x2 do
                                { Optimization from Thomas - mod 16 = and 15 }
                                {this optimization has been performed by the compiler
                                 for while as well (JM)}
                                if LinePatterns[PixelCount and 15] = TRUE then
                                  begin
                                        DirectPutPixelClip(PixelCount,y2+i);
                                  end;
                            end;
                     end;
               end
             else
             if x1 = x2 then
                  Begin
                   { Check if we must swap }
                   if y1 >= y2 then
                     Begin
                       swtmp := y1;
                       y1 := y2;
                       y2 := swtmp;
                     end;
                   if LineInfo.Thickness = NormWidth then
                     Begin
                       for PixelCount:=y1 to y2 do
                           { compare if we should plot a pixel here , compare }
                           { with predefined line patterns...                 }
                           if LinePatterns[PixelCount and 15] = TRUE then
                             begin
                           DirectPutPixel(x1,PixelCount);
                             end;
                     end
                   else
                     Begin
                       for i:=-1 to 1 do
                            Begin
                              for PixelCount:=y1 to y2 do
                              { compare if we should plot a pixel here , compare }
                              { with predefined line patterns...                 }
                                if LinePatterns[PixelCount and 15] = TRUE then
                                  begin
                                    DirectPutPixelClip(x1+i,PixelCount);
                                  end;
                            end;
                     end;
                  end
             else
                  Begin
                    oldCurrentColor := CurrentColor;
                    { Calculate deltax and deltay for initialisation }
                    deltax := abs(x2 - x1);
                    deltay := abs(y2 - y1);

                    { Initialize all vars based on which is the independent variable }
                    if deltax >= deltay then
                      begin

                        Flag := FALSE;
                        { x is independent variable }
                        numpixels := deltax + 1;
                        d := (2 * deltay) - deltax;
                        dinc1 := deltay Shl 1;
                        dinc2 := (deltay - deltax) shl 1;
                        xinc1 := 1;
                        xinc2 := 1;
                        yinc1 := 0;
                        yinc2 := 1;
                     end
                   else
                     begin

                       Flag := TRUE;
                       { y is independent variable }
                       numpixels := deltay + 1;
                       d := (2 * deltax) - deltay;
                       dinc1 := deltax Shl 1;
                       dinc2 := (deltax - deltay) shl 1;
                       xinc1 := 0;
                       xinc2 := 1;
                       yinc1 := 1;
                       yinc2 := 1;
                     end;

                   { Make sure x and y move in the right directions }
                   if x1 > x2 then
                     begin
                       xinc1 := - xinc1;
                       xinc2 := - xinc2;
                     end;
                   if y1 > y2 then
                     begin
                       yinc1 := - yinc1;
                       yinc2 := - yinc2;
                     end;

                   { Start drawing at <x1, y1> }
                   x := x1;
                   y := y1;

                   If LineInfo.Thickness=ThickWidth then

                    Begin
                      TmpNumPixels := NumPixels-1;
                      { Draw the pixels }
                      for i := 0 to TmpNumPixels do
                        begin
                            { all depending on the slope, we can determine         }
                            { in what direction the extra width pixels will be put }
                              If Flag then
                                 Begin
                                   { compare if we should plot a pixel here , compare }
                                   { with predefined line patterns...                 }
                                   if LinePatterns[i and 15] = TRUE then
                                     begin
                                       DirectPutPixelClip(x-1,y);
                                       DirectPutPixelClip(x,y);
                                       DirectPutPixelClip(x+1,y);
                                     end;
                                 end
                              else
                                 Begin
                                   { compare if we should plot a pixel here , compare }
                                   { with predefined line patterns...                 }
                                   if LinePatterns[i and 15] = TRUE then
                                    begin
                                      DirectPutPixelClip(x,y-1);
                                      DirectPutPixelClip(x,y);
                                      DirectPutPixelClip(x,y+1);
                                    end;
                                 end;
                          if d < 0 then
                                begin
                                  d := d + dinc1;
                                  x := x + xinc1;
                                  y := y + yinc1;
                                end
                          else
                                begin
                          d := d + dinc2;
                          x := x + xinc2;
                          y := y + yinc2;
                                end;
                       end;
                   end
                  else
                   Begin
                    { instead of putting in loop , substract by one now }
                    TmpNumPixels := NumPixels-1;
                   { NormWidth }
                    for i := 0 to TmpNumPixels do
                    begin
                         if LinePatterns[i and 15] = TRUE then
                           begin
                                 DirectPutPixel(x,y);
                           end;
                    if d < 0 then
                        begin
                          d := d + dinc1;
                          x := x + xinc1;
                          y := y + yinc1;
                        end
                    else
                        begin
                          d := d + dinc2;
                          x := x + xinc2;
                          y := y + yinc2;
                        end;
                    end;
                   end
               end;
       {******************************************}
       {  end patterned lines                     }
       {******************************************}
              { restore color }
              CurrentColor:=OldCurrentColor;
          end;
    end;
 end;  { Line }

{ multipage support could be done by using more than one background bitmap }
procedure SetVisualWin32GUI(page: word);

  begin
  end;

procedure SetActiveWin32GUI(page: word);
  begin
  end;

function queryadapterinfo : pmodeinfo;

  var
     mode: TModeInfo;
     ScreenWidth,ScreenHeight : longint;
     ScreenWidthMaximized,ScreenHeightMaximized : longint;

  procedure SetupWin32GUIDefault;

    begin
       mode.DirectPutPixel:={$ifdef fpc}@{$endif}DirectPutPixel16Win32GUI;
       mode.PutPixel:={$ifdef fpc}@{$endif}PutPixel16Win32GUI;
       mode.GetPixel:={$ifdef fpc}@{$endif}GetPixel16Win32GUI;
       mode.HLine := {$ifdef fpc}@{$endif}HLine16Win32GUI;
       mode.SetRGBPalette := {$ifdef fpc}@{$endif}SetRGBPaletteWin32GUI;
       mode.GetRGBPalette := {$ifdef fpc}@{$endif}GetRGBPaletteWin32GUI;
       mode.SetVisualPage := {$ifdef fpc}@{$endif}SetVisualWin32GUI;
       mode.SetActivePage := {$ifdef fpc}@{$endif}SetActiveWin32GUI;
       mode.InitMode := {$ifdef fpc}@{$endif}InitWin32GUI16colors;
       mode.OuttextXY:={$ifdef fpc}@{$endif}OuttextXYWin32GUI;
       mode.VLine := {$ifdef fpc}@{$endif}VLine16Win32GUI;
       // mode.circle := {$ifdef fpc}@{$endif}Circle16Win32GUI;
       // doesn't work yet
       // mode.Line:={$ifdef fpc}@{$endif}LineWin32GUI;
    end;

  begin
     SaveVideoState:={$ifdef fpc}@{$endif}savestate;
     RestoreVideoState:={$ifdef fpc}@{$endif}restorestate;
     { we must take care of the border and caption }
     ScreenWidth:=GetSystemMetrics(SM_CXSCREEN)-
       2*GetSystemMetrics(SM_CXFRAME);
     ScreenHeight:=GetSystemMetrics(SM_CYSCREEN)-
       2*GetSystemMetrics(SM_CYFRAME)-
       GetSystemMetrics(SM_CYCAPTION);
     { for maximozed windows it's again different }
     { here we've only a caption }
     ScreenWidthMaximized:=GetSystemMetrics(SM_CXFULLSCREEN);
     { neither GetSystemMetrics(SM_CYFULLSCREEN nor     }
     { SystemParametersInfo(SPI_GETWORKAREA)            }
     { takes a hidden try into account :( FK            }
     ScreenHeightMaximized:=GetSystemMetrics(SM_CYFULLSCREEN);

     QueryAdapterInfo := ModeList;
     { If the mode listing already exists... }
     { simply return it, without changing    }
     { anything...                           }
     if assigned(ModeList) then
       exit;
     { the first one becomes the standard mode }
     if (ScreenWidth>=640) and (ScreenHeight>=480) then
       begin
          InitMode(mode);
          mode.DriverNumber:= VGA;
          mode.HardwarePages:= 0;
          mode.ModeNumber:=VGAHi;
          mode.ModeName:='640 x 480 x 16 Win32GUI';
          mode.MaxColor := 16;
          mode.PaletteSize := mode.MaxColor;
          mode.DirectColor := FALSE;
          mode.MaxX := 639;
          mode.MaxY := 479;
          SetupWin32GUIDefault;
          mode.XAspect := 10000;
          mode.YAspect := 10000;
          AddMode(mode);
       end;
     if (ScreenWidth>=640) and (ScreenHeight>=200) then
       begin
          InitMode(mode);
          { now add all standard VGA modes...       }
          mode.DriverNumber:= VGA;
          mode.HardwarePages:= 0;
          mode.ModeNumber:=VGALo;
          mode.ModeName:='640 x 200 x 16 Win32GUI';
          mode.MaxColor := 16;
          mode.PaletteSize := mode.MaxColor;
          mode.DirectColor := FALSE;
          mode.MaxX := 639;
          mode.MaxY := 199;
          SetupWin32GUIDefault;
          mode.XAspect := 4500;
          mode.YAspect := 10000;
          AddMode(mode);
       end;
     if (ScreenWidth>=640) and (ScreenHeight>=350) then
       begin
          InitMode(mode);
          mode.DriverNumber:= VGA;
          mode.HardwarePages:= 0;
          mode.ModeNumber:=VGAMed;
          mode.ModeName:='640 x 350 x 16 Win32GUI';
          mode.MaxColor := 16;
          mode.PaletteSize := mode.MaxColor;
          mode.DirectColor := FALSE;
          mode.MaxX := 639;
          mode.MaxY := 349;
          SetupWin32GUIDefault;
          mode.XAspect := 7750;
          mode.YAspect := 10000;
          AddMode(mode);
       end;
     if (ScreenWidth>=640) and (ScreenHeight>=400) then
       begin
          InitMode(mode);
          mode.DriverNumber:= VESA;
          mode.HardwarePages:= 0;
          mode.ModeNumber:=m640x400x256;
          mode.ModeName:='640 x 400 x 256 Win32GUI';
          mode.MaxColor := 256;
          mode.PaletteSize := mode.MaxColor;
          mode.DirectColor := FALSE;
          mode.MaxX := 639;
          mode.MaxY := 399;
          SetupWin32GUIDefault;
          mode.XAspect := 8333;
          mode.YAspect := 10000;
          AddMode(mode);
       end;
     if (ScreenWidth>=640) and (ScreenHeight>=480) then
       begin
          InitMode(mode);
          mode.DriverNumber:= VESA;
          mode.HardwarePages:= 0;
          mode.ModeNumber:=m640x480x256;
          mode.ModeName:='640 x 480 x 256 Win32GUI';
          mode.MaxColor := 256;
          mode.PaletteSize := mode.MaxColor;
          mode.DirectColor := FALSE;
          mode.MaxX := 639;
          mode.MaxY := 479;
          SetupWin32GUIDefault;
          mode.XAspect := 10000;
          mode.YAspect := 10000;
          AddMode(mode);
       end;
     { add 800x600 only if screen is large enough }
     If (ScreenWidth>=800) and (ScreenHeight>=600) then
       begin
          InitMode(mode);
          mode.DriverNumber:= VESA;
          mode.HardwarePages:= 0;
          mode.ModeNumber:=m800x600x16;
          mode.ModeName:='800 x 600 x 16 Win32GUI';
          mode.MaxColor := 16;
          mode.PaletteSize := mode.MaxColor;
          mode.DirectColor := FALSE;
          mode.MaxX := 799;
          mode.MaxY := 599;
          SetupWin32GUIDefault;
          mode.XAspect := 10000;
          mode.YAspect := 10000;
          AddMode(mode);
          InitMode(mode);
          mode.DriverNumber:= VESA;
          mode.HardwarePages:= 0;
          mode.ModeNumber:=m800x600x256;
          mode.ModeName:='800 x 600 x 256 Win32GUI';
          mode.MaxColor := 256;
          mode.PaletteSize := mode.MaxColor;
          mode.DirectColor := FALSE;
          mode.MaxX := 799;
          mode.MaxY := 599;
          SetupWin32GUIDefault;
          mode.XAspect := 10000;
          mode.YAspect := 10000;
          AddMode(mode);
       end;
     { add 1024x768 only if screen is large enough }
     If (ScreenWidth>=1024) and (ScreenHeight>=768) then
       begin
          InitMode(mode);
          mode.DriverNumber:= VESA;
          mode.HardwarePages:= 0;
          mode.ModeNumber:=m1024x768x16;
          mode.ModeName:='1024 x 768 x 16 Win32GUI';
          mode.MaxColor := 16;
          mode.PaletteSize := mode.MaxColor;
          mode.DirectColor := FALSE;
          mode.MaxX := 1023;
          mode.MaxY := 767;
          SetupWin32GUIDefault;
          mode.XAspect := 10000;
          mode.YAspect := 10000;
          AddMode(mode);
          InitMode(mode);
          mode.DriverNumber:= VESA;
          mode.HardwarePages:= 0;
          mode.ModeNumber:=m1024x768x256;
          mode.ModeName:='1024 x 768 x 256 Win32GUI';
          mode.MaxColor := 256;
          mode.PaletteSize := mode.MaxColor;
          mode.DirectColor := FALSE;
          mode.MaxX := 1023;
          mode.MaxY := 768;
          SetupWin32GUIDefault;
          mode.XAspect := 10000;
          mode.YAspect := 10000;
          AddMode(mode);
       end;
     { add 1280x1024 only if screen is large enough }
     If (ScreenWidth>=1280) and (ScreenHeight>=1024) then
       begin
          InitMode(mode);
          mode.DriverNumber:= VESA;
          mode.HardwarePages:= 0;
          mode.ModeNumber:=m1280x1024x16;
          mode.ModeName:='1280 x 1024 x 16 Win32GUI';
          mode.MaxColor := 16;
          mode.PaletteSize := mode.MaxColor;
          mode.DirectColor := FALSE;
          mode.MaxX := 1279;
          mode.MaxY := 1023;
          SetupWin32GUIDefault;
          mode.XAspect := 10000;
          mode.YAspect := 10000;
          AddMode(mode);
          InitMode(mode);
          mode.DriverNumber:= VESA;
          mode.HardwarePages:= 0;
          mode.ModeNumber:=m1280x1024x256;
          mode.ModeName:='1280 x 1024 x 256 Win32GUI';
          mode.MaxColor := 256;
          mode.PaletteSize := mode.MaxColor;
          mode.DirectColor := FALSE;
          mode.MaxX := 1279;
          mode.MaxY := 1023;
          SetupWin32GUIDefault;
          mode.XAspect := 10000;
          mode.YAspect := 10000;
          AddMode(mode);
       end;
     { at least we add a mode with the largest possible window }
      InitMode(mode);
      mode.DriverNumber:= VESA;
      mode.HardwarePages:= 0;
      mode.ModeNumber:=mLargestWindow16;
      mode.ModeName:='Largest Window x 16';
      mode.MaxColor := 16;
      mode.PaletteSize := mode.MaxColor;
      mode.DirectColor := FALSE;
      mode.MaxX := ScreenWidth-1;
      mode.MaxY := ScreenHeight-1;
      SetupWin32GUIDefault;
      mode.XAspect := 10000;
      mode.YAspect := 10000;
      AddMode(mode);
      InitMode(mode);
      mode.DriverNumber:= VESA;
      mode.HardwarePages:= 0;
      mode.ModeNumber:=mLargestWindow256;
      mode.ModeName:='Largest Window x 256';
      mode.MaxColor := 256;
      mode.PaletteSize := mode.MaxColor;
      mode.DirectColor := FALSE;
      mode.MaxX := ScreenWidth-1;
      mode.MaxY := ScreenHeight-1;
      SetupWin32GUIDefault;
      mode.XAspect := 10000;
      mode.YAspect := 10000;
      AddMode(mode);
      { .. and a maximized window }
      InitMode(mode);
      mode.DriverNumber:= VESA;
      mode.HardwarePages:= 0;
      mode.ModeNumber:=mMaximizedWindow16;
      mode.ModeName:='Maximized Window x 16';
      mode.MaxColor := 16;
      mode.PaletteSize := mode.MaxColor;
      mode.DirectColor := FALSE;
      mode.MaxX := ScreenWidthMaximized-1;
      mode.MaxY := ScreenHeightMaximized-1;
      SetupWin32GUIDefault;
      mode.XAspect := 10000;
      mode.YAspect := 10000;
      AddMode(mode);
      InitMode(mode);
      mode.DriverNumber:= VESA;
      mode.HardwarePages:= 0;
      mode.ModeNumber:=mMaximizedWindow256;
      mode.ModeName:='Maximized Window x 256';
      mode.MaxColor := 256;
      mode.PaletteSize := mode.MaxColor;
      mode.DirectColor := FALSE;
      mode.MaxX := ScreenWidthMaximized-1;
      mode.MaxY := ScreenHeightMaximized-1;
      SetupWin32GUIDefault;
      mode.XAspect := 10000;
      mode.YAspect := 10000;
      AddMode(mode);
  end;

begin
  InitializeGraph;
  charmessagehandler:=nil;
  mousemessagehandler:=nil;
  commandmessagehandler:=nil;
  notifymessagehandler:=nil;
  OnGraphWindowCreation:=nil;
end.
